#
#  This file contains the scripted 'shims' for Keith's
# Loader code. These wrape the other functions defined in
# scripts in this directory to protect the DBMS from the
# pagan Atmospheric Scientists. :-)
#
#    Keith disavows these comments, but giggles none-the-less.
#
#
#  GCM_Get_CVariable ( name units stype )
#
#    Returns the oid of the Compound Variable being asked for.
#
proc GCM_Get_CVariable { compoundName instanceType CellUnits } {
	set VarOID [ Add_Variable_With_Abstract $compoundName $instanceType 1 $CellUnits "diagnostic" ]
	return [ Add_Compound $compoundName 1 linearArray [ list [ list $VarOID ] ] ]
}
#
# GCM_Get_Axis axisName axisType units positionOfOrigin gridSpacing gridDimensions
#
proc GCM_Get_Axis { axisName axisType units positionOfOrigin gridSpacing gridDimensions } {
	return [ Add_Axis $axisName $axisType $units $gridSpacing $positionOfOrigin 0.005 $gridDimensions ]
}
#
# GCM_Get_Grid count { List of Axis OIDS }
#
proc GCM_Get_Grid { productGridRank AxisGrids } {	
	global MD
	if { ( $productGridRank == 4 ) } {
		set OriginCnr "South-West-Early-Sea_Level"
	} else {
		set OriginCnr "South-West-Early"
	}
	set NameStub "GCMGrid$productGridRank"
	
	set Qry "
SELECT COUNT(*)
  FROM GCMGrid CV
 WHERE CV.gridName LIKE '$NameStub%';"
	set Cnt [ mi_select_first_row $MD(connection) $Qry ]
	if { $Cnt == "MI_ERROR" } {
		return "ERROR: Failed on $Qry"
	}
	if { $Cnt == "" } {
		set Cnt 1
	} else {
		incr Cnt
	}
	set NameStub "$NameStub.$Cnt"
	return [ Add_Grid $NameStub $productGridRank $OriginCnr "T" $AxisGrids ]
}
# 
# GCM_Get_Dataset fileName variableOID gridOID metaDataOID
#
proc GCM_Get_Dataset { fileName variableOID gridOID metaDataOID offSet } {
	global MD
#
# Get The DataSetId
#
	set Qry "
SELECT II.Data_Set_Id,
	    II.Beginning
  FROM Identification_Information II
 WHERE II.oid = '$metaDataOID'::oid;"

	set Results [ mi_select_first_row $MD(connection) $Qry ]
	if { $Results == "MI_ERROR" } {
		return "ERROR: Failed on $Qry"
	}
	mi_query_finish $MD(connection)
	set Data_Set_Id [ lindex $Results 0 ]
	set Beginning [ lindex $Results 1 ]
#
# . Now. From the Grid, I need to get the TEMPORAL Axis' origin (which is hours
#   since the start of the Experiment), the spacing (which is constant, but
#   I may as well look it up anyway) and the count.
#
	set Qry "
SELECT AX.positionOfOrigin::integer,
	    AX.gridDimensions,
	    AX.gridSpacing::integer
  FROM GCMAxis AX, GCMGridAxisRes GAR
 WHERE deref(GAR.productGrid).oid = '$gridOID'::oid
	AND deref(GAR.axis).oid = AX.oid
	AND AX.units = 'Hours'
	AND AX.axisType = 'Temporal'
	AND AX.axisName = 'TIMESTEP';"

	set Results [ mi_select_first_row $MD(connection) $Qry ]
	if { $Results == "MI_ERROR" } {
	   return "ERROR: Failed on $Qry"
	} elseif { $Results == "" } {
		return "ERROR: This Grid doesn't exist!
$Qry"
	}
	mi_query_finish $MD(connection)

	set positionOfOrigin [ lindex $Results 0 ]
	set gridDimensions [ lindex $Results 1 ]
	set gridSpacing [ lindex $Results 2 ]

	set Qry "
return ('$Beginning'::timestamp::abstime + ' @ $positionOfOrigin hours')::timestamp;"
	set NewBeginning [ mi_select_first_row $MD(connection) $Qry ]
	if { $NewBeginning == "MI_ERROR" } {
		return "ERROR: Failed on $Qry"
	}
	set NewBeginning [ lindex $NewBeginning 0 ]
#
	set Hours [ expr $gridDimensions * $gridSpacing ] 
	set Qry "
return ('$NewBeginning'::timestamp::abstime + ' @ $Hours hours')::timestamp;"
	set NewEnding [ mi_select_first_row $MD(connection) $Qry ]
	if { $NewEnding == "MI_ERROR" } {
	   return "ERROR: Failed on $Qry"
	}
	set NewEnding [ lindex $NewEnding 0 ]
#
# And now the Process.
#
#
#  The MD(procOID) must be populated in the init section.
#
	set AttrMods ""
	lappend AttrMods [ list Beginning "$NewBeginning" ]
	lappend AttrMods [ list Ending "$NewEnding" ]
	lappend AttrMods [ list Status_Progress "COMPLETE" ]
	lappend AttrMods [ list Lineage_Process $MD(ProcOID) ]

	return [ Add_Data $Data_Set_Id $fileName "NET-CDF" $offSet $gridOID $variableOID $AttrMods ]
}
#
#  GCM_Get_Metadata - 
# 
#       This takes a Data_Set_Id and creates a new second level, TOP_LEVEL 
#    II Tuple with all the stuff set up correctly.
# 
proc GCM_Get_Metadata { { Args } } {
	global MD
	set DataSetId ""
	foreach i $Args {
		set DataSetId "$DataSetId $i"
	}
	set DataSetId [ string trim $DataSetId ]
	set Qry "
SELECT oid
  FROM Identification_Information II
 WHERE II.Data_Set_Id = 'UCLA_GCM'
   AND deref(deref(II.Lineage_Process).Process).processName = 'Add Data Set'
   AND deref(deref(II.Lineage_Process).Process).processVersion = 1;"

	set RetVal ""
	set RetVal [ mi_select_first_row $MD(connection) $Qry ]

	if { ( $RetVal == "MI_ERROR" ) || ( $RetVal == "" ) } {
		return "ERROR: Failed on $Qry"
	} 

#
# Now get the interesting stuff.
#
set SklowerOID [ Add_Contact "Mr Keith Sklower" "Electronics Research Lab, UC Berkeley" \
"Staff Programmer" 0 "\[\]" "" "" "" "510-642-9587" "510-642-5153" "sklower@cs.Berkeley.EDU" "" ]
#
set UCBOID [ Add_Citation "Electronics Research Lab, UC Berkeley" "" "Random data for test pruposes" \
"1" "binary" "" "" "Berkeley, CA" "University of California, Berkeley" "" ""]
#
#
set PGBOID [ Add_Contact "Mr Paul Brown" "Electronics Research Lab, UC Berkeley" "Staff Programmer" \
2 "\[ UC Berkeley, Soda Hall \[Mail Drop 1776\], Room 380 \]" "Berkeley" "CA" "94703" \
"510-642-7379" "310-642-5153" "pbrown@postgres.Berkeley.EDU" "" ]
#
set AddDataSetOID [ Add_Process "Add Data Set" 1 $PGBOID "" "" ]
set AddDataSet1OID [ Add_Parameter $AddDataSetOID 1 "Lengthy File Name" "file_name" "" ]
#
set DSInstOID [ Add_ProcInstance $AddDataSetOID $PGBOID { "http://s2k-ftp.CS.Berkeley.EDU:8000/" } ""  ]
#

	set Qry "
INSERT INTO Identification_Information
( Data_Set_Id, Citation, Beginning, Status_Progress,
  Status_Update_Frequency, Spatial_Domain, Access_Constraints,
  Use_Constraints, Contact_Information, Data_Set_Credit,
  Security_Code, Native_Data_Set_Environment, Lineage_Process )
SELECT '$DataSetId',
       ref(C),
       '1982-09-01 12:00:00'::timestamp,
       'TOP_LEVEL',
       '',
       '(-180,-90,180,90)'::Box::Poly,
       'PUBLIC',
       'PERMISSION_REQUEST',
       ref(Ct),
       'The Atmospheric Modelling Team at UCLA',
       'READ_ONLY',
       'Data generated from a GCM run on a IBM SP-1',
       ref(LP)
  FROM Citation C, Contact_Information Ct, Lineage_Process LP
 WHERE Ct.oid  = '$SklowerOID'::oid
   AND C.oid   = '$UCBOID'::oid
   AND LP.oid  = '$DSInstOID'::oid;"

	set RetVal [ mi_exec $MD(connection) $Qry ]
	if { $RetVal == "MI_ERROR" } {
		puts "ERROR: Failed on $Qry"
		CloseDB
		exit
	}
	set Oid [ mi_result_object_id $MD(connection) ]

	return $Oid
}
#
proc CloseDB {} {
	global MD
	mi_close $MD(connection)
}
#
# 
#
