#
#  This script describes the functions which will handle the
# Lineage portion of the schema's functionality. 
#
#
proc Add_Process { { processName "" } { processVersion "" } { ContOID "" } \
{ processScript "" } { dateAdded "" } } {

	global MD
#
# . Need some minimums
#
	if { ($processName == "") || ($ContOID == "" ) } {
		return "ERROR: I need processName and a contact OID"
	}
	if { $dateAdded == "" } {
		set dateAdded [ lindex [ \
		mi_select_first_row $MD(connection) \
			"return current_timestamp;" ] 0 ]
	}
#
# . If there is no processVersion specified, get one.
#
	if { $processVersion == "" } {
		set processVersion [ mi_select_rows $MD(connection) "
SELECT Max(P.processVersion)
  FROM Process P
 WHERE P.processName = '$processName';" ]

		if { ( $processVersion == "MI_ERROR" ) || ( $processVersion == "") } {
			set processVersion 1
		}
	}
#
# . Check if such a process exists
#
	set Qry "
SELECT P.oid
  FROM Process P
 WHERE P.processName    = '$processName'
   AND P.processVersion = '$processVersion'::integer;"

	set RetVal [ mi_select_first_row $MD(connection) $Qry ]

	if { ( $RetVal != "MI_ERROR") && ($RetVal != "" ) } {
		return $RetVal
	}

#
# . OK It doesn't exist. So create it!
#
	set InsQry "
INSERT INTO Process
( processName, processVersion, Contact_Information, 
  processScript, dateAdded )
SELECT '$processName',
       '$processVersion'::integer,
       ref(CT),
       '$processScript',
       '$dateAdded'::timestamp
  FROM Contact_Information CT
 WHERE CT.oid = '$ContOID'::oid;"

	set RetVal [ mi_exec $MD(connection) $InsQry ]

	if { $RetVal == "MI_ERROR" } {
		return "ERROR: Error on Query $InsQry "
	}
#
# . Now find the oid and give it back.
#
	set RetVal [ mi_select_first_row $MD(connection) $Qry ]

	if { $RetVal == "MI_ERROR" } {
		return "ERROR: Failed with $Qry"
	} 

	return $RetVal
}
#
# Add_Parameter - Needs to use an OID from the Process table
#
proc Add_Parameter { ProcOID paramSeq paramName paramType paramDefault } {
#
#
#
	global MD
#
# . check for minimums
#
	if { ($ProcOID == "") || ($paramSeq == "") || ($paramName == "") } {
		return "ERROR: usage Add_Parameter ProcOID paramSeq paramName ?paramType? ?paramDefault? "
	}
#
# .Does it exist?
#
	set Qry "
SELECT P.oid
  FROM Parameter P
 WHERE P.paramName = '$paramName'
   AND P.paramSeq  = '$paramSeq'::integer
   AND deref(P.Process).oid = '$ProcOID'::oid;"

	set RetVal [ mi_select_first_row $MD(connection) $Qry ]

	if { ($RetVal != "MI_ERROR") && ($RetVal != "") } {
		return $RetVal
	}
#
# . OK. Not here. Add it.
#
	set InsQry "
INSERT INTO Parameter
( Process, paramSeq, paramName, paramType, paramDefault )
SELECT ref(P),
       '$paramSeq'::integer,
       '$paramName',
       '$paramType',
	    '$paramDefault'
  FROM Process P
 WHERE P.oid = '$ProcOID'::oid;"

	if { [ mi_exec $MD(connection) $InsQry ] == "MI_ERROR" } {
		return "ERROR: Failed on $InsQry"
	}
#
# . Now try the initial again
#
   set RetVal [ mi_select_first_row $MD(connection) $Qry ]

   if { $RetVal == "MI_ERROR" } {
		return "ERROR: Failed on $Qry"
   }

	return $RetVal
}
##
## This is a function which combines the ones above
##
proc Add_ProcInstance { ProcOID ContOID ParamList { TimeStamp "" } } {
#
#  ProcOID - This is the OID of the tuple in the Process table which this
#            is an instance of.
#
#  ParamList - This is an ordered list of the arguements for this Process.
#
#  ContOID   - This is the OID of the tuple in the contact table which
#              identifies the contact person.
#
#  paramName paramType paramDefault
#
#   They are added in the order that they are supplied.
#
#   Note that if you ask for a Process and version which exist, the
#   OID returned is for this Process tuple- the Params aren't checked!
#
	global MD
#
# . Do some checking 
#
	set Qry "
SELECT count(*)
  FROM Parameter LP
 WHERE deref(LP.Process).oid = '$ProcOID'::oid;"

	set PCnt [ mi_select_first_row $MD(connection) $Qry ]
	if { $PCnt == "MI_ERROR" } {
		return "ERROR: Query failed :$Qry"	
	} elseif { $PCnt != [ llength $ParamList ] } {
		return "ERROR: Expected [ llength $ParamList ] Params found $PCnt "
	}
#
# . Found Proc. Now look for an Instance with these Params.
#
	set Qry " 
SELECT LP.oid
  FROM Lineage_Process LP
 WHERE deref(LP.Process).oid = '$ProcOID'::oid
   AND deref(LP.processContact).oid = '$ContOID'::oid"

	set cnt 0
	foreach i $ParamList {
#
# . Need to find the Param for this. I hope this gets re-written!!
#
		incr cnt
		set Qry "$Qry
  AND EXISTS ( SELECT *
       FROM Lineage_Param LP$cnt
      WHERE deref(LP$cnt.Process).oid = deref(LP.Process).oid
        AND deref(LP$cnt.Process).oid = '$ProcOID'::oid
        AND deref(LP$cnt.Parameter).paramSeq = LP$cnt.argSeq 
        AND deref(LP$cnt.Parameter).paramSeq = '$cnt'::integer
        AND LP$cnt.argSeq   = '$cnt'::integer
        AND LP$cnt.argValue = '$i') "
	}
	set Qry "$Qry;"

	set RetVal [ mi_select_first_row $MD(connection) $Qry ]
   if { $RetVal == "MI_ERROR" } {
		return "ERROR: Failed on $Qry"
   }
	puts " $Qry"
   if { $RetVal != "" } {
		return $RetVal
   }
#
# . Nope. None there. So Add It.
#
	if { $TimeStamp == "" } {
		set TimeStamp [ lindex [ mi_select_first_row $MD(connection) \
"return current_timestamp;" ] 0 ]
	}
	set InsQry "
INSERT INTO Lineage_Process
( Process, processDate, processContact)
  SELECT ref(P),
         '$TimeStamp'::timestamp,
         ref(CT)
    FROM Process P, Contact_Information CT
   WHERE P.oid = '$ProcOID'::oid
     AND CT.oid = '$ContOID'::oid;"

	set RetVal [ mi_exec $MD(connection) $InsQry ]
	if { $RetVal == "MI_ERROR" } {
		return "ERROR: Failed Query $InsQry with $RetVal"
	}

	set NewQry "
SELECT LP.oid
  FROM Lineage_Process LP
 WHERE deref(LP.Process).oid = '$ProcOID'::oid
   AND deref(LP.processContact).oid = '$ContOID'::oid;"

	set LPOID [ mi_select_first_row $MD(connection) $NewQry ]

	if { ($LPOID == "" ) || ($LPOID == "MI_ERROR") } {
		return "ERROR: Failed on $NewQry"
	}

	set cnt 0 
	foreach i $ParamList {
		incr cnt
		set InsQry "
INSERT INTO Lineage_Param
( Parameter, Process, Lineage_Process, argSeq, argValue)
SELECT ref(PA),
       ref(P),
       ref(LP),
       PA.paramSeq,
       '$i'
  FROM Process P, Parameter PA, Lineage_Process LP
 WHERE P.oid  = '$ProcOID'::oid
   AND deref(PA.Process).oid = '$ProcOID'::oid
   AND P.oid = deref(PA.Process).oid
   AND PA.paramSeq = '$cnt'::integer
   AND LP.oid = '$LPOID'::oid;"

		set RetVal [ mi_exec $MD(connection) $InsQry ]

		if { $RetVal == "MI_ERROR" } {
 			mi_exec $MD(connection) "abort transaction;"
#
# . Cleanup
#
 			mi_exec $MD(connection) "DELETE FROM Lineage_Process WHERE oid = '$LPOID'::oid;" 
			return "ERROR: Failed on $InsQry"	
		}
	}
#
# . Now return the oid
#
	return $LPOID
}
