#
#  This file contains the scripts which add the Meta Data information
#  to the schema. Meta Data in this context is everything which relates
#  to the 
#
#  Add_Citation - Adds to the Citation table and returns the
#					  oid just inserted. If this exists ( same
#					  Originator, and Title/Edition) then the
#					  function returns the OID.
#
proc Add_Citation { { Originator ""} { Publication_DateTime ""} {Title ""}
{ Edition "" } { GeoSpatial_Data_Form ""}  { Series_Name ""}
{ Issue_Identification ""} { Publication_Place ""}
{ Publisher ""} {Other_Citation_Details ""}  { Online_Linkage ""} } {
#
#  Each of the params has the same name as the Attribute in the 
# Citation relation.
#
	global MD
	set COid ""
#
#
# . Check to see if the tuple identified by this logical key already exists
#
	if { ( $Originator == "" ) || ( $Title == "" ) || ( $Edition == "" ) } {
		return "ERROR: Must specify (at least) Originator Title Edition"
	}

	set Qry "
SELECT C.oid
  FROM Citation C
 WHERE C.Originator = '$Originator'
	AND C.Title		= '$Title'
	AND C.Edition	 = '$Edition';"

	set COid [ mi_select_first_row $MD(connection) $Qry]

	if { $COid != "" } {
		return $COid
	}
#
# . Check and set some stuff
#
	if { $Publication_DateTime == "" } {
		set Publication_DateTime [ lindex [ \
			mi_select_first_row $MD(connection) \
			"return current_timestamp;" ] 0 ]
	}
	
#
# . Don't have this tuple. Insert and get back.
#
	set InsQry "
INSERT INTO Citation
( Originator, Publication_DateTime, Title, Edition, 
  Geospatial_Data_Form, Series_Name, Issue_Identification, 
  Publication_Place, Publisher, 
  Other_Citation_Details, Online_Linkage )
VALUES
( '$Originator', '$Publication_DateTime' ,'$Title', '$Edition',
  '$GeoSpatial_Data_Form', '$Series_Name', '$Issue_Identification',
  '$Publication_Place', '$Publisher', '$Other_Citation_Details',
  '$Online_Linkage' );"

	set Ret [ mi_exec $MD(connection) $InsQry ]

	if { $Ret == "MI_ERROR" } {
		return "ERROR: Error on Insert to Citation"
	}

	set COid [ mi_result_object_id $MD(connection) ]
	mi_query_finish $MD(connection)

	if { $COid != "" } {
		return $COid
	}

	return "ERROR: Error on re-select of new oid value"
}
#
#  Add_Contact	- Takes information about a contact person and returns
#						the oid refering to that person's Tuple - if it exists.
#						If it doesn't exist, then a new Tuple is inserted and
#						the oid for the new Tuple is inserted.
#
proc Add_Contact {{Primary_Contact_Person ""} {Primary_Contact_Organization ""}
{ Contact_Position "" } { Address_Row_Count 0} { Address "" } { City "" }
{ State "" } { Postal_Code "" } { Contact_Phone "" } { Contact_Fax "" }
{ Contact_E_Mail "" } { Hours_of_Service "" } } {
#
#  Each of these parameters is named after its related attribute in the
# Contacts relation.
#
	global MD
	set CtOid ""
	
	if { ( $Primary_Contact_Person == "" ) || 
		  ( $Primary_Contact_Organization == "" ) ||
		  ( $Contact_Position == "" ) } {
		return "ERROR: Must specify Primary_Contact_Person, Primary_Contact_Organization, Contact_Position"
	}

	set Qry "
SELECT C.oid
  FROM Contact_Information C
 WHERE C.Primary_Contact_Person		 = '$Primary_Contact_Person'
	AND C.Primary_Contact_Organization = '$Primary_Contact_Organization'
	AND C.Contact_Position				 = '$Contact_Position';"

	set CtOid [ mi_select_first_row $MD(connection) $Qry ]

	if { $CtOid != "" } {
		return $CtOid
	}
#
# . now build the Insert Statement and do it, before trying the Query
#	again.
#
	set InsQry "
INSERT INTO Contact_Information
( Primary_Contact_Person, Primary_Contact_Organization, 
  Contact_Position, Address_Row_Count, Address,
  City, State_or_Province, Postal_Code, Contact_Phone, Contact_Fax,
  Contact_E_Mail, Hours_of_Service )
VALUES
( '$Primary_Contact_Person', '$Primary_Contact_Organization',
  '$Contact_Position', '$Address_Row_Count', '$Address',
  '$City', '$State', '$Postal_Code', '$Contact_Phone', '$Contact_Fax',
  '$Contact_E_Mail', '$Hours_of_Service' );"

	set RetVal [ mi_exec $MD(connection) $InsQry ]

	if { $RetVal == "MI_ERROR" } {
		return "ERROR: Error on insert into Contact with $InsQry"
	}
#
# . Having done the insert, try again.
#

	set CtOid [ mi_result_object_id $MD(connection) ]
	mi_query_finish $MD(connection)

	if { $CtOid != "" } {
		return $CtOid
	} 

	return "ERROR: Error re-selecting the Row just inserted."
}
#
#  Add_IdentInfo - Adds to the Identification_Information and Lineage
#						tables.
#
proc Add_Data { DataSetId File format offSet GridOID CellOID AttrModList } {
#
#  DataSetId - This is the Identification_Information.Data_Set_Id which will
#				  be replicated in the new tuple excepting the changes 
#				  specified in the AttrModList.
#
#  File		- This is the file in the local file-system space which is the
#				  objective of the whole bug hunt.
#
#  format	 - This is the format that the file's in.
#
#  offSet	 - This is the offSet within the file to seek to for the start of
#				  of Data.
#
#  GridOID	- This is the OID of the grid. This is gotten from a previous
#				  call to Add_Grid (or whatever).
#
#  CellOID	- This is the OID of the Compound Cell. This is gotten with
#				  a previous Call to Add_CompoundVariable (or whatever.)
#
#  AttrModList - You may specify a list of attributes to be changed and the
#					 values they are to be changed to in this call. Only
#					 attributes which are part of Identification Information
#					 may be changed in this way.
#
#					 If a ref() attribute is to be changed, then the OID of
#					 the new target tuple is to be passed in.
#

	global MD
	
#
# . There must be something in the AttrModList
#
	if { [ llength $AttrModList ] == 0 } {
		return "ERROR: you must alters something, preferably Spatial_Domain or Beginning"
	}
#
# . All that I'll do is perform the inserts. The rules on the table should
#	deal with integrity. Note that each of the lists below must be in
#	*precisely* the correct order.
#
	set SelList { { Data_Set_Id text} { Beginning timestamp} \
{ Ending timestamp } { Status_Progress text }  { Status_Update_Frequency text }\
 { Spatial_Domain Poly } { Access_Constraints text } { Use_Constraints text } \
{ Data_Set_Credit text } { Security_Code text } \
{ Native_Data_Set_Environment text } }
#
# . List of ref()
#
	set InsQryHdr "
INSERT INTO Identification_Information
( Data_Set_Id, Beginning, Ending, Status_Progress, 
  Status_Update_Frequency, Spatial_Domain, Access_Constraints, 
  Use_Constraints, Data_Set_Credit, Security_Code, 
  Native_Data_Set_Environment, 
  Citation, Contact_Information, Lineage_Process)"
#
# . This is going to be slow, but fuck it.
#
# If 'attribute_in_question' is in 'AttrModList' then change it. Otherwise
# select the default.
#
	set InsQrySel "SELECT"
	foreach i $SelList {
		set Attr [ lindex $i 0 ]
		set Type [ lindex $i 1 ]
		set Val [ GetAttrMod $Attr $AttrModList ]
		if { $Val != ":-)" } {
					 	set InsQrySel "$InsQrySel
		'$Val'::$Type,"
		  	} else {
					 	set InsQrySel "$InsQrySel
		II.$Attr,"
		}
		  }
#
# . Now for the other three; Citation, Contact_Information, Lineage_Process
#
	set Val [ GetAttrMod Citation $AttrModList ]
	if { $Val != ":-)" } {
		set InsQrySel "$InsQrySel
		ref(Cit),"
		set CitOID $Val
	} else {
		set InsQrySel "$InsQrySel
		 II.Citation,"
		set CitOID 0
	}
		  set Val [ GetAttrMod Contact_Information $AttrModList ]
		  if { $Val != ":-)" } {
					 set InsQrySel "$InsQrySel
		ref(CI),"
					 set ContactOID $Val
		  } else {
					 set InsQrySel "$InsQrySel
		II.Contact_Information,"
					 set ContactOID 0
		  }
		  set Val [ GetAttrMod Lineage_Process $AttrModList ]
		  if { $Val != ":-)" } {
					 set InsQrySel "$InsQrySel
		ref(LP)"
					 set LPOID $Val
		  } else {
					 set InsQrySel "$InsQrySel
		II.Lineage_Process"
					 set LPOID 0
		  }
#
# . OK. Now build the FROM clause
#
	set InsQryFrom ""
	set InsQryFrom "$InsQryFrom
	 FROM Identification_Information II"
	if { $CitOID != 0 } { set InsQryFrom "$InsQryFrom, Citation Cit" }
	if { $ContactOID != 0 }  { set InsQryFrom "$InsQryFrom, Contact_Information CI" }
	if { $LPOID != 0 }  { set InsQryFrom "$InsQryFrom, Lineage_Process LP" }
#
# . And the WHERE Clause
#
	set InsQryWhere ""
	set InsQryWhere "$InsQryWhere
	WHERE II.Data_Set_Id = '$DataSetId'
	  AND deref(deref(II.Lineage_Process).Process).processName = 'Add Data Set'
	  AND deref(deref(II.Lineage_Process).Process).processVersion = 1"

	if { $CitOID != 0 }  {
		set InsQryWhere "$InsQryWhere
	  AND Cit.oid = '$CitOID'::oid"
	}
	if { $ContactOID != 0 }  {
		set InsQryWhere "$InsQryWhere
	  AND CI.oid = '$ContactOID'::oid"
	}
	if { $LPOID != 0 }  {
					 set InsQryWhere "$InsQryWhere
	  AND LP.oid = '$LPOID'::oid"
		  }

	set InsQryWhere "$InsQryWhere;"
	set InsQry "$InsQryHdr
$InsQrySel
$InsQryFrom
$InsQryWhere"

#
# . I've got the query. Now do it and check that a row was inserted. Note
#	that more than one row amounts to an error, so I'll need to ensure
#	integrity with a transaction.
#

	set RetVal ""
 	set RetVal [ mi_exec $MD(connection) $InsQry ]
	if { $RetVal == "MI_ERROR" } {
		return "ERROR: Failed on $InsQry"
	}
	set RowCnt 0 
	set RowCnt [ mi_result_row_count $MD(connection) ]
	if { $RowCnt != 1 } {
		mi_exec $MD(connection) "abort transaction;"
		return "ERROR: Query inserted $RowCnt rows;
$InsQry"
	} else {
		set RowOID [ mi_result_object_id $MD(connection) ]
	}
	mi_query_finish $MD(connection)

#
# . Now do the MetaData insert 
#
	set BSOIns "
INSERT INTO Metadata_Reference_Information
(  Metadata_Date, Metadata_Contact, Metadata_Standard_Name, 
	Metadata_Standard_Version,
	Metadata_Time_Convention, Metadata_Access_Constraints, 
	Metadata_Use_Constraints, Metadata_Security_Code )
VALUES
(  current_time, Concat(username(), Concat('@',hostname())), 'FGDC', 
	'June 8th 1994','Standard','PUBLIC','UNRESTRICTED','NONE');"

	set RetVal ""
	set RetVal [ mi_exec $MD(connection) $BSOIns ]
	if { $RetVal == "MI_ERROR" } {
		mi_exec $MD(connection) "abort transaction;"
		return "ERROR: Failed on $BSOIns"
	} else {
		set MDOID [ mi_result_object_id $MD(connection) ]
	}
	mi_query_finish $MD(connection)
#
# . Now do the Insert into the Metadata table
#
	set InsQry "
INSERT INTO MetaData
( Identification_Information, Metadata_Reference_Information )
SELECT ref(II),
		 ref(MR)
  FROM Identification_Information II, Metadata_Reference_Information MR
 WHERE II.oid = '$RowOID'::oid
	AND MR.oid = '$MDOID'::oid;"

 	set RetVal [ mi_exec $MD(connection) $InsQry ]
	if { $RetVal == "MI_ERROR" } {
		mi_exec $MD(connection) "abort transaction;"
		return "ERROR: Failed on $InsQry"
	} else {
		set MDROID [ mi_result_object_id $MD(connection) ]
	}
	mi_query_finish $MD(connection)
	
#
# . And - finally, the BigSurObject
#
	set InsQry "
INSERT INTO BigSurDataObject
( MetaData, BigSurExternal, fileOffSet, storageFormat, BigSurGrid, CellVector )
SELECT ref(MD),
	'$File',
	'$offSet'::integer,
       '$format',
		 ref(GR),
		 ref(CV)
  FROM MetaData MD, GCMCompoundVariable CV, GCMGrid GR
 WHERE MD.oid = '$MDROID'::oid
	AND CV.oid = '$CellOID'::oid
	AND GR.oid = '$GridOID'::oid;"

	set RetVal ""
	set RetVal [ mi_exec $MD(connection) $InsQry ]
	set RowCnt [ mi_result_row_count $MD(connection) ]

	if { ($RetVal == "MI_ERROR") || ( $RowCnt == 0 ) } {
		mi_exec $MD(connection) "abort transaction;"
		return "ERROR: Failed on $InsQry"
	} else {
		set TopOID [ mi_result_object_id $MD(connection) ]
	}
	mi_query_finish $MD(connection)
	return $TopOID

}
