#
#   variable.tcl - This file contains the routines which interact with
#                  the BigSur Variable components of the Schema. 
#
#                  The basic pattern here is also to take the information
#                  which identifies a tuple, and return an oid corresponding
#                  to a tuple with these attribute values.
#
#  Add_Variable_With_Abstract - Takes a set of attributes which describe a
#                               variable value and return an oid corresponding
#                               to a row with these variables.
#                               
proc Add_Variable_With_Abstract { attributeName instancePrimitive 
elementsPerValue attributeUnits class { Abstract "" } } {

 global MD
 set Exists ""

 set Qry "
select distinct V.oid
  from Variable V
 where V.attributeName = '$attributeName'
   and V.instancePrimitive = '$instancePrimitive'
   and V.elementsPerValue = $elementsPerValue
   and V.attributeUnits = '$attributeUnits';"
#
#  Note: Ignoring Class for the time being. It's not being supplied by
#        the model.
#

 set Exists [ mi_select_rows $MD(connection) $Qry ]

 if { $Exists != "" } { return $Exists }

 set InsertQry "
insert into Variable
( attributeName, instancePrimitive, elementsPerValue, 
  attributeUnits, class, Abstract, Supplemental )
values
 ( '$attributeName', '$instancePrimitive', $elementsPerValue,
   '$attributeUnits', '$class', '$Abstract', 'Added Automatically on user request.');"

  mi_exec $MD(connection) $InsertQry

 set Exists [ mi_select_rows $MD(connection) $Qry ]

 if { $Exists != "" } { return $Exists }

 return "ERROR: unable to add non-existant Variable $Qry"
}
#
#  Add_Compound - Take attributes describing a Compound Variable and a set
#                 of oids describing basic variables. This function returns
#                 an oid corresponding to the row.
#
proc Add_Compound { compoundName variableCount indexingScheme Variable } {

 global MD
 set Exists ""
#
# Check the list length
#
  if { ( $variableCount != [ llength $Variable ] ) &&
       ( $variableCount > 0 ) } {
      return "ERROR: variable count doesn't match the number of variables passed or no variables passed.' }
#
#  Write the 'find' query
#
set CmpQry "
select C.oid
  from GCMCompoundVariable C
 where C.compoundName = '$compoundName'
   and C.variableCount = $variableCount
   and C.indexingScheme = '$indexingScheme'"

set Seq 1
set Qry $CmpQry
foreach i $Variable {
 if { [ llength $i ] == 1 } {
   set Uncert 0.005
   set ScalA 1.0
   set ScalB 1.0 
 } else if { [ llength $i ] == 4 } {
   set Uncert [ lindex $i 1 ]
   set ScalA [ lindex $i 2 ]
   set ScalB [ lindex $i 3 ] 
 } else {
   return "ERROR: Must specify VarOID, Uncertainty, Scalar A, Scalar B or just VarOID" }

 set Var [ lindex $i 0 ]

 set Qry "$Qry
   and EXISTS ( select 1
                  from GCMVariableCompRes R$Seq, GCMVariable V$Seq
                 where deref(R$Seq.GCMCompoundVariable).oid = C.oid
                   and deref(R$Seq.GCMVariable).oid = V$Seq.oid
                   and V$Seq.oid = '$Var'::oid
                   and R$Seq.variableUncertainty = '$Uncert'::double precision
                   and R$Seq.variableScalarA     = '$ScalA'::double precision
                   and R$Seq.variableScalarB     = '$ScalB'::double precision
                   and R$Seq.seqNum = $Seq)"
}
set Qry "$Qry;"
puts $Qry

 set Exists [ mi_select_rows $MD(connection) $Qry ]
 if { $Exists != "" } { return $Exists }
#
# Check that a compound of this name but with another
# variable coniguration doesn't exist
#
 set CmpQry "$CmpQry;"
 set Exists ""
 set Exists [ mi_select_rows $MD(connection) $CmpQry ]

 if { $Exists != "" } { return "ERROR: Another Compound Variable with this Name exists $CmpQry" }

 set InsertQry "
insert into CompoundVariable
( compoundName, variableCount, indexingScheme )
values
( '$compoundName', $variableCount, '$indexingScheme');"

 mi_exec $MD(connection) $InsertQry
 set CmpOID [ mi_select_rows $MD(connection) $CmpQry ]
 if { $CmpOID == "" } { return "ERROR: Unable to add compound variable record $CmpQry" }

 set Seq 1
 set InsQry ""
 foreach i $Variable {

 if { [ llength $i ] == 1 } { 
   set Uncert 0.005 
   set ScalA 1.0
   set ScalB 1.0 } else if { [ llength $i ] == 4 } { 
   set Uncert [ lindex $i 1 ] 
   set ScalA [ lindex $i 2 ]
   set ScalB [ lindex $i 3 ] } else {
   return "ERROR: Must specify VarOID, Uncertainty, Scalar A, Scalar B or just VarOID" }

   set VarOID [ lindex $i 0 ]
#
#  Each of the Variable pairs must be;  { VariableOID uncertainty }
#
#   The default uncertainty is 0.005
#  
 set InsQry "$InsQry
insert into GCMVariableCompRes
(  GCMCompoundVariable, GCMVariable, variableUncertainty, 
   variableScalarA, variableScalarB, seqNum )
select ref(C),
       ref(V),
       '$Uncert'::double precision,
       '$ScalA'::double precision,
       '$ScalB'::double precision,
       $Seq
  from GCMCompoundVariable C, GCMVariable V
 where C.oid = '$CmpOID'::oid
   and V.oid = '$VarOID'::oid;"
}

 mi_exec $MD(connection) $InsQry

 set Exists [ mi_select_rows $MD(connection) $Qry ]
 if { $Exists == "" } { return "ERROR: Unable to Add_Compound and none found to match $Qry" }

 return $Exists

}
