/*
 *  dlobh.h  - distributed large object handle.
 *
 *   These are the types etc that are used throughout the
 *   new dlobh type.
 *
*/
#include "mi.h"
#include <stdio.h>
#include <math.h>
#include <memory.h>
#include <strings.h>
#include <sys/param.h>

#define BUFSIZE 8129
#define SERVER_NAME_LEN 32
#define SERVER_HOST_LEN 128
#define USER_NAME_LEN 32
#define DBASE_NAME_LEN 32
#define PASSWORD_LEN 32

#define COLON_DELIM ':'
#define SLASH_DELIM '/'
#define END_PARAM ')'

#define EXTERNAL_FILE 'X'
#define LARGE_OBJECT  'I'
#define LARGE_TEXT    'T'

#define VALID_DLOB_T(X)  (((X=='X')||(X=='I')||(X=='T'))?1:0)
#define NULL_IF_NULL(X) ((X[0]=='\0')?NULL:X)
#define IFNULL_NULL(X) ((X==NULL)?"\0":X)

#define DLOBH_URL_TAG "dlobh"

#define ONE_K 1024
#define EIGHT_K 8192
/*
 * These are the names of Schemas and Relations defined and created by
 * the scripts ../sql. 
 */
#define DLOBH_SCHEMA 		   "SYS_DLOBH"
#define LOCAL_EXTERNAL_RELATION	   "Local_ExternalFiles"
#define LOCAL_LARGETEXT_RELATION   "Local_LargeTexts"
#define LOCAL_LARGEOBJECT_RELATION "Local_LargeObjects"

#define DLOBH_INCOMPLETE_STATUS "INCOMPLETE"
#define DLOBH_ERROR_STATUS "DERROR"

/*
 *  Right! Bogus stuff number 1. I'm just going to assign arbitrary lengths to
 *  each of these attributes rather than do the right thing and use VARLENA
 *  types, which is more complicated.  Version 2. of the dlobh blade
 *  will support this.
*/
typedef struct {
	char ServerName[SERVER_NAME_LEN];  /* Server name to connect to.     */
	char ServerHost[SERVER_HOST_LEN];  /* Host name running this server. */
	int  ServerPort;                   /* Port number to attempt the     */
					   /* connection to.                 */
	char DBaseUserName[USER_NAME_LEN]; /* User Name needed to get to     */
					   /* this database                  */
	char DBaseName[DBASE_NAME_LEN];    /* Name of the database to connect*/
					   /* to on the server with the user */
	char DBPassword[PASSWORD_LEN];     /* Password to apply. May be      */
					   /* encrypted.                     */
	char LOBHData[MAXPATHLEN + 1];     /* Text of LOHandle. Could be an  */
	                                   /* external file                  */
        char ObjectType;                   /* Type of LO. eXternal, Internal */
					   /* or Text.                       */
        long  mtime;                       /* Used for external files to     */
	long  size;                        /* ensure consistency.            */
} dlobh;
/*
 * These are the functions I'll use.
 */
mi_text * dlobh_test ( /* mi_text *  */);
dlobh * dlobh_in( /*  mi_text *  */);
mi_text * dlobh_out(  /* dlobh *  */);
mi_boolean dlobh_equal ( /*  dlobh *, dlobh * */);
mi_boolean dlobh_not_equal ( /*  dlobh *, dlobh *);
void dlobhGetConnectionInfo ( /* dlobh *, MI_CONNECTION_INFO * */);
void dlobhGetDatabaseInfo (  /* dlobh *, MI_DATABASE_INFO * */);
void dlobhPrint ( /* dlobh * */ );
