/*
 *  DistDLOBH.c
 *
 *  Paul Brown
 *
 *   The long October of the soul (Oct 1994)
 *
 *   Right. In a nutshell - which is where you'll be once all of this
 *  has been explained to you - these are the two functions which
 *  implement the DLOBH in the server, *without* *Tcl* !
 *
 *   They are to be used;
 *
 *   * select DLOBHToFile(R.Raster_Data, '/scratch/FF') from Rasters R where etc;
 *
 *   And;
 *   * insert into Rasters
 *     ( Raster_Data, ... )
 *     values
 *     ( FileToDLOBH('/scratch/FF.new', ... );
 *
 *
 *  Note that the DLOBHToFile is not much good without the stuff that lives
 *  in the ../BE_Exec source tree. That's where the work gets done; all this
 *  does is to execl() the code created there, which does the work. We need
 *  to jump through this hoop because an Illustra BE function can only
 *  do stuff with the Database/Server it's connected to, and can't connect
 *  to another. mi_server_connect(foo,bar,foobar) in a BE function has
 *  unpredictable results.
 *
 *
 */
#include "DistDLOBH.h"
mi_text *
DLOBHToFile( pDLOBH, szFileName )
dlobh * pDLOBH;
mi_text  * szFileName;
{
 
	int 	MyPipe[2],
		n;
	pid_t	pid;
	char    szMesgBuf[100],
		cPipeBuf,
		szRetVal[MAXPATHLEN],
		* szFileStub,
	        * RetVal,
                * pchDLOBH,
		szMiHome[32],
 		szPipe[3];
        MI_CONNECTION * pstrDefaultConnection;

	pstrDefaultConnection = mi_open(NULL,NULL,NULL); /* BE default conn  */

	if (pDLOBH->ServerPort < 1)
	{
	   sprintf(szMesgBuf, "Can't get DLOBH. (DERROR or INCOMPLETE)");
           mi_db_error_raise(pstrDefaultConnection, MI_EXCEPTION, szMesgBuf);
        }

        pchDLOBH   = mi_text_to_string(dlobh_out(pDLOBH));
	szFileStub = mi_text_to_string(szFileName);

	if ((pipe(MyPipe)) < 0)
	{
		sprintf(szMesgBuf, "pipe() command failed.");
		mi_db_error_raise(pstrDefaultConnection, MI_EXCEPTION, szMesgBuf);
	}

	if ((pid = fork()) < 0)                /* ERROR - fork failed        */
	{
		sprintf(szMesgBuf, "fork() command failed.");
		mi_db_error_raise(pstrDefaultConnection, MI_EXCEPTION, szMesgBuf);

        } else if (pid == 0) {                 /* Child process              */
		/*
	 	* Need to construct the call and execle() it
         	*/            
	 	close(MyPipe[0]);
		if (strcpy(szMiHome,mi_getsysparam(mi_sysname((char *)NULL),
				 	  "MI_HOME"))==NULL)
		{
		   sprintf(szMesgBuf, "Can't get MI_HOME environment variable");
		   mi_db_error_raise(pstrDefaultConnection, MI_EXCEPTION, szMesgBuf);
                }
		sprintf(szRetVal,"%s/bin/get_dlobh",szMiHome);
                sprintf(szPipe,"%d",MyPipe[1]);
		if (n=execl(szRetVal, "get_dlobh", szPipe, pchDLOBH, szFileStub, (char *) 0) < 0)
                {
		   sprintf(szMesgBuf, "exec");
                   mi_db_error_raise(pstrDefaultConnection, MI_EXCEPTION, szMesgBuf);
                }

        } else {                               /* Parent process              */
	/*
	 * Suck at the pipe for all it's worth until something interesting
         * arrives. Should be a filename.
	 */
         close(MyPipe[1]);
         memset(szRetVal,NULL,sizeof(szRetVal));
         while(n=read(MyPipe[0],szRetVal,sizeof(szRetVal))!=NULL);
       }
  return mi_string_to_text(szRetVal);
}

/*
 * FileToDLOBH - This function takes a connection and a filename
 *               on the client machine. It registers the new Large
 *               Object, and returns a dlobh URL as the handle to
 *               that dlobh.
 *
 * 
 */
dlobh *
FileToDLOBH ( pFileName, chType, pTxtStrgMgr, pTxtArchMgr) 
mi_text * pFileName;
mi_char1 chType;
mi_text * pTxtStrgMgr, * pTxtArchMgr;
/*
 *mi_text * pTxtStrgMgr;
 * mi_text * pTstArchMgr;
*/
{

  dlobh             *  pLocalNewDLOBH = NULL;
  MI_CONNECTION_INFO   bConnInfo;
  MI_DATABASE_INFO     bDataBInfo;
  MI_CONNECTION *      pLocalConnection;
  char   szErrorText[ONE_K];
  char   szQueryText[ONE_K + 3 * MAXPATHLEN];/* might take up complete MAXPATHLEN */
  char   szRelationName[32];
  char   * pServerName;
  char   * pHostName;
  char   * pPortId;
  int    nPortId = 0;
  char  * pResult,
	* pLocalFileName,
	* pchTxtStrgMgr,
	* pchTstArchMgr,
        * pChar;
  mi_text * szErr;
  struct stat  StatBuf;

  if ((chType=='I')||(chType=='T')||(chType=='X'))/* Legal types of LO   */
  {
    pLocalConnection = mi_open(NULL, NULL, NULL); /* Get the conn, Spock! */
    szErrorText[0] = '\0';
    bzero(szQueryText,ONE_K + MAXPATHLEN);

    pLocalFileName = mi_text_to_string(pFileName);

    pchTxtStrgMgr = mi_text_to_string(pTxtStrgMgr);
    pchTstArchMgr = mi_text_to_string(pTxtArchMgr);
    
    if (mi_get_database_info(pLocalConnection,
                             &bDataBInfo)!=MI_ERROR)
    {
       /*
        * Right! Because I can't seem to use the mi_get_connection_info()
        * in the BE to get anything but bogus *rubbish* (localhost,0,NULL)
        * I'm going to get it the *hard way*. (Yeah!)
        */
       pServerName=mi_sysname(NULL);
       if ((pPortId=mi_getsysparam(pServerName,"MI_PORT"))==NULL)
       {
         nPortId = 7531;      /* This is the default port, which is going to */
                              /* be the one used when none is defined in     */
                              /* MI_SYSPARAMS.                               */
       } else {
         nPortId=atoi(pPortId);
       }
       pHostName=mi_getsysparam(pServerName,"MI_HOST");
       /*
        * There is a valid connection, and a valid database. Now I try to
        * build the dlobh
        */
       pLocalNewDLOBH=(dlobh *) mi_alloc (sizeof(dlobh));
       strncpy(pLocalNewDLOBH->ServerName,pServerName,SERVER_NAME_LEN);
       strncpy(pLocalNewDLOBH->ServerHost,pHostName,SERVER_HOST_LEN);
       pLocalNewDLOBH->ServerPort=nPortId;
       strncpy(pLocalNewDLOBH->DBaseUserName,IFNULL_NULL(bDataBInfo.user_name),USER_NAME_LEN);
       strncpy(pLocalNewDLOBH->DBaseName,IFNULL_NULL(bDataBInfo.database_name),DBASE_NAME_LEN);
       strncpy(pLocalNewDLOBH->DBPassword,IFNULL_NULL(bDataBInfo.password),PASSWORD_LEN);
       pLocalNewDLOBH->ObjectType=chType;

       switch(chType)
       {
           case 'I':
		
             if ( strcmp(pchTxtStrgMgr,"")==0)
	         sprintf(szQueryText,"return FileToLO('%s');",
                         pLocalFileName);
             else
	         sprintf(szQueryText,"return FileToLO('%s', '%s');",
                         pLocalFileName,
                         pchTxtStrgMgr);
 
 /*
  *             sprintf(szQueryText,"return FileToLO('%s');",
  *                    pLocalFileName);
  */

              if (mi_exec(pLocalConnection,szQueryText) != MI_ERROR)
              {
                 if((pResult=(char *) get_result(pLocalConnection))!=NULL)
                     strcpy(pLocalNewDLOBH->LOBHData,pResult);
               } else
               sprintf(szErrorText, "FileToDLOBH: Error getting result of FileToLO().");
             break;

           default:

              if (stat(pLocalFileName,&StatBuf)==0)
              {
                 pLocalNewDLOBH->mtime = StatBuf.st_mtime;
                 pLocalNewDLOBH->size  = StatBuf.st_size;
/*
 * If this file already exists, then don't do the insert. It causes an error to
 * be raised and - in the worst of all worlds - causes a BE crash.
 *
 */
                 sprintf(szQueryText, "insert into DLOBH.Externals (File_Name, Ext_LO_Name ) select '%s', '%s' from DLOBH.Externals E1 WHERE NOT EXISTS ( SELECT * FROM DLOBH.Externals E2 WHERE E2.File_Name = '%s');", pLocalFileName, pLocalFileName, pLocalFileName);

                 if (mi_exec(pLocalConnection,szQueryText) != MI_ERROR)
                 {
                    if((mi_query_finish(pLocalConnection))!=MI_ERROR)
                    strcpy(pLocalNewDLOBH->LOBHData,pLocalFileName);
                    else  sprintf(szErrorText, "FileToDLOBH: mi_query_finish error.");

                 } else
                  sprintf(szErrorText, "FileToDLOBH: Error on Insert.");
              } else
              sprintf(szErrorText, "FileToDLOBH: Error getting stat() of file >%s<.",
                                   pLocalFileName);
             break;
          }
          if (szErrorText[0]==NULL)
             return (pLocalNewDLOBH);
     } else 
     sprintf(szErrorText, "FileToDLOBH: Failed to get Database Info.  Logged in?");
   } else
   sprintf(szErrorText,"FileToDLOBH: Invalid dlobh type. Must be 'I'nternal, e'X'ternal or Large 'T'ext. (I,X,T)");

/*
 * Now I have an error condition. Raise the Error, and return the NULL
 * value (just in case).
 */
    szErr = mi_string_to_text(szErrorText);
    szErr = mi_string_to_text("This is the error.");
    mi_db_error_raise(pLocalConnection,
                      MI_EXCEPTION,
                      mi_string_to_text("This is the error"),
                      NULL);

    return (dlobh *) NULL;

}
