#include "mi.h"
#include "dlobh.h"

int dlobh_parse_url( /* dlobh *, char * */);

/*
 *  dlobh_in - Takes the textual stuff for a distributed large object handle
 *             and parses it into the dlobh type. This doesn't do any
 *             testing for the existance of the LO, as I can't assume that it
 *             even exists at _in time.
 */
dlobh *
dlobh_in ( TextPassed )
mi_text * TextPassed;
{

  char          * pStringPassed;
  MI_CONNECTION * conn;
  dlobh         * pbNewDLObHandle = NULL;
  int		  nRetVal;

  pStringPassed = mi_text_to_string(TextPassed); /* Get the text data.       */

  if ((pbNewDLObHandle=(dlobh *)mi_alloc(sizeof(dlobh)))==(dlobh *)NULL)
  {
    return ((dlobh *) NULL);
  }

  bzero(pbNewDLObHandle,sizeof(dlobh));

  if (strncmp(pStringPassed,DLOBH_URL_TAG,5)==0)
  {
    if (!(dlobh_parse_url(pbNewDLObHandle,pStringPassed)))
    {
      return ((dlobh *) NULL);
    }
  } else if ((strcmp(pStringPassed,DLOBH_INCOMPLETE_STATUS) == 0) ||
	     (strcmp(pStringPassed,DLOBH_ERROR_STATUS) == 0))
  {
     pbNewDLObHandle->ServerPort = -1;
     strcpy(pbNewDLObHandle->ServerHost,pStringPassed);
  }

  return (pbNewDLObHandle);

}

/*
 *   dlobh_parse_url - populates a dlobh block with the contents of a string.
 *
 *                     This function takes the passed string and breaks it
 *                     out into dlobh components. The string will look like;
 *
 *    "dlobh://HOSTNAME:PORT/SERVER/DATABASE/USER/PASSWORD/TYPE/LOHANDLE"
 *
 */
int
dlobh_parse_url ( pbNewDLObHandle, pStringPassed )
dlobh * pbNewDLObHandle;
char  * pStringPassed;
{
  char * pChar,
       * pOldChar,
       * pScr;
  int  nParamCount = 6,
       nScratch    = 0,
       nStrLen;
  char   szBucket[MAXPATHLEN+2]; /*Will use this for sloshing strings around */
  char errorBuf[128];            /* Buffer to hold any error messages        */
  char szMtime[24];
  char szSize[16];
  char * pFirst, 
       * pLast;

   pOldChar=pStringPassed+sizeof(DLOBH_URL_TAG)+2; /* Step over the dlobh:// */

   for(nScratch=0,pChar=pOldChar+1;((nScratch<6)&&(*pChar!=NULL));pChar++)
   {
      switch (*pChar)
      {
         case SLASH_DELIM:
         {
	    bzero(szBucket,MAXPATHLEN+2);
	    strncpy(szBucket,pOldChar,(pChar-pOldChar));
	    switch(nScratch)
	    {
	      case 0:  /*HOSTNAME:PORT*/
	      {
		pScr = index(szBucket,COLON_DELIM);
		strncpy(pbNewDLObHandle->ServerHost,szBucket,(pScr-szBucket));
		pbNewDLObHandle->ServerPort=atoi((pScr+1));
              }
	      break;   

	      case 1:  /* SERVER        */
		strncpy(pbNewDLObHandle->ServerName,szBucket,SERVER_NAME_LEN);
              break;

	      case 2:  /* DATABASE      */
	        strncpy(pbNewDLObHandle->DBaseName,szBucket,DBASE_NAME_LEN);
	      break;

	      case 3:  /* USER          */
	        strncpy(pbNewDLObHandle->DBaseUserName,szBucket,USER_NAME_LEN);
              break;

	      case 4:  /* PASSWORD      */
	        strncpy(pbNewDLObHandle->DBPassword,szBucket,PASSWORD_LEN);
	      break;

	      case 5: /* TYPE           */
                 pbNewDLObHandle->ObjectType=szBucket[0];
		 if (szBucket[0] == EXTERNAL_FILE)
		 {
		   pFirst = (char *)strchr(szBucket,COLON_DELIM);
		   pLast  = (char *)strrchr(szBucket,COLON_DELIM);
		   if ((pFirst==0)||(pLast==0)) return (0);
                   strncpy(szMtime,(pFirst+1),(pLast-pFirst-1));
		   strcpy(szSize,(pLast+1));
		   pbNewDLObHandle->mtime = atol(szMtime);
		   pbNewDLObHandle->size = atol(szSize);
		   if (pbNewDLObHandle->mtime == 0) return (0);

		 }
	      break;

	    }
         nScratch++;
         pOldChar=pChar+1;
         }
	 break;
      }
   }
   /*
    * What's left should be of the form;
    *
    * '/foo/bar/file' or 'IO87456'
    */
    bzero(szBucket,MAXPATHLEN+1);            /* Zero it out                  */
    strcpy(szBucket,pOldChar);               /* Copy the rest into szBucket  */
    strcpy(pbNewDLObHandle->LOBHData,szBucket);
  return (1);

}

/*
 * dlobh_out - This is the _out function for the type. It takes the type and
 *             renders it into a mi_text type.
 */
mi_text *
dlobh_out ( pbDLObHandle )
dlobh * pbDLObHandle;
{
 char szOutBuf[SERVER_NAME_LEN+SERVER_HOST_LEN+          /* This is the sum of*/
	       USER_NAME_LEN+DBASE_NAME_LEN+PASSWORD_LEN /* all the internal  */
	       +MAXPATHLEN+15];                          /* sizes with extra. */
 mi_text  * MiTResult;

 if (pbDLObHandle->ServerPort < 0) {
    sprintf(szOutBuf,"%s",pbDLObHandle->ServerHost);

 } else if (pbDLObHandle->ObjectType!=EXTERNAL_FILE)
 {
 sprintf(szOutBuf,"dlobh://%s:%d/%s/%s/%s/%s/%c/%s",
	 pbDLObHandle->ServerHost,
	 pbDLObHandle->ServerPort,
	 pbDLObHandle->ServerName,
	 pbDLObHandle->DBaseName,
	 pbDLObHandle->DBaseUserName,
	 pbDLObHandle->DBPassword,
	 ((pbDLObHandle->ObjectType=='\0')?' ':pbDLObHandle->ObjectType),
	 pbDLObHandle->LOBHData);

  } else {
 sprintf(szOutBuf,"dlobh://%s:%d/%s/%s/%s/%s/%c:%d:%d/%s",
	 pbDLObHandle->ServerHost,
	 pbDLObHandle->ServerPort,
	 pbDLObHandle->ServerName,
	 pbDLObHandle->DBaseName,
	 pbDLObHandle->DBaseUserName,
	 pbDLObHandle->DBPassword,
	 pbDLObHandle->ObjectType,
	 pbDLObHandle->mtime,
	 pbDLObHandle->size,
	 pbDLObHandle->LOBHData);

  }

  MiTResult = mi_string_to_text(szOutBuf);  /* This seems a bit bogus, but   */
  return (MiTResult);                       /* I'll just assume this is more */
}                                           /* than just a cast.             */

/*
 *  These functions are included to make the DLOBH type amenable to a
 *  b-tree index type.
 */
mi_boolean
dlobh_equal( bDLOBHOne, bDBLOHTwo )
dlobh * bDLOBHOne,
      * bDBLOHTwo;
{
  return ((strcmp(bDLOBHOne->ServerName,bDBLOHTwo->ServerName)==0)&&
          (strcmp(bDLOBHOne->ServerHost,bDBLOHTwo->ServerHost)==0) &&
	  (bDLOBHOne->ServerPort==bDBLOHTwo->ServerPort)&&
          (strcmp(bDLOBHOne->DBaseUserName,bDBLOHTwo->DBaseUserName)==0) &&
          (strcmp(bDLOBHOne->DBaseName,bDBLOHTwo->DBaseName)==0) &&
          (strcmp(bDLOBHOne->DBPassword,bDBLOHTwo->DBPassword)==0) &&
          (strcmp(bDLOBHOne->LOBHData,bDBLOHTwo->LOBHData)==0));
};
 
mi_boolean
dlobh_not_equal (bDLOBHOne, bDBLOHTwo )
dlobh * bDLOBHOne,
      * bDBLOHTwo;
{
    return (!(dlobh_equal(bDLOBHOne, bDBLOHTwo )));
}
mi_boolean
dlobh_less_than ( bDLOBHOne, bDBLOHTwo )
dlobh * bDLOBHOne,
      * bDBLOHTwo;
{
  return(((strcmp(bDLOBHOne->ServerHost,bDBLOHTwo->ServerHost)<0)||
          (strcmp(bDLOBHOne->ServerName,bDBLOHTwo->ServerName)<0)||
          (bDLOBHOne->ServerPort<bDBLOHTwo->ServerPort)||
          (strcmp(bDLOBHOne->LOBHData,bDBLOHTwo->LOBHData)<0)||
          (strcmp(bDLOBHOne->DBaseUserName,bDBLOHTwo->DBaseUserName)<0)||
          (strcmp(bDLOBHOne->DBaseName,bDBLOHTwo->DBaseName)<0)||
          (strcmp(bDLOBHOne->DBPassword,bDBLOHTwo->DBPassword)<0)));
}
mi_boolean
dlobh_greater_than ( bDLOBHOne, bDBLOHTwo )
dlobh * bDLOBHOne,
      * bDBLOHTwo;
{
  return (!(dlobh_less_than(bDLOBHOne, bDBLOHTwo)));
}

mi_boolean
dlobh_greater_than_or_equal_to( bDLOBHOne, bDBLOHTwo )
dlobh * bDLOBHOne,
      * bDBLOHTwo;
{
        return ((dlobh_greater_than ( bDLOBHOne, bDBLOHTwo))||
                  (dlobh_equal( bDLOBHOne, bDBLOHTwo)));
}
mi_boolean
dlobh_less_than_or_equal_to( bDLOBHOne, bDBLOHTwo )
dlobh * bDLOBHOne,
      * bDBLOHTwo;
{
        return ((dlobh_less_than ( bDLOBHOne, bDBLOHTwo))||
                  (dlobh_equal( bDLOBHOne, bDBLOHTwo)));
}

mi_text *
dlobh_get_servername ( bDLOBH )
dlobh * bDLOBH;
{
     char * pStr;

     pStr = mi_alloc(strlen(bDLOBH->ServerHost) + 1);
     strncpy(pStr,bDLOBH->ServerHost,strlen(bDLOBH->ServerHost)+1);
     return (mi_string_to_text(pStr));
}

/*
 *  dlobh_test  - This function simply tests the other two in an MSQL
 *                situation.
 */
mi_text *
dlobh_test ( TextPassed )
mi_text * TextPassed;
{
  dlobh  * pLocalDlobh;

  pLocalDlobh = dlobh_in(TextPassed);

  return (dlobh_out(pLocalDlobh));

}
/*
 * dlobhGetConnectionInfo  - Takes a dlobh structure and populates the 
 *                           MI_CONNECTION_INFO structure passed byref. 
 *                           This is used *everywhere*!
 *
 */
void
dlobhGetConnectionInfo ( pbDLOBH, pbMiConnection )
dlobh * pbDLOBH;
MI_CONNECTION_INFO * pbMiConnection;
{

   pbMiConnection->server_name=NULL_IF_NULL(pbDLOBH->ServerName);
   pbMiConnection->server_host=NULL_IF_NULL(pbDLOBH->ServerHost);
   pbMiConnection->server_port=pbDLOBH->ServerPort;

}

/*
 * dlobhGetDatabaseInfo  - Takes a dlobh structure and populates the 
 *                         MI_DATABASE_INFO structure passed byref. 
 *
 */
void
dlobhGetDatabaseInfo ( pbDLOBH, pbMiDatabase )
dlobh * pbDLOBH;
MI_DATABASE_INFO * pbMiDatabase;
{
   pbMiDatabase->user_name=NULL_IF_NULL(pbDLOBH->DBaseUserName);
   pbMiDatabase->database_name=NULL_IF_NULL(pbDLOBH->DBaseName);
   pbMiDatabase->password=NULL_IF_NULL(pbDLOBH->DBPassword);
}
/*
 *  dlobhPrint           - Prints out each of the populated
 *                         elements ( or non- populated elements )
 *                         depending on the case. Useful for
 *                         testing.
 */
void
dlobhPrint ( pbDLOBH ) 
dlobh * pbDLOBH;
{
  printf("DLOBH->{\n\
\tServerName = >%s<\n\
\tServerHost = >%s<\n\
\tServerPort = >%d<\n\
\tDBaseUserName = >%s<\n\
\tDBaseName = >%s<\n\
\tDBPassword = >%s<\n\
\tLOBHData = >%s<\n\
\tObjectType = >%c<\n\
\tmtime = >%d<\n\tsize = >%d< } ",
pbDLOBH->ServerName, pbDLOBH->ServerHost, pbDLOBH->ServerPort,
pbDLOBH->DBaseUserName, pbDLOBH->DBaseName, pbDLOBH->DBPassword,
pbDLOBH->LOBHData, pbDLOBH->ObjectType, pbDLOBH->mtime, pbDLOBH->size);

}
