#
# action.tcl - Monbrowse Action procedures
#
# Actions are user-defined Tcl scripts.  See $MIBROWSE/actions/README
# for instructions on adding your own action.
#


#
# ActionInit - setup Actions by scanning site directories
#
proc ActionInit {} {
	global MB MBAc env
	# actions already scanned
	if [info exists MBAc] {return 1}

	# scan for actions
	Tku_Working "Scanning for actions..."
	set actions $MB(Home)/actions
	if {![file isdirectory $actions] || ![file readable $actions]} return
	foreach action [glob -nocomplain $actions/*/] {
    	regsub {/$} $action "" action; # remove trailing /
	    set init $action/init.tcl
		if ![file readable $init] continue;  # no init
	    # the init file must set the Name and optionally the
	    # Help variables and must also have a procedure
	    # called "Action" which accepts a single argument, the
	    # action directory.
		set Name ""
		set Help ""
		if [catch {source $init} err] {
			Tku_Error "Error reading action \"$init\":  $err"
			continue
	    }
		set MBAc($action) [list $Name $Help]
	}
	Tku_Waiting
	if [info exists MBAc] {return 1} else {return 0}
}


#
# ActionButton - adds an action button to a Tku_MenuBar menu
#
# input:
#    w    Tku_MenuBar frame
#    type action type (default: all actions)
#
proc ActionButton {menubar} {
	global MB MBAc

	if ![ActionInit] return

	set actions ""
	foreach action [lsort [array names MBAc]] {
		set name [lindex $MBAc($action) 0]
		set help [lindex $MBAc($action) 1]
		lappend actions \
			"command -label {$name...} -command {ActionExec $action}" $help
	}
	if {$actions != ""} {
		eval Tku_MenuButton $menubar "{Actions -under 0}" $actions
	}
}

#
# ActionExec -    execute an action
#
# Input:
#	actiondir	the action directory, passed to the Action
#
proc ActionExec {action} {
	global MB

	set init $action/init.tcl
	if {![file readable $init]} {
	    Tku_Error "Can't run action in directory:  $actiondir"
	    return 0
	}
	# remove current Action
	proc Action {} {}
	# source new Action
	source $init
	if {[info body Action] == {}} {
	    Tku_Error "No Action procedure defined in $init, can't run action."
	    return 0
	}
	# run Action
	Action $action
}
