#
#  admin.tcl - procedure which do handle the Administrative stuff.
#
#  $Header: /usr/local/devel/montage/contrib/MiBrowse/lib/RCS/admin.tcl,v 1.2 1995/01/02 23:31:02 pbrown Exp pbrown $
#
#
#   First thing to do is to define all the queries this section will use.
#  This is done to minimize the code and centralize these interfaces. I've
#  employed the system catalogs wherever possible rather than the dba_funcs(),
#  which may or may not be employed. In addition, the columns returned
#  require their own names.
#
#-----------------------------------------------------------------------------------------
# Admin:SessQuery
#
#   Return the query to use when making the selection of User Sessions from
#  the database
#
#-----------------------------------------------------------------------------------------
proc Admin:SessQuery {} {
 return "\
SELECT P.proc_user as \"Process User\",
       P.proc_host as \"User's Host\",
       P.proc_pid as \"Users PID\",
       P.proc_xid as \"Users Trans Id\",
       P.proc_status as \"Process Status\",
       (SELECT UNIQUE COUNT(*) 
          FROM xactlocks 
         WHERE xlock_xid = P.proc_xid) as \"Lock Count\" 
  FROM procs P;" } 
#----------------------------------------------------------------------------------------- 
# Admin:LocksQuery  
# 
#   Return the select to use when the locks are requested. If no TransactionId, or list
#  of transaction Ids is requested, then list them all. The idea is that the user
#  selects a list of sessions (including TransactionIds from the sheet) and shows the
#  list of locks these sessions hold. (Looking for Deadlocks, ladder deadlocks etc).
#
#-----------------------------------------------------------------------------------------
proc Admin:LocksQuery { {Txs ""}} {

	set Qry "
SELECT P.proc_user as \"Process User\",
       P.proc_host as \"User's Host\",
       P.proc_pid as \"Users PID\",
       T.table_name as \"Table with Lock\",
       T.table_schema as \"Locked Table's Schema\",
       X.xlock_pid as \"Process holding Lock\",
       X.xlock_tid as \"Page being locked\",
       X.xlock_holders as \"Holding Info for the Process\"
  FROM xactlocks X, procs P, tables T
 WHERE T.oid         = X.xlock_table::oid
   AND X.xlock_pid   = P.proc_pid
   AND X.xlock_xid   = P.proc_xid"
	
	if { [ llength $Txs ] != 0 } {
		set Qry "$Qry
   AND ( "
		set first 0
		foreach i $Txs {
			if { $first == 0 } {
				set Qry "$Qry X.xlock_xid = '$i'::xid"
				set first 1
			} else {
				set Qry "$Qry OR X.xlock_xid = '$i'::xid"
			}
		}
		set Qry "$Qry )"
	}
	set Qry "$Qry;"
return $Qry
}

#-----------------------------------------------------------------------------------------
# Admin:Sessions 
#
#   Show the number of active sessions connected to this database.
#
#-----------------------------------------------------------------------------------------
proc Admin:ShowSessions {} {
	global MB
#
#  Check to see that I am the 'miadmin' user. If not, then
#  can't get to the stuff needed here.
#
	if {$MB(user) != "miadmin" } {
		Tku_Notice {Only the miadmin user has permissions on what this section reports on.}
		return 0
	}
#
# Create a new toplevel window.
#
	if [winfo exists .as] {destroy .as}
	toplevel .as
	wm title .as {DB Aministration}
	wm protocol .as WM_DELETE_WINDOW Admin:CloseSessions
	wm geom .as +150+150
	wm minsize .as 50 50
	wm withdraw .as
#
#  Fill the Menu options with interesting things to do.
#
	Tku_MenuBar .as.menu \
		{About	{This window is for displaying inforamation about current sessions for an
				Illustra database. This window automatically brings
				up the state of the system at the time you open this window. By 
				hitting 'Refresh', you get an updated view of the system. By hitting
				'Close' you Close this window and return.}}

	Tku_MenuButton .as.menu "File -under 0" \
		"command -label {Show Log} -command LogOpen -under 5" \
			{Opens the log window.} \
		"command -label Close -command Query:CloseSessions -under 0" \
			{Closes the Query Dialog and Query Results window.}
	Tku_MenuButton .as.menu "Options -under 0" \
		"checkbutton -label {Verify msql} -variable MB(-verify) -under 0" \
			{Modify and/or confirm msql before sending to the backend} \
		"command -label {Column width...} -command QRes:SetWidth -under 0" \
			{Set the width of the columns in the results window spread sheet}

	Tku_StatusBar	.as.info "Sessions Information: Server $MB(server), Database $MB(database)"
	pack .as.menu .as.info -fill x

    set QryText [ Admin:SessQuery ]

	if ![mi_check_exec MI_ROWS $QryText] {
		Tku_Error "Select failed"
		Admin:CloseSession
	}

	mi_get data cols

	if {$data(nrows) == 0} {
		Tku_Notice {ERROR! Something *really* wierd man.}
		Admin:CloseSession
	}

	loop n 0 $data(ncols) {
		set name [lindex $data(names) $n]
		lappend colinfo [list $name $MB(-width) "" ]
	}
	Tku_Sheet create .as.sheet $data(ncols) $data(nrows) $colinfo
	loop n 0 $data(ncols) {Tku_Sheet putcol .as.sheet $n $data($n)}
	Tku_Sheet ignore .as.sheet oid 

	frame .as.button
	Tku_Button	.as.fresh "Refresh" Admin:RefreshSessions "Refresh the lock list table."
	Tku_Button	.as.close "Close" Admin:CloseSessions "Close the Lock List window."
	Tku_Button	.as.locks "Locks" {Admin:ShowSelectedLocks .as.sheet} "Show Locks for selected sessions."
	pack .as.fresh -side left -in .as.button -pady 2m -padx 5m
	pack .as.locks -side left -in .as.button -pady 2m -padx 5m
	pack .as.close -side left -in .as.button -pady 2m -padx 5m
	pack .as.button -fill x -side bottom
	pack .as.sheet -fill both -expand 1

	wm deiconify .as
	Tku_Waiting

}
#-----------------------------------------------------------------------------------------
# Admin:RefreshSessions
#
#   Refresh the widget showing the sessions with new information. 
#
#-----------------------------------------------------------------------------------------

proc Admin:RefreshSessions {} {

    set QryText [ Admin:SessQuery ]

    if ![mi_check_exec MI_ROWS $QryText] {
        Tku_Error "Select failed"
        Admin:CloseSession
    }

    mi_get data cols

    if {$data(nrows) == 0} {
        Tku_Notice {ERROR! Something *really* wierd man.}
        Admin:CloseSession
    }
	Tku_Sheet clear .as.sheet
        loop n 0 $data(ncols) {Tku_Sheet putcol .as.sheet $n $data($n)}
	Tku_Sheet renum .as.sheet
}
#-----------------------------------------------------------------------------------------
#  Admin:ShowSelectedLocks 
#
#     Show the locking information for a set of transactions selected from the
#   Tku_Sheet.
#
#-----------------------------------------------------------------------------------------
proc Admin:ShowSelectedLocks { sheet } {
	list Tixs
	if {[set toedit [Tku_Sheet currentsel $sheet]] == ""} return
	foreach i $toedit {
		lappend Tixs [ Tku_Sheet get .as.sheet 3 $i ]
        }
	Admin:ShowLocks $Tixs
	return
}
#-----------------------------------------------------------------------------------------
# Admin:ShowLocks
#
#   Show the locking information requested by the user. The user may either
#  ask for all Locks, or may specify a list of transactions to show locks
#  for from the ShowSessions proc. 
#
#-----------------------------------------------------------------------------------------
proc Admin:ShowLocks { { Tixs ""}} {
	global MB
#
#  Check to see that I am the 'miadmin' user. If not, then
#  can't get to the stuff needed here.
#
	if {$MB(user) != "miadmin" } {
		Tku_Notice {Only the miadmin user has permissions on what this section reports on.}
		return 0
	}
#
# Create a new toplevel window.
#
	if [winfo exists .al] {destroy .al}
	toplevel .al
	wm title .al {DB Aministration}
	wm protocol .al WM_DELETE_WINDOW Admin:CloseLocks
	wm geom .al +150+150
	wm minsize .al 50 50
	wm withdraw .al
#
#  Fill the Menu options with interesting things to do.
#
	Tku_MenuBar .al.menu \
		{About	{This window is for displaying inforamation about transactions locking for an
				Illustra database. This window automatically brings
				up the state of the system at the time you open this window displaying
				whatever transactions you've requested. By hitting 'Refresh', you get 
				an updated view of the system. By hitting
				'Close' you Close this window and return.}}

	Tku_MenuButton .al.menu "File -under 0" \
		"command -label {Show Log} -command LogOpen -under 5" \
			{Opens the log window.} \
		"command -label Close -command Query:CloseLocks -under 0" \
			{Closes the Query Dialog and Query Results window.}
	Tku_MenuButton .al.menu "Options -under 0" \
		"checkbutton -label {Verify msql} -variable MB(-verify) -under 0" \
			{Modify and/or confirm msql before sending to the backend} \
		"command -label {Column width...} -command QRes:SetWidth -under 0" \
			{Set the width of the columns in the results window spread sheet}

	Tku_StatusBar	.al.info "Locking Information: Server $MB(server), Database $MB(database)"
	pack .al.menu .al.info -fill x

	set QryText [ Admin:LocksQuery $Tixs ]
#
# 	puts $QryText
#

	if ![mi_check_exec MI_ROWS $QryText] {
		Tku_Error "Select failed"
		return 0
	}

	mi_get data cols

	if {$data(nrows) == 0} {
		Tku_Notice {Selected Transaction(s) are holding no locks at this time.}
		return 0
	}

	loop n 0 $data(ncols) {
		set name [lindex $data(names) $n]
		lappend colinfo [list $name $MB(-width) "" ]
	}
	Tku_Sheet create .al.sheet $data(ncols) $data(nrows) $colinfo
	loop n 0 $data(ncols) {Tku_Sheet putcol .al.sheet $n $data($n)}
	Tku_Sheet ignore .al.sheet oid 

	frame .al.button
	Tku_Button	.al.fresh "Refresh" Admin:RefreshLocks "Refresh the lock list table."
	Tku_Button	.al.close "Close" Admin:CloseLocks "Close the Lock List window."
	pack .al.fresh -side left -in .al.button -pady 2m -padx 5m
	pack .al.close -side left -in .al.button -pady 2m -padx 5m
	pack .al.button -fill x -side bottom
	pack .al.sheet -fill both -expand 1

	wm deiconify .al
	Tku_Waiting

}
#-----------------------------------------------------------------------------------------
# Admin:RefreshSessions
#
#   Refresh the widget showing the sessions with new information. 
#
#-----------------------------------------------------------------------------------------

proc Admin:RefreshLocks {{ Tixs ""}} {

	set QryText [ Admin:LocksQuery ]

	if ![mi_check_exec MI_ROWS $QryText] {
		Tku_Error "Select failed"
		Admin:CloseLocks
		return 0
	}

	mi_get data cols

	if {$data(nrows) == 0} {
		Tku_Notice {Selected transactions have released all locks.}
		Admin:CloseLocks
		return 0
	}

	Tku_Sheet clear .al.sheet
        loop n 0 $data(ncols) {Tku_Sheet putcol .al.sheet $n $data($n)}
	Tku_Sheet renum .al.sheet
}
#-------------------------------------------------------------------------------------------
# Admin:CloseSessions - Close the Sessions sucker down.
#-------------------------------------------------------------------------------------------
proc Admin:CloseSessions {} {
	global MB
	update
	catch [destroy .as]
}
#-------------------------------------------------------------------------------------------
# Admin:CloseLocks - Close the Locks window down.
#-------------------------------------------------------------------------------------------
proc Admin:CloseLocks {} {
        global MB
        update
        if {[winfo exists .al]} { destroy .al }
}
