#
# browse.tcl - procedures for the top level window of MBrowse
#
# $Header: /usr/local/devel/montage/contrib/MiBrowse/lib/RCS/browse.tcl,v 1.3 1995/01/02 23:29:08 pbrown Exp pbrown $
#

# ---------------------------------------------------------------------------
# File:Server - connect to a new server
# ---------------------------------------------------------------------------
proc File:Server {} {
	set server [Tku_Entry "Server:" default "Set Server"]
	if {$server == ""} {
		return
	}
	OpenDb template1 $server
	GetDbs
}

# ---------------------------------------------------------------------------
# File:Close - close the browser
# ---------------------------------------------------------------------------
proc File:Close {} {Query:CloseAll; destroy .}

# ---------------------------------------------------------------------------
# GetDbs - get a list of databases on the host, using the template1
#		database as a start, and update the .dbs.listbox
# ---------------------------------------------------------------------------
proc GetDbs {} {
	global MB

	Tku_ListboxClear .dbs.listbox
	Tku_ListboxClear .schemas.listbox
	Table reset
	# retrieve and save the list of databases using template1
	if ![mi_check_exec MI_ROWS \
		"select database_name from databases;" "getting databases..."] {
		Tku_Error {Could not get database list (are the DBA functions installed?)}
		return 0
	}
	mi_get data cols
	Tku_ListboxFill .dbs.listbox [lsort $data(0)]
	Tku_DefFocus .dbs.listbox
	return 1
}

# ---------------------------------------------------------------------------
# DebugReSource - re-source a file in the $MB(Home)/lib directory
# ---------------------------------------------------------------------------
proc DebugReSource { file } { 
	global MB
	set f $MB(Home)/lib/$file
	if [file exists $f] {
			puts stderr "re-reading $f..."
			uplevel #0 "source $f"
	}
}

# ---------------------------------------------------------------------------
# Table - open, close, and maintain information about tables in the db
# ---------------------------------------------------------------------------
proc Table {cmd {arg ""}} {
	global MB MBA MBTK

	if ![info exists MBA] {set MBA(num) 0}

	case $cmd in {
		num {return $MBA(num)}
		tables {Tku_ListboxDump .tables.listbox}
		numopened {return [llength [Table opened]]}
		opened {
			if ![info exists MBA(opened)] {return ""} \
			else {return $MBA(opened)}
		}
		open {
			if {$arg == ""} {error "Table:  must provide a table to open."}
			if [info exists MBA(n,$arg)] {
				wm iconify .table$MBA(n,$arg)
				wm deiconify .table$MBA(n,$arg)
				update
				focus .table$MBA(n,$arg)
				return
			}
			if {[lsearch [Table tables] $arg] == -1} {
				Tku_Error "No such table:  $arg"
				return 0
			}
			if {$MBTK($arg) == "i"} {
				ShowIndex $arg
				return
			}
			lappend MBA(opened) $arg
			# get the next available window name
			set n 0
			while {[winfo exists .table$n]} {incr n}
			set MBA(n,$arg) $n
			set f .table$n
			toplevel $f
			wm title $f "Table: $arg"
			wm iconname $f $arg
#			wm group . $f
			wm minsize $f 1 1

			Tku_Listbox $f.box {TypeOpen [selection get]} -geom 30x10
			$f.box.title conf -text "Column(s) in $arg:"
			Tku_Listbox $f.indices {ShowIndex [selection get]} \
				-geom 30x10
			$f.indices.title conf -text "Indices on $arg:"
			frame $f.butts 
			Tku_Button $f.select "Select *" "Query:Table $arg" \
				"Show all rows in $arg" 1
			Tku_Button $f.query Query "Query $arg" \
				"Open the query tool on $arg"
			Tku_Button $f.close Close "destroy .table$n" \
				"Close and return"
			bind $f <Destroy> "Table close $arg"
			pack $f.select $f.query $f.close \
				-pady 2m -padx 5m -side left -in $f.butts
#			frame $f.rules
#			label $f.rules.l -text "Rules on $arg:"
#			scrollbar $f.rs -command "$f.r yview"
#			text $f.r -yscroll "$f.rs set" -relief sunken -bd 2
#			pack $f.rules.l -fill x
#			pack $f.rs -in $f.rules -side right -fill y
#			pack $f.r -in $f.rules -side right -fill both \
#				-expand 1
#			$f.r insert 0.0 [join [GetRules $arg] \n]
			Tku_Listbox $f.rules {} -geom 30x10
			$f.rules.title conf -text "Rules on $arg:"
			pack $f.butts -side bottom -fill x
			pack $f.rules -side bottom -fill both -expand 1
			pack $f.box $f.indices -side left -fill x -expand 1
			foreach e [Column names $arg] {lappend defs [Column def $arg $e]}
			Tku_ListboxFill $f.box.listbox $defs
			Tku_ListboxFill $f.indices.listbox \
				[mi_select_rows $MB(connection) \
					"select index_indextable from indices 
					 where index_heaptable = '$arg';"]
			set lines ""
			foreach rule [GetRules $arg] {
				foreach line [split $rule \n] {
					lappend lines $line
				}
			}
			Tku_ListboxFill $f.rules.listbox $lines
			bind $f.box.listbox <Control-w> "$f.close.b invoke"
			bind $f.box.listbox <Control-q> "$f.query.b invoke"
			Tku_SetTabbing $f.box.listbox $f.query.b $f.close.b
			update
			focus $f.box.listbox
			return
		}
		close {
			 set MBA(opened) [ldelete $MBA(opened) $arg]
			 unset MBA(n,$arg)
		}
		reset {
			if [info exists MBA] {
				# close any query results
				Query:CloseAll
				# delete all possible open table windows
				set n 0
					foreach c [Table tables] {
					if [winfo exists .table$n] {destroy .table$n}
					incr n
					}
				# clear the table and attribute information
				unset MBA
				if [winfo exists .tables.listbox] {
					Tku_ListboxClear .tables.listbox
					.tables.title conf -text $arg
				}
				Column reset
				Type reset
			}
		}
	}
}

# ---------------------------------------------------------------------------
# OpenDb - open a database
# ---------------------------------------------------------------------------
proc OpenDb {db {server ""}} {
	global MB

	if ![info exists MB(database)] {
		set MB(database) ""
	}

	if ![info exists MB(server)] {
		set MB(server) ""
	}

	if {$server == ""} {
		set server $MB(server)
	}

	if {$db == $MB(database) && $server == $MB(server)} {
		return
	}

	set MB(database) $db
	set MB(server) $server

	Tku_Working "Connecting to $db on $server..."

	if [info exists MB(connection)] {
		mi_close $MB(connection)
	}
	set MB(connection) [mi_server_connect $MB(server)]
	mi_login $MB(connection) $MB(database) miadmin
	Tku_Waiting
}

		
# ---------------------------------------------------------------------------
# GetSchemas - query the db for a list of tables
# ---------------------------------------------------------------------------
proc GetSchemas {{db ""}} {
	global MB

	if {$db != ""} {
		OpenDb $db
	}

	.schemas.listbox delete 0 end
	.schemas.title conf -text $db
	Table reset
	set q "select schema_name from schemas;"
	if ![mi_check_exec MI_ROWS $q "getting schemas..."] {
		Tku_Error "Could not get schema list"
		return ""
	}
	mi_get data cols
	set schemas $data(0)
	lappend schemas public
	Tku_ListboxFill .schemas.listbox [lsort $schemas]
	return $schemas
}

# ---------------------------------------------------------------------------
# GetTables - query the db for a list of tables
# ---------------------------------------------------------------------------
proc GetTables {{schema ""}} {
	global MB MBA MBTK
	if { $schema == "" } {
		set schema $MB(schema)
	} else {
		set MB(schema) $schema
	}

	Table reset $schema
	mi_check_exec MI_DML "set schema $schema;"
	set q "select table_name, table_kind from tables
	where table_schema = '$schema'"
	if !$MB(show.tables) {append q "\nand table_kind <> 't'"}
	if !$MB(show.views) {
		append q "\nand table_kind <> 'v'"
		append q "\nand table_kind <> 'V'"
	}
	if !$MB(show.indices) {append q "\nand table_kind <> 'i'"}
	if !$MB(show.special) {append q "\nand table_kind <> 's'"}
	switch -exact -- $MB(include) {
		user	{append q "\nand table_issystem = 'f';"}
		system	{append q "\nand table_issystem = 't';"}
		all	{append q ";"}
	}
	if ![mi_check_exec MI_ROWS $q "getting tables..."] {
			Tku_Error "Could not get tables list"
			return 0
	}
	mi_get data cols
	Tku_ListboxClear .tables.listbox
	Tku_ListboxFill .tables.listbox [lsort $data(0)]
	catch {unset MBTK}
	set n 0
	foreach table $data(0) {
		set MBTK($table) [lindex $data(1) $n]
		incr n
	}
	return 1
}

proc TypeOpen {sel} {
	set type [string trimright [lindex [split [lindex $sel 0] (] 1] )]
#
#  If this is a ref, then we need the actual type being ref'ed
#
	if { $type == "ref" } {
         set type [string trimright [lindex [split [lindex [split \
                   [lindex $sel 0] )] 0] (] 2] ]
        }
	if {[Type num $type] == 0} {Tku_Notice "$type is a base type"; return}
	foreach e [Type names $type] {lappend defs [Type def $type $e]}
	Tku_Notice [join $defs \n]
}

# ---------------------------------------------------------------------------
# Type - information about the types in the db
# ---------------------------------------------------------------------------
proc Type {cmd {type ""} {arg ""} {arg2 ""}} {

	global MBT MB

	# get the attributes of a table if we haven't already done so
	if {$type != ""} {
		if ![info exists MBT($type,num)] {
			if ![Type:Query $type] return 
		}
	}

	case $cmd in {
		reset	{catch {unset MBT}}
		num		{return $MBT($type,num)}
		names	{return $MBT($type,names)}
		name	{return $MBT($type,name,$arg)}
		def		{return $MBT($type,def,$arg)}
		type	{return $MBT($type,type,$arg)}
		default {error "Type error: no such command $cmd"}
	}
}

# ---------------------------------------------------------------------------
# Type:Query - get the columns of a type (if any)
# ---------------------------------------------------------------------------
proc Type:Query {type} {
	global MB MBT

	set cmd \
"select column_name, column_type, column_number
	from	types, columns
	where	column_member = type_name
	and	type_name = '$type'
	order	by column_number;"
	if ![mi_check_exec MI_ROWS $cmd "opening $type..."] {
		Tku_Error "Could not get columns for type:  $type"
		return 0
	}

	mi_get data cols
	lappend MBT(types) $type
	set MBT($type,num) $data(nrows)

	set num 0
	loop n 0 $data(nrows) {
		set name [lindex $data(0) $n]
		set ctype [lindex $data(1) $n]
		incr num 
		set MBT($type,name,$num) $name
	   	set MBT($type,type,$name) $ctype
		lappend MBT($type,names) $name
		set MBT($type,def,$name) "$name  ($ctype)"
	}
	return 1
}

# ---------------------------------------------------------------------------
# Column - information about the columns of tables in the db
# ---------------------------------------------------------------------------
proc Column {cmd {table ""} {arg ""} {arg2 ""}} {
	global MBC MB

	# get the attributes of a table if we haven't already done so
	if {$table != ""} {
		if ![info exists MBC($table,num)] {
			if ![Column:Query $table] return 
		}
	}

	case $cmd in {
		reset	{catch {unset MBC}}
		num		{return $MBC($table,num)}
		names	{return $MBC($table,names)}
		name	{return $MBC($table,name,$arg)}
		def		{return $MBC($table,def,$arg)}
		type	{return $MBC($table,type,$arg)}
		allnum	{return $MBC($table,allnum)}
		allnames {return $MBC($table,allnames)}
		default {error "Column error: no such command $cmd"}
	}
}

# ---------------------------------------------------------------------------
# Column:Query - get the possible columns for a table
# ---------------------------------------------------------------------------
proc Column:Query {table} {
	global MB MBC
#
# Illustra doesn't store the public schema as a schema. Rather, it
# only stores schemas within the public schema. Hence this hack.
#
	if { $MB(schema) == "public" } {
		set cmd "
select C.column_name,
       C.column_type,
       C.column_number
   from columns C, tables T
 where C.column_member = T.table_type
   and T.table_name = '$table'
 order by column_number;"
	} else {
        	set cmd "
select C.column_name,
       C.column_type,
       C.column_number
   from columns C, tables T, schemas S
 where C.column_member = T.table_type
   and T.table_name = '$table'
   and S.schema_name = '$MB(schema)'
   and T.table_schema = S.oid::schemaoid
 order by column_number;"
	}

	if ![mi_check_exec MI_ROWS $cmd "opening $table..."] {
		Tku_Error "Could not get columns for table:  $table"
		return 0
	}

	mi_get data cols
	lappend MBC(tables) $table
	set MBC($table,num) $data(nrows)
        set AllRows 0

	loop n 0 $data(nrows) {
		incr AllRows
		set name [lindex $data(0) $n]
		set type [lindex $data(1) $n]
		set MBC($table,name,$n) $name
	   	set MBC($table,type,$name) $type
		lappend MBC($table,names) $name
		set MBC($table,def,$name) "$name  ($type)"
#
# I'm adding a couple of new ones;
#
#    MBC($table, allnum)
#
#    MBC($table, allname, x)  - Name of the all, which will be deref(Name).Attr
#
		set MBC($table,allname,$AllRows) $name
		set MBC($table,alltype,$name) $type
		lappend MBC($table,allnames) $name
                set MBC($table,alldef,$name) "$name ($type)"
#
#  If this is a ref() type, then I need to add the attributes which can be 'ref''d to
#
		if {[string trimright [lindex [split [lindex $type 0] (] 0] ] \
                    == "ref" } {
			set Ptype [lindex [split [lindex [split $type )] 0] (] 1]
			foreach e [Type names $Ptype] {
				incr AllRows
				set PName "deref($name).$e"
				set MBC($table,allname,$AllRows) $PName
				lappend MBC($table,allnames) $PName
                	}
                }
	}
	set MBC($table,allnum) $AllRows
	return 1
}

proc File:Open {} {
	global MB

	if {[.dbs.listbox cur] != ""} {
		GetTables [selection get]
	} elseif {[.tables.listbox cur] != ""} {
		puts stdout [selection get]
		foreach table [selection get] {Table open $table}
	}
}

proc Debug:SendTK {} {
	if ![Tku_Edit "Tk To Execute" "" cmd] return
	send [winfo name .] $cmd
}

proc File:SendSQL {} {
    global MB

    if ![Tku_Edit "msql:" "" cmd] return
    if {[set cmd [string trim $cmd]] == ""} return
	set ret [mi_exec_log $cmd]
	case $ret in {
		MI_ERROR {Tku_Error "msql failed"}
		MI_ROWS	{
			if [Tku_Confirm "View Results?"] {
				QRes $cmd 0
#				set tmp [Tku_Tmp]
#				mi_dump $tmp
#				Tku_ViewFile $tmp
			}
		}
    }
}

proc File:SendSQLDisp {cmd} {
    global MB
    if {[set cmd [string trim $cmd]] == ""} return
        set ret [mi_exec_log $cmd]
        case $ret in {
                MI_ERROR {Tku_Error "msql failed"}
                MI_ROWS {
                        if [Tku_Confirm "View Results?"] {
                                QRes $cmd 0
#                               set tmp [Tku_Tmp]
#                               mi_dump $tmp
#                               Tku_ViewFile $tmp
                        }
                }
    }
}



proc File:Drop {} {
	set tables [Tku_ListboxGet .tables.listbox]
	if {$tables == ""} return
	if ![Tku_Confirm "Drop:  $tables?"] return
	foreach t $tables {
		if ![mi_check_exec MI_DML \
			"drop table $t;" "Dropping $t..."] return
	}
	GetTables
}


proc collapse s {
	regsub -all { +} [string trim [translit \t\n "  " $s]] { } s
	regsub -all {\( } $s {(} s
	regsub -all { \)} $s {)} s
	return $s
}

proc GetRules {table} {
	global MB
	set c $MB(connection)
	set ret ""

	# collect all the rules by event type
	foreach rule [mi_select_rows $c  "select rule_ev_type,rule_when,rule_name,rule_qual,rule_action from rules where rule_table = '$table';"] {
		switch [lindex $rule 0] {
			3	{lappend ev(select) $rule}
			4	{lappend ev(insert) $rule}
			5	{lappend ev(delete) $rule}
			6	{lappend ev(update) $rule}
		}
	}

	# for each event type, collect rules by when and
	# then dump them out
	foreach e {select insert update delete} {
		if ![info exists ev($e)] continue
		foreach rule $ev($e) {
			switch [lindex $rule 1] {
				1	{lappend when(before) $rule}
				2	{lappend when(after) $rule}
				3	{lappend when(while) $rule}
				4	{lappend when(instead) $rule}
			}
		}
		foreach rule $ev($e) {
			foreach w {before while after instead} {
				if ![info exists when($w)] continue
				foreach r $when($w) {
					set action [collapse [lindex $r 4]]
					set qual [collapse [lindex $r 3]]
#				keylset k event $e  when $w  name [lindex $r 2]   qual $qual action $action
#		lappend ret $k
#					unset k
set do Do
if {$w != "while"} {
	append do " $w"
}
if {$qual != ""} {
	set qual "\n       $qual"
}
lappend ret "On $e:$qual\n$do:\n        $action\n"
				}
				unset when($w)
			}
		}
		unset ev($e)
	}
	return $ret
}


proc ShowIndex {index} {
	set q "select index_heaptable,index_key from indices \
		where index_indextable = '$index';"
	if ![mi_check_exec MI_ROWS $q "Getting index $index..."] {
		Tku_Error "Could not check attributes of index:  $index"
		return 0
	}
	mi_get data cols
	Tku_Notice "$index is an index on the $data(0) table on $data(1)"
	return 1
}
