#
# init.tcl - set variables, parse arguments, and initialize the . window
#

#set tk_strictMotif 1
set MB(Home) $env(MIBROWSE)
set MB(Bin) $MB(Home)/bin
lappend auto_path $MB(Home)/lib

# set defaults and parse the comand line args
set MB(show.tables) 1
set MB(show.views) 1
set MB(show.special) 0
set MB(show.indices) 0
set MB(show.refs) 0
set MB(include) user
set MB(-debug) 0
set MB(-verify) 0
set MB(-log) 0
set MB(-width) 20
set MB(schema) public
if [info exists env(MI_SERVER)] {
	set server $env(MI_SERVER)
} else {
	set server default
}
if [info exists env(MI_DATABASE)] {
	set db $env(MI_DATABASE)
} else {
	set db $env(USER)
}
if [info exists env(USER) ] {
	set MB(user) $env(USER)
} else {
	set MB(user) nobody
}
set table ""
option add *fixedfont -adobe-courier-bold-r-normal-*-*-120-*

for {set i 0} {$i < [llength $argv]} {incr i} {
	switch -glob -- [lindex $argv $i] {
		-debug {set MB(-debug) 1}
		-l* {set MB(-log) 1}
		-v* {set MB(-verify) 1}
		-big {
			option add *Font -adobe-helvetica-bold-r-normal-*-*-180-*
			option add *Entry.font -adobe-courier-medium-r-normal-*-*-180-*
			option add *Text.font -adobe-courier-medium-r-normal-*-*-180-*
			option add *fixedfont -adobe-courier-bold-r-normal-*-*-180-*
			option add *Pad 10
		}
		-s* {set server [lindex $argv [incr i]]}
		-d* {set db [lindex $argv [incr i]]}
		-t* {set table [lindex $argv [incr i]]}
		-w* {set MB(-width) [lindex $argv [incr i]]}
		-* {
			puts stderr {
Usage:
	mibrowse [-big] [-db database] [-table table]
Where:
	-big			use big font
	-server server  Illustra server
	-db database	inital database
	-table table	inital database
}
			exit 1
		}
	}
}

# Initialize the log window
	toplevel .log
	wm withdraw .log
	wm title .log "Illustra Query Log"
	wm minsize .log 100 100 
	wm geom .log +1-1
	frame .log.frame
	text .log.text -yscroll {.log.vs set} -state disabled -wrap word
	scrollbar .log.vs -command {.log.text yview}
	Tku_MenuBar .log.menu \
		{About {This window displays the commands sent to the
		Illustra server and the messages received.  You can
		save the log at anytime and can add notes by clicking
		at the desired location and typing}}
	
	Tku_MenuButton .log.menu "File -under 0" \
		{command -label "Save Log" -command LogSave -under 0} \
			{Save the log to a file} \
		{command -label "Clear Log" -command LogClear -under 1} \
			{Clears the log window} \
		{command -label Close -command LogClose -under 0} \
			{Closes the log window}
	
	Tku_StatusBar .log.info "Illustra Query Log"
	.log.info conf 
	pack .log.menu .log.info -fill x
	pack .log.text -side left -fill both -expand 1 -in .log.frame
	pack .log.vs -side left -fill y -in .log.frame
	pack .log.frame -fill both -expand 1
	bind .log.text <Button-1> {focus %W}
	bind .log.text <FocusIn> {%W conf -relief sunken}
	bind .log.text <FocusOut> {%W conf -relief raised}
	bind .log.text <Map> {focus .log.text}
	bind .log.text <Unmap> {focus [focus default]}
	bind .log <FocusIn> {focus .log.text}
	bind .log <FocusOut> {focus [focus default]}
	wm protocol .log WM_DELETE_WINDOW LogClose


# Hide or initialized the root window

   # set window manager stuff
	wm command . [concat pgbrowse $argv]
	wm client . [exec hostname]
	wm title . "Illustra Browser"
	wm protocol . WM_DELETE_WINDOW File:Close
	wm geom . +75+75
	wm minsize . 30 10
	wm maxsize . 30 50

	Tku_StatusBar .info "Illustra Browser"

	Tku_MenuBar .menu {About {Illustra Browser v0.2a Complaints, bugs and suggestions for improvement, e-mail pbrown@postgres.Berkeley.EDU}}
	
	Tku_MenuButton .menu "File -under 0" \
		"command -label {Open} -command File:Open -under 0" \
			{Opens the selected database or table} \
		"command -label {Server...} -command File:Server -under 0" \
			{Connects to a different server}  \
		"command -label {Open Log} -command LogOpen -under 5" \
			{Opens the msql query log} \
		"command -label {Send msql...} -command {Tku_MISQL MISQL {} 0 }" \
			{Sends an msql query to the Illustra server} \
		"command -label {Drop Table(s)} -command File:Drop -under 0" \
			{Drop selected tables (do you know what you're doing???} \
		separator {} \
		"command -label Close -command File:Close -under 0"

	Tku_MenuButton .menu "Admin -under 0" \
		"command -label {Show Active Sessions} -command Admin:ShowSessions -under 0" \
			{Show users with open sessions to the Database} \
		"command -label {Show Locks} -command Admin:ShowLocks -under 0" \
			{Show users locking} \
		"command -label {Show Statistics} -command Admin:ShowStats -under 0" \
			{Show statistics for selected tables}
	
	Tku_MenuButton .menu "Options -under 0" \
		"checkbutton -label {Show Regular Tables} -variable MB(show.tables) \
			-comm GetTables -under 5" \
			{Include regular tables in list of tables} \
		"checkbutton -label {Show View Tables} -variable MB(show.views) \
			-comm GetTables -under 5" \
			{Include views in list of tables} \
		"checkbutton -label {Show Index Tables} -variable MB(show.indices) \
			-comm GetTables -under 5" \
			{Include indices in list of tables} \
		"checkbutton -label {Show Special Tables} -variable MB(show.special) \
			-comm GetTables -under 5" \
			{Include special tables in list of tables} \
	"checkbutton -label {Show ref() attributes} -variable MB(show.refs) \
			-comm GetTables -under 5" \
		{For each ref() attribute, show all deref(attribute).attributes} \
		separator {} \
		"radiobutton -label {Include User Tables} -variable MB(include) \
			-comm GetTables -under 8 -value user" \
			{Include just user tables in list of tables} \
		"radiobutton -label {Include System Tables} -variable MB(include) \
			-comm GetTables -under 9 -value system" \
			{Include just system tables in list of tables} \
		"radiobutton -label {Include All Tables} -variable MB(include) \
			-comm GetTables -under 8 -value all" \
			{Include all tables in list of tables} \
		separator {} \
		"checkbutton -label {Verify msql} -variable MB(-verify) -under 7" \
			{Modify and/or confirm msql before sending to the server}
	
	if $MB(-debug) {
		foreach file [glob $MB(Home)/lib/*.tcl] {
			set file [file tail $file] 
			lappend debugs \
				"command -label $file -command {DebugReSource $file}" ""
		}
		lappend debugs \
			"command -label {Send Tk...} -command Debug:SendTK" \
			"Execute Tk code in the mibrowse interpreter"

		eval Tku_MenuButton .menu "{Debug -under 0}" $debugs
	}
	
	pack .menu .info -fill x
	
	Tku_Listbox .dbs {GetSchemas [selection get]} -geom 30x10
	.dbs.title conf -text databases
	Tku_Listbox .schemas {GetTables [selection get]} -geom 30x10
	Tku_Listbox .tables {Table open [selection get]} -geom 30x10
	pack forget .dbs.hs .schemas.hs .tables.hs
	pack .dbs .schemas .tables -fill y -side left
	pack .tables -expand 1 -fill both -side left
	Tku_SetTabbing .dbs.listbox .tables.listbox
	tk_listboxSingleSelect .dbs.listbox .schemas.listbox
	Tku_DefFocus .dbs.listbox
	bind .tables.listbox <Double-Button-2> {
	Query:Table [%W get [%W nearest %y]]
	}


#
# get list of databases, list of schemas, and tables in public schema
#
if $MB(-log) LogOpen
mi_add_callback Exception LogCallbackMsg
mi_add_callback Client_Library_Error LogCallbackMsg
OpenDb $db $server
unset server db
GetDbs
GetSchemas
GetTables
