proc mi_column {what rdesc {flat 1}} {
	case $what in {
		t*	{set f mi_column_type_name}
		n*	{set f mi_column_name}
		i*	{set f mi_column_is_composite}
		default {error "unknown option:  $what}
	}
	set r ""
	loop n 0 [mi_column_count $rdesc] {
		if {[mi_column_is_composite $rdesc $n] && $flat} {
			foreach e [mi_column $what \
					   [mi_composite_column_desc $rdesc $n] $flat] {
				lappend r $e
			}
		} else {
			lappend r [$f $rdesc $n]
		}
	}
	return $r
}
proc mi_get {array {type rows} {flat 1} {start 20} {update 5}} {
	global MB

	set MB(get) 1
	set MB(update) $update
	set MB(start) $start
	upvar $array a
	catch {unset a}
	set a(nrows) 0
	set rdesc [mi_get_row_desc_without_row $MB(connection)]
	set a(names) [mi_column names $rdesc $flat]
	set a(types) [mi_column types $rdesc $flat]
	if !$flat {set a(is_composite) [mi_column is_composite $rdesc $flat]}
	set a(ncols) [llength $a(names)]
	if [string match r* $type] {
		while {[set row [mi_next_row $MB(connection)]] != ""} {
			if ![mi_check_progress $a(nrows)] return
			set a($a(nrows)) [mi_get_row_values $row $flat]
			incr a(nrows)
		}
	} else {
		loop n 0 $a(ncols) {set a($n) ""}
		while {[set row [mi_next_row $MB(connection)]] != ""} {
			set n 0
			foreach val [mi_get_row_values $row $flat] {
				lappend a($n) $val; incr n
			}
			if ![mi_check_progress [incr a(nrows)]] return
		}
	}
	mi_stop_progress
}
proc mi_stop_progress {} {
	if [winfo exists .get] {destroy .get}
	update
}
proc mi_start_progress {} {
	global MB
	set MB(row) $MB(start)
	set MB(next) [expr $MB(start) + $MB(update)]
	catch {destroy .get}
	toplevel .get
	wm title .get "Fetch"
	frame .get.bot
	Tku_Button .get.can Cancel {set MB(get) 0}
	pack .get.can -in .get.bot
	frame .get.top
	label .get.bit -bitmap hourglass
	label .get.label -text "Row:  "
	label .get.row -textvar MB(row) -anchor w -width 5
	pack .get.bit .get.label .get.row -side left -padx 5m -pady 5m \
		-in .get.top
	pack .get.top .get.bot -fill x
	update
}
proc mi_check_progress {row} {
	global MB
	update
	if !$MB(get) {	
		mi_query_interrupt $MB(connection) 1
		destroy .get
		return 0
	}
	if {$row < $MB(start)} {return 1}
	if {$row == $MB(start)} {mi_start_progress; return 1}
	if {$row > $MB(next)} {
		set MB(row) $MB(next)
		incr MB(next) $MB(update)
		update
	}
	return 1
}
	
proc mi_dump {file {flat 1}} {
	write_file $file [mi_row_values $MB(connection)]
}
proc mi_exec_log {query {msg querying...}} {
	global MB

	if $MB(-verify) {if ![Tku_Edit "Verify msql" $query query] {return 0}}
	LogSend $query
	Tku_Working $msg
	set result [mi_exec $MB(connection) $query]
	LogRec $result
	Tku_Waiting
	return $result
}
proc mi_check_exec {expect query {msg "selecting rows..."}} {
	if {[mi_exec_log $query $msg] != $expect} {return 0} \
	else {return 1}
}
