#
# query.tcl - procedures to support the query tool and query results window
#

# ---------------------------------------------------------------------------
# Query - Simple query system
# ---------------------------------------------------------------------------
proc Query {table} {
	global MB MBOpers

	set MB(queryTable) $table
	if $MB(show.refs) { set names allnames } else { set names names }


	if [winfo exists .q] {destroy .q}
	toplevel .q
	wm title .q Query
	wm protocol .q WM_DELETE_WINDOW Query:Close
	wm geom .q +150+150
	wm minsize .q 50 50
	wm withdraw .q

	# pack together menu and info bars
	Tku_MenuBar .q.menu \
	{About {This window allows you to query the table you're
	interested in.  First, set "on" the checkbutton's of the columns
	you're interested in selecting.  Then, either select "Select
	All" to return all rows or enter qualifications for one or more
	columns and select "Select."  You can insert a new row by filling
	in the columns and selecting "Insert".}}

	Tku_MenuButton .q.menu "File -under 0" \
	  "command -label {Show Log} -command LogOpen -under 5" \
	  {Opens the log window.} \
	  "command -label Close -command Query:Close -under 0" \
	  {Closes the Query Dialog and Query Results window.}
	Tku_MenuButton .q.menu "Options -under 0" \
		"checkbutton -label {Verify msql} -variable MB(-verify) -under 0" \
		{Modify and/or confirm msql before sending to the backend} \
		"command -label {Column width...} -command QRes:SetWidth -under 0" \
		{Set the width of the columns in the results window spread sheet}

	Tku_StatusBar .q.info "Table Query:  $table"
	pack .q.menu .q.info -fill x

	# get the width of the longest column name
	set maxline 0
	foreach name [Column $names $table] {
		if {[set n [string length $name]] > $maxline} {set maxline $n}
	}
	incr maxline 20

	# create the toplevel buttons for select or insert
	set MB(do) Select
	frame .q.do -bd 0
	radiobutton .q.do.sel -text Select... -var MB(do) -val Select \
		-relief raised -command Query:SetTo
	radiobutton .q.do.ins -text Insert... -var MB(do) -val Insert \
		-relief raised -command Query:SetTo
	pack .q.do.sel .q.do.ins -side left -ipadx 1c -ipady 5m -fill x
	pack .q.do -fill x

	# make a frame for the canvas and it's scrollbars
	frame .q.cframe -bd 0
	scrollbar .q.xs -command {.q.c xview} -orient horiz -relief flat
	scrollbar .q.ys -command {.q.c yview} -relief flat
	canvas .q.c -yscroll {.q.ys set} -bd 0 -xscroll {.q.xs set}

	# create the frame on the canvas
	frame .q.c.f -bd 0
	.q.c create window 0 0 -window .q.c.f -anchor nw

	# create the insert frame (hidden for now)
	set frame [frame .q.c.f.ins -bd 0]
	label $frame.vals -text "Values:" -anchor w
	label $frame.default -text Default: -anchor e
	pack $frame.vals -side left
	pack $frame.default -side right

	# create and pack the select frame
	set frame [frame .q.c.f.sel -relief flat]
	label $frame.select -text Select: -anchor w
	pack $frame.select -fill x
	set MB(queryDistinct) 0
	checkbutton $frame.distinct -text Distinct -var MB(queryDistinct) \
		-relief flat -anchor w
	set MB(query,sel) 1
	radiobutton $frame.all -text "All (*)" -var MB(query,sel) -val 1 \
		-anchor w -relief flat -command Query:AllCols
	radiobutton $frame.count -text "Count(*)" -var MB(query,sel) -val 2 \
		-anchor w -relief flat -command Query:Count
	radiobutton $frame.some -text "Some:" -var MB(query,sel) -val 0 \
		-anchor w -relief flat -command Query:NoCols
	pack $frame.distinct $frame.count $frame.all $frame.some -anchor w
	set checks [frame $frame.radios -relief flat]
	set ncol 4
	if $MB(show.refs) { set numbers allnum } else { set numbers num }
	set nrow [expr [Column $numbers $table] / ($ncol - 1)]
	set i 0
	set row 0
	foreach name [Column $names $table] {
		if {$row == 0} {set rowframe [frame $checks.$i -relief flat]}
		set MB(sel,$name) 1
		checkbutton $rowframe.$i -text $name -var MB(sel,$name) \
			-anchor w -relief flat 
		pack $rowframe.$i -anchor w 
		incr i
		incr row
		if {$row == $nrow} {
			pack $rowframe -padx 5m -side left
			set row 0
		}
	}
	if {$row != 0} {
			pack $rowframe -padx 5m -side left -anchor n
	}
	pack $checks -fill x

	# add the where button
	set MB(query,where) 1
	checkbutton $frame.where -text Where... -anchor w -var MB(query,where) \
		-command Query:Where -relief flat
	pack $frame.where -anchor w
	pack $frame -fill x

	# build and pack each label/entry pair
        #
        #  Hmm. This should be built up from the system catalogs .  .  .
        #  and some means of getting to the Functions should be 
        #  incuded too . . .
        #
        set opers [ Query:GetAllOpers ]
        lappend opers like "not like" 
# set opers [list = <> > >= < <= like "not like"]
	set i 0
	Tku_Working "opening query form..."
	foreach name [Column $names $table] {
		set MB(query,name,$i) $name

		set frame [frame .q.c.f.$i -bd 0]
		# set type [Column type $table $name]

		# add the label
		label $frame.r -text $name -width $maxline -anchor e

		# add an entry for this column
		Tku_EntryBox $frame.entry "" "" -width 40

		# add the default save button
		checkbutton  $frame.save -relief flat -var MB(save,$i)

		# add an operator pulldown or = label
		label $frame.label -text = -padx 2m
		Tku_PullDown $frame.opts {} $opers = -width 6

		# pack what's to be visible now
		pack $frame.r $frame.entry -side left 

		# add to tablist and set <Return> binding
	  	lappend tablist $frame.entry.e
		if {$i > 0} {
	    	bind .q.c.f.[expr $i-1].entry.e <Return> "focus $frame.entry.e"
		}
		
		# pack this column's frame
		pack $frame -fill x -anchor w
		incr i
	}

	# add order by frame
	set order [frame .q.c.f.order -bd 0]
	checkbutton $order.c -relief flat -text "Order By ..." -var MB(queryOrder)
	menubutton $order.mb -menu $order.mb.m -relief raised \
		-bitmap [Tku_PullDown:Bitmap]
	menu $order.mb.m
	foreach name [Column $names $table] {
		$order.mb.m add command -label $name -command "Query:OrderBy $name"
	}
	pack $order.c $order.mb -side left -pady 2m
	pack $order -fill x

	# pack the canvas together
	pack .q.xs -in .q.cframe -side bottom -fill x
	pack .q.ys -in .q.cframe -fill y -side right
	pack .q.c -in .q.cframe -fill both -expand 1

	# add the button buttons
	frame .q.button 
	Tku_Button .q.clear "Clear" Query:Clear "Clear all entries"
	Tku_Button .q.close "Close" Query:Close "Close and return."
	Tku_Button .q.doit "" Query:Do "" 1
	.q.doit.b conf -textvar MB(do)
	pack .q.clear -side left -in .q.button -pady 2m -padx 5m
	pack .q.doit -side left -in .q.button -pady 2m -padx 5m
	pack .q.close -side left -in .q.button -pady 2m -padx 5m
	pack .q.button -fill x -side bottom

	# pack on the canvas
	pack .q.cframe -fill both -expand 1

	# set tabbing and <Return> binding for last label/entry pair
	bind .q.c.f.[expr $i-1].entry.e <Return> "focus .q.doit.b"
	lappend tablist .q.doit.b
	eval Tku_SetTabbing $tablist

	# initialize .q and make it visible
	Query:SetTo
	Query:Resize
	set th [set h [winfo reqheight .q.c.f]]
	set maxh [expr [winfo screenheight .q]*.5]
	if {$th > $maxh} {set th $maxh}
	.q.c conf -width [winfo reqwidth .q.c.f] -height $th
#    wm minsize .q [winfo reqwidth .q.c.f] 200
	wm geom .q +10+10
	set MB(query,where) 0
	Query:Where
	wm deiconify .q
	Tku_Waiting
}

# ---------------------------------------------------------------------------
# Query:OrderBy - add a column to order by
# ---------------------------------------------------------------------------
proc Query:OrderBy {name} {
	set order .q.c.f.order
	set menu $order.mb.m
	set i [$menu index $name]
	$menu entryconfig $i -state disabled
	button $order.$i -text $name -cursor pirate -command \
		"destroy $order.$i; $menu entryconfig $i -state normal"
	pack $order.$i -before $order.mb -side left -padx 1m
}

# ---------------------------------------------------------------------------
# Query:Where - display or hide the where entries
# ---------------------------------------------------------------------------
proc Query:Where {} {
	global MB
	if $MB(show.refs) { set numbers allnum } else { set numbers num }
	loop n [expr [Column $numbers $MB(queryTable)]-1] -1 -1 {
		if $MB(query,where) {
			pack .q.c.f.$n -fill x -after .q.c.f.sel
		} else {
			pack forget .q.c.f.$n
		}
	}
}

# ---------------------------------------------------------------------------
# Query:SetTo - modify the canvas for inserts or selects
# ---------------------------------------------------------------------------
proc Query:SetTo {} {
	global MB
	set frame .q.c.f
        if $MB(show.refs) { set numbers allnum } else { set numbers num }
	if {$MB(do) == "Select"} {
		if ![winfo exists .q.c.f] return
		if [winfo ismapped .q.c.f.sel] return
		pack forget .q.c.f.ins
		pack .q.c.f.sel -fill x -before .q.c.f.0
		loop n 0 [Column $numbers $MB(queryTable)] {
			set MB(save,$n) 0
			pack forget $frame.$n.save
			pack forget $frame.$n.label
			pack $frame.$n.opts -before $frame.$n.entry -side left
		}
		Query:Where
		pack .q.c.f.order -fill x
	} else {
		if ![winfo exists .q.c.f] return
		if [winfo ismapped .q.c.f.ins] return
		if !$MB(query,where) {
			set MB(query,where) 1
			Query:Where
			set MB(query,where) 0
		}
		pack forget .q.c.f.sel
		pack forget .q.c.f.order
		pack .q.c.f.ins -fill x -before .q.c.f.0
		loop n 0 [Column $numbers $MB(queryTable)] {
			pack forget $frame.$n.opts
			pack $frame.$n.label -before $frame.$n.entry -side left
			pack $frame.$n.save -after $frame.$n.entry -side left
		}
	}
	Query:Resize
}

# ---------------------------------------------------------------------------
# Query:Reszie - attempt to resize the canvas
# ---------------------------------------------------------------------------
proc Query:Resize {} {
	update
	set reqh [winfo reqheight .q.c.f]
	.q.c conf -scrollregion "0 0 [winfo reqwidth .q.c.f] $reqh"
	set h [lindex [.q.c conf -height] 4]
	if {$h > $reqh} {.q.c conf -height $reqh}
}

# ---------------------------------------------------------------------------
# Query:GetQual - get the qualification for a column/entry pair
# ---------------------------------------------------------------------------
proc Query:GetQual {i} {
	global MB
	set name $MB(query,name,$i)
	set oper [.q.c.f.$i.opts.e get]
	set data [.q.c.f.$i.entry.e get]
	if {$data == {}} {return ""}
	if ![regexp ' $data] {set data "'$data'"}
	return "$name $oper $data"
}

# ---------------------------------------------------------------------------
# Query:AllCols - set all or no columns to be selected
# ---------------------------------------------------------------------------
proc Query:NoCols {} {
	Query:AllCols 0
}
proc Query:AllCols {{allnone 1}} {
	global MB 
	if $MB(show.refs) { set names allnames } else { set names names }
	foreach table $MB(queryTable) {
		foreach name [Column $names $table] \
  	    {set MB(sel,$name) $allnone}
	}
}

# ---------------------------------------------------------------------------
# Query:Count - Select the count of all rows specified
# ---------------------------------------------------------------------------
proc Query:Count {} {
	Query:AllCols 0
}





# ---------------------------------------------------------------------------
# Query:Clear -    clear all the query entry boxes
# ---------------------------------------------------------------------------
proc Query:Clear {} {
	global MB
        if $MB(show.refs) { set numbers allnum } else { set numbers num }

	loop i 0 [Column $numbers $MB(queryTable)] {
		if !$MB(save,$i) {.q.c.f.$i.entry.e delete 0 end}
		if [winfo exists .q.c.f.n$i] {destroy .q.c.f.n$i}
	}

	set fon .q.c.f.0.entry.e
	if $MB(save,0) {
		loop i 1 [Column $numbers $MB(queryTable)] {
			if !$MB(save,$i) {
				set fon .q.c.f.$i.entry.e
				break
			}
		}
	}
	focus $fon

	Query:Resize
} 

# ---------------------------------------------------------------------------
# Query:Do - execute the bottom button
# ---------------------------------------------------------------------------
proc Query:Do {} {global MB; Query:$MB(do)}

# ---------------------------------------------------------------------------
# Query:Insert -    add an instance to the table being queried
# ---------------------------------------------------------------------------
proc Query:Insert {} {
	global MB

	set table $MB(queryTable)
	if $MB(show.refs) { set numbers allnum } else { set numbers num }

	# build up the insert command
	set nulls 0
	set cols ""
	set vals ""
	loop i 0 [Column $numbers $table] {
		set name [Column name $table $i]
		set data [.q.c.f.$i.entry.e get]
	    if {$data == {}} {
			set nulls 1
			continue
	  	}
		lappend cols $name
		lappend vals "'$data'"
	}

	# warn the user if they forgot to enter some data
	if $nulls {
		if ![Tku_Confirm "Some fields are null, continue?"] {return 0}
	}

	set cols "([join $cols ,])"
	set vals "([join $vals ,])"
	set cmd "insert into $table $cols values $vals;"

	# send the msql
	if ![mi_check_exec MI_DML $cmd "inserting row..."] {
		Tku_Error "Insert failed"
		return 0
	}

	# clear the entry boxes and focus on the first entry box again
	Query:Clear
	return 1
} 

# ---------------------------------------------------------------------------
# Query:Select - select qualifying rows
# ---------------------------------------------------------------------------
proc Query:Select {} {
	global MB

	set where ""
	if $MB(show.refs) { set numbers allnum } else { set numbers num }
	if $MB(query,where) {
		loop i 0 [Column $numbers $MB(queryTable)] {
			if {[set qual [Query:GetQual $i]] != ""} {
				if {$where == ""} {
					set where "where $qual"
				} else {
					append where " and $qual"
				}
			}
		}
		if {$where == ""} {
			Tku_Notice {Please qualify your search or uncheck the "Where..." button}
			return 0
		}
	}

	# make sure the user wants at least a few columns retrieved
	if {[set cols [Query:ColsToSel]] == ""} return

	set select select
	if $MB(queryDistinct) {append select " distinct"}
	append select " $cols from $MB(queryTable) $where"

	if $MB(queryOrder) {
		set slaves [pack slaves .q.c.f.order]
		lvarpop slaves 0
		lvarpop slaves end
		set order ""
		foreach slave $slaves {
			lappend order [lindex [$slave conf -text] 4]
		}
		if {$order != ""} {
			append select "\norder by [join $order ,]"
		}
	}
	append select ";"

	# send query and view results
	if ![mi_check_exec MI_ROWS $select selecting...] {
	  Tku_Error "Select failed"
	  return 0
	}
	if $MB(queryDistinct) {QRes $select 0} else {QRes $select 1}
}

# ---------------------------------------------------------------------------
# Query:ColsToSel - create a list of columns for select
# ---------------------------------------------------------------------------
proc Query:ColsToSel {} {
	global MB 

	set all 1
	set cols ""
	if $MB(show.refs) { set names allnames } else { set names names }
	if { $MB(query,sel) != 2 } {
		foreach name [Column $names $MB(queryTable)] {
			if $MB(sel,$name) {
				lappend cols $name
			} else {
	    			set all 0
			}
		}
	} else {
		set cols "count(*)"
		set all 0
	}

	# use * if all columns in this table are selected
	if $all {set cols *}

	if {$cols == ""} {
		Tku_Notice "Please select one or more columns."
		return ""
	}

	# add oid for updating, except where selecting count(*)
	if { $MB(query,sel) != 2 } {
		if !$MB(queryDistinct) {lappend cols oid}
	}

	return [join $cols ,]
}

# ---------------------------------------------------------------------------
# Query:Close -    closes the query and query results windows
# ---------------------------------------------------------------------------
proc Query:Close {} {
	global MB 
	update
	catch {destroy .qres}
	catch {destroy .q}
}

proc Query:CloseAll {} {
	Query:Close
	foreach w [winfo children .] {
		if [string match .qres* $w] {destroy $w}
	}
}

# ---------------------------------------------------------------------------
# QRes - show the results of a query in a spreadsheet like set of listboxes
# ---------------------------------------------------------------------------
proc QRes:SetWidth {} {
	global MB

	if {[set w [Tku_Entry "Column Width" $MB(-width)]] == ""} {return}
	set MB(-width) $w
	return
}
proc QRes {select {buttons 1}} {
	global MB

	# get data
	Tku_Working fetching...
	mi_get data cols

	if {$data(nrows) == 0} {
		Tku_Waiting
		Tku_Notice {No rows were selected.}
		return 0
	}

	set MB(qr,ncols) $data(ncols)
	set MB(qr,nlines) $data(nrows)

	Tku_Working "extracting $MB(qr,nlines) rows..."

	if $buttons {
		set w .qres
		catch {destroy .qres}
	} else {
	set n 0
	while {[winfo exists [set w .qres$n]]} {incr n}
	}
	toplevel $w
	wm title $w "Query Results"
	wm minsize $w 100 300
	wm geom $w 600x400+300+300
	wm withdraw $w

	set about {
		This window shows the results of your query.  To resize
		columns, drag their right side border.  You can select entries
		by pressing MB1 and dragging.  To extend the selection or add a
		column, press Shift-MB1.
	}
	if $buttons {append about {  You can edit or delete selected rows using
		the buttons at the bottom of the spread sheet."}}
	Tku_MenuBar $w.menu [list About $about]

	Tku_MenuButton $w.menu "File -under 0" \
		"command -label Save... -command {QResSave $w} -under 0" \
		{Saves columns and lines you have selected to a file.} \
		"command -label Print... -command {QResPrint $w} -under 0" \
		{Prints columns and lines you have selected.} \
		"command -label Close -command {update; destroy $w} -under 0" \
		{Closes the results of your query.}

	pack $w.menu -fill x

	Tku_StatusBar $w.info $select
	pack $w.info -fill x


	if $buttons {
		set cmd "QRes:Edit $w.sheet"
	} else {
		set cmd ""
	}
	loop n 0 $MB(qr,ncols) {
		set name [lindex $data(names) $n]
		lappend colinfo [list $name $MB(-width) $cmd]
	}
	Tku_Sheet create $w.sheet $MB(qr,ncols) $MB(qr,nlines) $colinfo
	loop n 0 $MB(qr,ncols) {Tku_Sheet putcol $w.sheet $n $data($n)}
	Tku_Sheet ignore $w.sheet oid

	# add the buttons at the bottom
	frame $w.butts
	Tku_Button $w.all All "Tku_Sheet selectall $w.sheet" \
		"Selects all lines"
	Tku_Button $w.clear Clear "Tku_Sheet selectclear $w.sheet" \
		"Clears the selection"
	lappend tablist $w.all.b $w.clear.b
	pack $w.all $w.clear -side left -padx 2m -pady 2m -in $w.butts
	if $buttons {
		Tku_Button $w.edit "Edit..." "QRes:Edit $w.sheet" \
			"Edits selected instances"
		Tku_Button $w.delete "Delete..." "QRes:Delete $w.sheet" \
	    	"Deletes selected instances"
		lappend tablist $w.edit.b $w.delete.b
		pack $w.edit $w.delete -side left -padx 2m -pady 2m \
			-in $w.butts
	}
	Tku_Button $w.close Close "destroy $w" "Closes the table"
	pack $w.close -side left -padx 2m -pady 2m -in $w.butts
	lappend tablist $w.close.b
	pack $w.butts -fill x -side bottom
	eval Tku_SetTabbing $tablist
	# add actions
	if $buttons {ActionButton $w.menu}
	pack $w.sheet -fill both -expand 1
	wm deiconify $w
	Tku_Waiting
}

# ---------------------------------------------------------------------------
# QRes:Delete - delete a line from the database
# ---------------------------------------------------------------------------
proc QRes:Delete {s} {
	global MB

	set table $MB(queryTable)

	if {[set insts [Tku_Sheet cur $s]] == ""} return
	if [Tku_Confirm "Delete the selected rows?" No Yes] return

	# delete instances 10 at a time
	set oidbox $s.c.[expr $MB(qr,ncols) - 1]
	set n [llength $insts]
	set i 0
	set j 0
	while {$i < $n} {
	incr i 9
	if {$i >= $n} {
 	    set ten [lrange $insts $j end]
	} else {
	    set ten [lrange $insts $j $i]
	}
	set j [incr i]

		foreach inst $ten {
	      set oid "'[$oidbox get $inst]'"
	      if ![info exists cmd] {
	      	set cmd "delete from $table where oid = $oid"
	      } else {
	      	append cmd "\n\tor oid = $oid"
	      }
		}
		append cmd ";"

		if ![mi_check_exec MI_DML $cmd deleting..] {
			Tku_Error "Could not delete rows"
			return 0
		}
		unset cmd
	}

	# delete the lines from the list boxes and update line number column
	foreach i [lreverse $insts] {Tku_Sheet delete $s $i}
}

# ---------------------------------------------------------------------------
# QRes:Edit - start editing one or more rows
# ---------------------------------------------------------------------------
proc QRes:Edit {s} {
	global MB

	if [winfo exists .qed] return
	if {[set toedit [Tku_Sheet currentsel $s]] == ""} return

	# if editing more than one line, let the user cancel mid way through
	set MB(CancelAll) 0; set MB(MultiEdit) 0
	if {[llength $toedit] != 1} {set MB(MultiEdit) 1}
	foreach i $toedit {if !$MB(CancelAll) {QRes:EditOne $s $i}}
}

# ---------------------------------------------------------------------------
# QRes:EditOne - edit a single instance
# ---------------------------------------------------------------------------
proc QRes:EditOne {s i} {
	global MB

	set table $MB(queryTable)
	if $MB(show.refs) { set names allnames } else { set names names }
	set w [toplevel .qed]
	wm title $w "Edit Row #$i"
	wm iconname $w "Edit #$i"

	canvas $w.c -yscroll "$w.ys set"
	scrollbar $w.ys -command "$w.c yview"
	frame $w.c.f
	$w.c create window 0 0 -anchor nw -window $w.c.f

	set ndatacol [expr $MB(qr,ncols) - 1]
	set max 0
	foreach name [Column $names $table] {
		if {[set n [string length [Column def $table $name]]] > $max} {
			set max $n
		}
	}
	loop c 0 $ndatacol {
	set box $w.c.f.$c
	set lastbox $w.c.f.[expr $c-1]
	set name [Tku_Sheet name $s $c]
	Tku_EntryBox $box [Column def $table $name] \
		[Tku_Sheet get $s $c $i] -width 50
	$box.label conf -width $max -anchor e
	  if $c {bind $lastbox.e <Return> "focus $box.e"}
	  lappend tablist $box.e
	pack $box -anchor w -fill x
	}
	if {$ndatacol > 1} {bind $lastbox.e <Return> "focus $w.ok.b"}
	frame $w.bs
	Tku_Button $w.ok Commit "QRes:EditOk $s $i" \
	"Commit changes to this instances" 1
	Tku_Button $w.cancel Cancel "destroy $w" "Cancel editing of this row"
	pack $w.ok $w.cancel -in $w.bs -side left -padx 5m -pady 5m
	lappend tablist $w.ok.b $w.cancel.b

	# add a button to cancel multiple instance editing
	if $MB(MultiEdit) {
	  Tku_Button $w.canall "Cancel All" \
	    "set MB(CancelAll) 1; destroy .qed" \
	    "Cancel editing this and all other selected rows"
 	pack $w.canall -in $w.bs -padx 5m -side left
	  lappend tablist $w.canall.b
	}

	eval Tku_SetTabbing $tablist
	pack $w.bs -side bottom -fill x
	pack $w.ys -fill y -side right
	pack $w.c -fill both -expand 1
	wm withdraw $w
	update
	set wid [winfo reqwidth $w.c.f]
	set th [set h [winfo reqheight $w.c.f]]
	set maxh [expr [winfo screenheight $w]*.8]
	if {$th > $maxh} {set th $maxh} else {pack forget $w.ys}
	$w.c conf -scrollregion "0 0 $wid $h" -width $wid -height $th
	wm minsize $w 100 100
	Tku_Center $w
	tkwait visibility $w.c.f.0.e
	focus $w.c.f.0.e
	tkwait window $w
}

# ---------------------------------------------------------------------------
# QRes:EditOk - submit update command
# ---------------------------------------------------------------------------
proc QRes:EditOk {s e} {
	global MB

	set table $MB(queryTable)

	set ncols $MB(qr,ncols)
	incr ncols -1
	set oid [Tku_Sheet get $s $ncols $e]

	set nulls 0
	set sets ""
	loop col 0 $ncols {
		set name [Tku_Sheet name $s $col]
		set data [.qed.c.f.$col.e get]
		if {$data == [Tku_Sheet get $s $col $e]} continue
		if {$data == "(null)"} {set data {}}
		if {$data == {}} {
			set nulls 1
		} else {
			lappend sets "$name = '$data'"
		}
	}
	if {$sets == ""} {
		destroy .qed
		Tku_Notice "No changes made to instance #$e"
		return
	}

	if $nulls {
		if ![Tku_Confirm "Some fields are null, continue?"] return
	}

	set q "update $table set [join $sets ,] where oid = '$oid';"
	if ![mi_check_exec MI_DML $q updating...] {
		Tku_Error "Update failed"
		return 0
	}
	loop col 0 $ncols {
		Tku_Sheet replace $s $col $e [.qed.c.f.$col.e get]
	}
	update
	destroy .qed
}

proc QResSave {w} {Tku_Sheet save $w.sheet}
proc QResPrint {w} {Tku_Sheet print $w.sheet}
proc ActionColNum {name {allok 1} {confirm 1}} {
	Tku_Sheet:ColNum .qres.sheet $name $allok $confirm
}
proc ActionGetCol {n {allok 1}} {Tku_Sheet:GetCol .qres.sheet $n $allok}
proc ActionFields {} {Tku_Sheet:ColNames .qres.sheet}


proc Query:Table {table} {
	set select "select * from $table;"
	# send query and view results
	if ![mi_check_exec MI_ROWS $select selecting...] {
	  Tku_Error "Select failed"
	  return 0
	}
	QRes $select 0
}
#------------------------------------------------------------------------------- 
#  Query:GetAllOpers
#
#   Returns a list of all operators which are allows in the Where
#  specification. These are all binary functions returning Boolean
#  values.
#------------------------------------------------------------------------------- 
proc Query:GetAllOpers {} {
	set Qry "
SELECT DISTINCT O.operator_name
  FROM operators O, functions F
 WHERE O.operator_function = F.function_name
   AND F.function_nargs    = 2
   AND F.function_rettype  = (SELECT UNIQUE oid::typeoid
                                FROM types
                               WHERE type_name = 'boolean');"
	if ![mi_check_exec MI_ROWS $Qry selecting...] {
		Tku_Error "Select failed"
		return 0
        }

	mi_get data cols

        if {$data(nrows) == 0} {
                return {}
        }

	list RetOps   
	puts $data(nrows)

	foreach i $data(0) {
		lappend RetOps $i
	}

  return $RetOps
}

