#
# Tku Utilities
#
# $Header: /usr/local/devel/montage/contrib/MiBrowse/lib/RCS/tku.tcl,v 1.4 1995/01/02 23:29:39 pbrown Exp pbrown $
#
# A bunch of utilites for Tcl/Tk
# Send bug reports to: jimbo@crseo.ucsb.edu
# 

proc Tku_HelpWindow {frame} {
    if {$frame == "."} {set hw .tku_help} else {set hw $frame.tku_help}
    label $hw -relief sunken -textvariable TKU(popup,msg) -anchor w
    pack $hw -side bottom -fill x
}    
# set popup help for buttons and menu entries
proc Tku_Popup {help w {entry ""}} {
    global TKU
    if {$entry != ""} {
    	set TKU(popup,$w,$entry) [Tku_HelpText $help]
    } else {
    	set TKU(popup,$w) [Tku_HelpText $help]
    }
}
# popup help at the edge of a button
proc Tku_Popup:Edge {w X Y} {
    global TKU
    if [catch {set TKU(popup,msg) $TKU(popup,$w)}] return
    if !$TKU(popup,enabled) return
    set px [expr [winfo rootx $w]+[winfo width $w]-10]
    set py [expr [winfo rooty $w]+[winfo height $w]-10]
    if {$px < $X && $py < $Y} {incr px 15 ; incr py 15}
    Tku_Popup:Post $w $TKU(popup,$w) $px $py
}
# popup help at the edge of a menu entry
proc Tku_Popup:Menu {w y} {
    global TKU
    set i [$w index @$y]
    if {$TKU(popup,entry) == "$w,$i"} return
    Tku_Popup:UnPost
    if [catch {set TKU(popup,msg) $TKU(popup,$w,$i)}] return
    set TKU(popup,entry) "$w,$i"
    if !$TKU(popup,enabled) return
    Tku_Popup:Post $w $TKU(popup,$w,$i) \
        [expr [winfo rootx $w]+[winfo width $w]+2] \
        [expr [winfo rooty $w]+[$w yposition $i]+2]
}
# post a popup help message
proc Tku_Popup:Post {w h x y} {
    global TKU
    catch {destroy .popup_help}
    toplevel .popup_help -bd 4 -relief ridge
    wm overrideredirect .popup_help 1
    wm sizefrom .popup_help user
    wm transient .popup_help $w
    message .popup_help.m -text $h -width 2i
    pack .popup_help.m -padx 1m -pady 1m
    wm geom .popup_help +$x+$y
    wm deiconify .popup_help
}
# unpost any popup help message
proc Tku_Popup:UnPost {} {
    global TKU
    catch {destroy .popup_help}
    set TKU(popup,entry) ""
    set TKU(popup,msg) ""
}

#
# Tku_Listbox -    make a vertically and horizontally scrollable listbox
#        with gobs of bindings
# inputs:
#    frame   frame to create for listbox and scrollbars
#    bind    <Double-Button-1> binding
#    args    more args to listbox
#
proc Tku_Listbox {frame {bind {}} args} {
    frame $frame
    label $frame.title -text ""
    frame $frame.box -relief flat -bd 0
    frame $frame.active -relief flat
    eval listbox $frame.listbox -yscroll \"$frame.vs set\" -setgrid 1 \
	-xscroll \"$frame.hs set\" $args
    scrollbar $frame.vs -command "$frame.listbox yview" 
    scrollbar $frame.hs -command "$frame.listbox xview" -orient horizontal
    pack $frame.title -fill x
    pack $frame.hs -fill x -in $frame.box -side bottom
    pack $frame.vs -in $frame.box -side right -fill y
    pack $frame.listbox -in $frame.active -expand 1 -fill both
    pack $frame.active -expand 1 -fill both -in $frame.box
    pack $frame.box -expand 1 -fill both
    if {$bind != {}} {
        bind $frame.listbox <Return> $bind
        bind $frame.listbox <Double-Button-1> $bind
    }
    pack $frame.active -expand 1 -fill both
    set fg [lindex [$frame.listbox conf -fg] 4]
    set bg [lindex [$frame.listbox conf -bg] 4]
    bind $frame.listbox <ButtonRelease-1> {+focus %W}
    bind $frame.listbox <FocusIn> "$frame.active conf -bg $fg"
    bind $frame.listbox <FocusOut> "$frame.active conf -bg $bg"
    return $frame
}

#
# Tku_ListboxFill - fill in a list box
#
proc Tku_ListboxFill {box tofill} {foreach i $tofill {$box insert end $i}}

#
# Tku_ListboxClear - clear a list box
#
proc Tku_ListboxClear box {$box delete 0 end}

#
# Tku_ListboxDump -	do a $w get 0 end
#
proc Tku_ListboxDump {w} {
    if ![$w size] {return ""}
    loop n 0 [$w size] {lappend r [$w get $n]}
    return $r
}

#
# Tku_ListboxGet - get selected elements of a listbox
#
proc Tku_ListboxGet {w} {
    if {[set sel [$w curselection]] == ""} {return ""}
    foreach e $sel {lappend r [$w get $e]}
    return $r
}

#
# Almost Macintoch keyboard control of a listbox
#
# set the view to include the selected entry in a nice way
proc Tku_Listbox:Yview {w {sel ""}} {
    if {$sel != ""} {
	$w select from $sel
    } else {
	if {[set sel [$w cursel]] == ""} return
    	if {[llength $sel] > 1} return
    }
    set topvis [$w nearest 0]
    set botvis [$w nearest [winfo height $w]]
    if {$sel >= $topvis && $sel <= $botvis} return
    if {$sel < $topvis} {
	$w yview $sel
    } else {
	set l [lindex [split [lindex [$w conf -geom] 4] x] 1]
	$w yview [expr $sel-$l+1]
    }
}
proc Tku_Listbox:MoveUp {w} {
    set sel [$w cur]
    case [llength $sel] in {
	0    {set sel [$w nearest 0]}
	1    {if {[incr sel -1] < 0} {set sel 0}}
	default {set sel [lindex $sel 0]}
    }
    Tku_Listbox:Yview $w $sel
}
proc Tku_Listbox:MoveDown {w} {
    set sel [$w cur]
    case [llength $sel] in {
            0    {set sel [$w nearest 0]}
            1    {incr sel}
            default {set sel [llast $sel]}
    }
    if {$sel == [$w size]} {incr sel -1}
    Tku_Listbox:Yview $w $sel
}
proc Tku_Listbox:MoveTop {w} {Tku_Listbox:Yview $w 0}
proc Tku_Listbox:MoveBottom {w} {Tku_Listbox:Yview $w [expr [$w size]-1]}
set TKU(inListbox) ""
proc Tku_Listbox:MoveTo {w a} {
    global TKU
    if {$a == ""} return
    if {$TKU(inListbox) != $w} {
	set TKU(inListbox) $w
	set TKU(matchString) $a
	after 500 {set TKU(inListbox) ""}
	set start 0
    } else {
	append TKU(matchString) $a
	set start [lindex [$w cur] 0]
    }
    loop n $start [$w size] {
	if [string match "[string tolower $TKU(matchString)*]" \
		[string tolower [$w get $n]]] {
	    Tku_Listbox:Yview $w $n; return
	}
    }
}

#
# Tku_Tab - Tab between widgets
#
proc Tku_Tab {args} {
    set cur [lsearch $args [focus]]
    set next [expr $cur+1]
    if {$next >= [llength $args]} {
        set next 0
    }
    set curw [lindex $args $cur]
    set nextw [lindex $args $next]
    focus $nextw
}

#
# Tku_SetTabbing - setup forward and backward tabbing for a list of windows
#
proc Tku_SetTabbing {args} {
    global TKU
    if {[llength $args] < 2} return
    set m 1
    set num [llength $args]
    loop n 0 $num {
	set cur [lindex $args $n]
	if {$m < $num} {set next [lindex $args $m]} \
	else {set next [lindex $args 0]}
	incr m
	foreach b {<Tab> <Right> <Down>} {
	    if {[bind $cur $b] == "" && [bind [winfo class $cur] $b] == ""} {
		bind $cur $b "focus $next"
	    }
	}
	foreach b {<Shift-Tab> <Left> <Up>} {
	    if {[bind $next $b] == "" && [bind [winfo class $next] $b] == ""} {
		bind $next $b "focus $cur"
	    }
	}
    }
}



#
# Tku_Entry - simple modal entry dialog
#
# inputs:
#    prompt        prompt string for dialog
#    default        default value for entry box
#
# returns:
#    valid entry string or "" if canceled
#
proc Tku_Entry {prompt {default {}} {title Entry}} {
    global TKU
    catch {destroy .ed}
    toplevel .ed
    wm title .ed $title
    frame .ed.bot 
    Tku_Button .ed.ok OK {set TKU(Return) [.ed.entry.e get]} {} 1
    Tku_Button .ed.cancel Cancel {set TKU(Return) {}}
    pack .ed.ok -in .ed.bot -side left -padx 2m -pady 2m
    pack .ed.cancel -in .ed.bot -side left -padx 2m -pady 2m

    frame .ed.top
    Tku_EntryBox .ed.entry "$prompt:" $default
    bind .ed.entry.e <Return> {.ed.ok.b flash; .ed.ok.b invoke}
    pack .ed.entry -in .ed.top -side left -padx 2m -pady 5m
    pack .ed.top -ipadx 5m -fill x
    pack .ed.bot -fill x
    Tku_SetTabbing .ed.entry.e .ed.ok.b .ed.cancel.b
    Tku_Center .ed
    tkwait visibility .ed
    grab .ed
    focus .ed.entry.e
    tkwait variable TKU(Return)
    destroy .ed
    return $TKU(Return)
}

#
# Tku_Center -    place a window in the center of the screen
#
proc Tku_Center {w} {
    wm withdraw $w
    update idletasks
    set x [expr [winfo screenwidth $w]/2 - [winfo reqwidth $w]/2 \
            - [winfo vrootx [winfo parent $w]]]
    set y [expr [winfo screenheight $w]/2 - [winfo reqheight $w]/2 \
            - [winfo vrooty [winfo parent $w]]]
    wm geom $w +$x+$y
    wm deiconify $w
}
    

proc Tku_SelectBtn {block msg {title Notice} {bitmap info} args} {
    global TKU

    if ![llength $args] {set args OK}
    set TKU(Return) -1
    set n 0
    while {[winfo exists [set w .mb$n]]} {incr n}
    toplevel $w
    wm title $w $title
    set pad {padx 1c pady 1c}
    frame $w.top
    label $w.bitmap -bitmap $bitmap
    message $w.msg -text $msg -width 3i
    pack $w.bitmap -in $w.top -padx 5m -side left
    pack $w.msg -in $w.top  -padx 5m -pady 5m -fill x -side left -expand 1
    frame $w.bot
    set n 0
    set def 1
    foreach button $args {
	if $block {
		set cmd "set TKU(Return) $n"
	} else {
		set cmd "destroy $w"
	}
        Tku_Button $w.$n $button $cmd {} $def
        set def 0
        pack $w.$n -in $w.bot -side left -padx 1c -pady 5m
        lappend tablist $w.$n.b
        incr n
    }
    eval Tku_SetTabbing $tablist
    pack $w.top $w.bot -fill x -expand 1
    Tku_Center $w
    if $block {
    focus $w.0.b
    grab $w
    tkwait variable TKU(Return)
    destroy $w
    return $TKU(Return)
    }
}
proc Tku_Confirm {msg {y Yes} {n No} } {
    case [Tku_SelectBtn 1 $msg Confirm error $y $n] in {
        0    {return 1}
        1    {return 0}
    }
}
proc Tku_Confirm3 {msg 1 2 3} {
    Tku_SelectBtn 1 $msg Confirm error $1 $2 $3
}
proc Tku_Notice { msg } {
    Tku_SelectBtn 0 $msg Notice
}
proc Tku_Error { msg } {
    Tku_SelectBtn 1 $msg Error error
}
proc Tku_Info { msg } {
    Tku_SelectBtn 0 $msg Info
}
proc Tku_Help { msg } {
    Tku_SelectBtn 0 [Tku_HelpText $msg] Help
}

#
# Tku_HelpText - read a help file or format a short help message
#
proc Tku_HelpText {msg} {
    # If msg does not begin with an @ sign, assume it's a simple
    # help message that just needs to be trimed and formatted
    if ![regsub {^@} $msg {} file] {
        foreach line [split $msg \n] {lappend r [string trim $line]}
        return [string trim [join $r]]
    }

    
    # Otherwise, assume the msg is a filename to be opened and read
    # from the help directory
    if {[set dir [Tku_HelpDir]] == ""} return ""
    if [catch {set fp [open $dir/$file]} err] {
	Tku_Error "Couldn't open help file:  $dir/$file:  $err"
        return ""
    } else {
        set msg [read $fp]
        close $fp
        return $msg
    }
}

#
# Tku_HelpDir - set and/or return the help directory
#
proc Tku_HelpDir {{dir ""}} {
    global TKU

    if {$dir != ""} {
    	# check that the new help directory exists
	if ![file readable $dir] {
	    Tku_Error "Can't open help directory:  $dir"
	    return ""
	} else {
	    set TKU(Helpdir) $dir]
	}
    } elseif ![info exists TKU(Helpdir)] {
	# have the user locate the help directory
        set TKU(Helpdir) [Tku_FdDir -msg "Please locate the help directory"]
    }
    return $TKU(Helpdir)
}

#
# Tku_Button - make a frame and a button with <Return> and <Focus> binding
#
# inputs:
#    w		button window
#    text	button text
#    cmd	command for <Button-1> and <Return>
#    help	popup help text
#    default	add an extra frame to signify a default choice
#
proc Tku_Button {w text cmd {help {}} {default 0}} {
    if {[set l [string length $text]] < 10} {
        set l 10
    }
#    incr l 2
    frame $w -relief flat -bd 2
    frame $w.f1 -relief flat -bd 1
    set bg [lindex [$w conf -bg] 4]
    frame $w.f -relief flat -bd 0
    button $w.b -text $text -command $cmd -width $l -relief raised
    set bd [lindex [$w.b conf -bd] 4]
    set bd2 [expr $bd*2]
    bind $w.b <FocusIn> "$w.b conf -state active;$w conf -bg black"
    bind $w.b <FocusOut> "$w.b conf -state normal;$w conf -bg $bg"
    bind $w.b <Return> "$w.b flash;$w.b invoke"
    pack $w.b -in $w.f -padx 2 -pady 2
#    pack $w.f -padx 1 -pady 1 -in $w.f1
    pack $w.f -in $w.f1
    pack $w.f1 -in $w -padx 0 -pady 0
    if $default {$w.f conf -bd 1 -relief sunken}
    if {$help != {}} {Tku_Popup $help $w.b}
    return $w
}

proc Tku_PullDown:Bitmap {} {
    global TKU

    if ![info exists TKU(pulldown)] {
	set TKU(pulldown) [Tku_TmpDir]/pulldown.bit
	if ![file exists $TKU(pulldown)] {
	    set fp [open $TKU(pulldown) w]
	    puts $fp {
#define Down_width 12
#define Down_height 12
static char Down_bits[] = {
   0xf0, 0x00, 0xf0, 0x00, 0xf0, 0x00, 0xf6, 0x06, 0xfe, 0x07, 0xfc, 0x03,
   0xf8, 0x01, 0xf0, 0x00, 0x60, 0x00, 0xff, 0x0f, 0xff, 0x0f, 0xff, 0x0f};
}
	    close $fp
	}
    }
    return @$TKU(pulldown)
}

proc Tku_EntryBox {w {label ""} {def ""} args} {
    global TKU

    frame $w -bd 0
    frame $w.active -bd 2 -relief flat
    frame $w.space -bd 2 -relief flat
    eval entry $w.e $args
    pack $w.e -in $w.space -side left
    pack $w.space -in $w.active
    bind $w.e <FocusIn> [bind Entry <FocusIn>]
    bind $w.e <FocusOut> [bind Entry <FocusIn>]
    bind $w.e <FocusIn> "+$w.active conf -bg black"
    bind $w.e <FocusOut> "+$w.active conf -bg [lindex [$w.space conf -bg] 4]"
    $w.e insert 0 $def
    if {$label != ""} {
	label $w.label -text $label
	pack $w.label -side left -padx 2m
    }
    pack $w.active -side left -padx 2m
    return $w
}

proc Tku_PullDown {w {label ""} {choices ""} {defchoice ""} args} {
    global TKU

    frame $w -bd 0 -relief flat
    frame $w.active -bd 2 -relief flat
    frame $w.space -bd 2 -relief flat
    eval entry $w.e $args
    menubutton $w.b -menu $w.b.m -bitmap [Tku_PullDown:Bitmap] \
	-relief raised -bd 2
    pack $w.e $w.b -in $w.space -side left
    pack $w.space -in $w.active
    if {$label != ""} {
	label $w.label -text $label
	pack $w.label -side left -padx 1m
    }
    pack $w.active -side left
    bind $w.e <FocusIn> [bind Entry <FocusIn>]
    bind $w.e <FocusOut> [bind Entry <FocusIn>]
    bind $w.e <FocusIn> "+$w.active conf -bg black"
    bind $w.e <FocusOut> "+$w.active conf -bg [lindex [$w.space conf -bg] 4]"
    bind $w.e <Down> "tk_mbPost $w.b; $w.b.m activate 0"
    bind $w.e <Up> "tk_mbPost $w.b; $w.b.m activate last"
    $w.e insert 0 $defchoice
    menu $w.b.m
    foreach c $choices {
	$w.b.m add command -label $c \
	    -command "$w.e delete 0 end; $w.e insert 0 {$c}"
    }
    return $w
}


#
# Tku_MenuBar -	make a menu bar with optional Help button
#
# Inputs:
#    w		menu bar frame
#    args	list of 2 element lists, 0 = help entry, 1 = HelpText
#
proc Tku_MenuBar {w args} {
    global TKU
    frame $w
    if {[llength $args] != 0 || [info exists TKU(popup,enabled)]} {
        menubutton $w.help -text "Help" -menu $w.help.m -under 0
        menu $w.help.m
        $w.help.m add checkbutton -offvalue 0 -onvalue 1 -under 0 \
            -variable TKU(popup,enabled) -label "Popup Help" 
        foreach help $args {
            $w.help.m add command -label [lindex $help 0] \
                 -command "Tku_Help {[lindex $help 1]}"
        }
        pack $w.help -side right -padx 2m -in $w
    }
    Tku_MenuBar:Setup $w
    bind $w <Destroy> "tk_menuBar $w {}"
    return $w
}

proc Tku_MenuBar:Accel {mod w key} {
    global tk_priv

    if [catch {set menubar $tk_priv(menuBarFor[winfo toplevel $w])}] return

    foreach mb $tk_priv(menusFor$menubar) {
	set menu [lindex [$mb conf -menu] 4]
	loop n 0 [expr [$menu index last]+1] {
	    if [catch {set ac [lindex [$menu entryconf $n -accel] 4]}] continue
	    if {$ac == ""} continue
	    if [regexp -nocase ^$mod.*-$key $ac] {
		Tku_Menu:Invoke $mb $menu $n; return
	    }
	}
    }
    # if we didn't find an accelerator, flash the window
    Tku_Flash $w
}

proc Tku_Menu:Invoke {mb menu entry} {
    global TKU
    if {[lindex [$mb config -state] 4] == "disabled"} return
    set popup $TKU(popup,enabled)
    set TKU(popup,enabled) 0
    set bind [bind $menu <Any-Enter>]
    bind $menu <Any-Enter> { # nothing to do }
    $mb conf -state active
    tk_mbPost $mb
    for {set n 0} {$n < 3} {incr n} {
        $menu entry $entry -state normal; update; after 50
        $menu entry $entry -state active; update; after 50
    }
    tk_mbUnpost
    $mb conf -state normal
    $menu invoke $entry
    set TKU(popup,enabled) $popup
    # catch this bind because the accelator may have deleted $menu
    catch {bind $menu <Any-Enter> $bind}
}

proc Tku_MenuBar:Setup {w} {
    global tk_priv
    set n 0
    set bs ""
    while {[winfo exists $w.$n]} {lappend bs $w.$n;incr n}
    if [winfo exists $w.help] {lappend bs $w.help}
    eval tk_menuBar $w $bs

}
    
#
# Tku_MenuButton - Add a button to the left of a Tku_MenuBar frame
#
# Inputs:
#	w	menu bar frame
#	name	name of button
#	args	list of menu add commands and corresponding HelpText,
#		i.e., "cmd1" "help1" "cmd2" "help2" "cmd3" help3" ...
#
proc Tku_MenuButton {w name args} {
    set n 0
    while {[winfo exists $w.$n]} {incr n}
    eval menubutton $w.$n -text $name -menu $w.$n.m 
    menu $w.$n.m
    if $n {
        pack $w.$n -after $w.[expr $n-1] -side left -padx 2m
    } else {
    	if [winfo exists $w.help] {
            pack $w.$n -before $w.help -side left -padx 2m
	} else {
            pack $w.$n -side left -padx 2m
	}
    }
    if {[llength $args] > 0} {
	set m 0
	for {set e 0} {$e < [llength $args]} {incr e 2} {
	    eval $w.$n.m add [lindex $args $e]
	    set h [lindex $args [expr $e+1]]
	    if {$h != ""} {Tku_Popup $h $w.$n.m $m}
	    incr m
	}
    }
    Tku_MenuBar:Setup $w
    return $w.$n
}

#
# Tku_ExpTilde - expand ~dir to full path name
#
proc Tku_ExpTilde {d} {
    case $d in {
        ~/    {return [file dirname ~/]}
        default {return [file dirname $d/]}
    }
}

#
# Tku_SearchPath -    search for an executable in user's PATH
#
# inputs:
#    prog    prog to search for
#    apath    other directories to search, separated by :
#
# returns:
#    location of prog, perhaps adding the location to env(PATH),
#    or "" if not found
#
proc Tku_SearchPath {prog {apath {}}} {
        global env
        foreach p [split $env(PATH) :] {
        if [catch {file exists $p}] continue
                if [file exists $p/$prog] {
                        return [Tku_ExpTilde $p]/$prog
                   }
        }
        foreach p [split $apath :] {
        if [catch {file exists $p}] continue
                   if [file exists $p/$prog] {
                           set env(PATH) $env(PATH):[Tku_ExpTilde $p]
                           return $p/$prog
        }
        }
        return ""
}

proc Tku_DumpArray { array } {
    upvar 1 $array a
    foreach n [array names a] {
        lappend r "set ${array}($n) {$a($n)}"
    }
    if [info exists r] {return $r} else {return ""}
}


proc llast {l} {
    lindex $l [expr [llength $l]-1]
}
proc lfirst {l} {
    lindex $l 0
}
proc lreverse {l} {
    for {set n [expr [llength $l]-1]} {$n >= 0} {incr n -1} {
	lappend r [lindex $l $n]
    }
    return $r
}
proc ldelete {l e} {
   while {[set i [lsearch  $l $e]] != -1} {
	set l [lreplace $l $i $i]
   }
   return $l
}


proc Tku_Edit {title text edited} {
    global TKU
    toplevel .qedit
    wm minsize .qedit 100 100
    pack [label .qedit.l -text $title -relief raised -bd 2] -fill x
    frame .qedit.box
    pack [text .qedit.text -width 80 -height 10 -wrap char \
        -yscroll ".qedit.s set"] \
        -side left -expand 1 -fill both -in .qedit.box
    .qedit.text insert 0.0 $text
    pack [scrollbar .qedit.s -command ".qedit.text yview"] \
        -side right -fill y -in .qedit.box
    pack .qedit.box -expand 1 -fill both
    set TKU(qedit) 0
    frame .qedit.frame
    pack [Tku_Button .qedit.ok OK {set TKU(qedit) 1} {} 1] \
    	-side left -padx 5m -pady 2m -in .qedit.frame
    if {$edited != -1} {
    	wm title .qedit Edit
    	pack [Tku_Button .qedit.cancel Cancel {set TKU(qedit) 0}] \
    	    -side left -padx 5m -pady 2m -in .qedit.frame
    } else {
    	wm title .qedit View
	.qedit.text conf -state disabled
    }
    pack .qedit.frame -fill x
    Tku_SetTabbing .qedit.text .qedit.ok.b .qedit.cancel.b
    Tku_Center .qedit
    focus .qedit.text
    grab .qedit
    tkwait variable TKU(qedit)
    if {$TKU(qedit) && $edited != -1} {
	upvar $edited myedited
	set myedited [.qedit.text get 0.0 end]
    }
    destroy .qedit
    update
    return $TKU(qedit)
}

proc Tku_MISQL { title text edited } {
    global TKU
    toplevel .qedit
    wm minsize .qedit 100 100
    pack [label .qedit.l -text $title -relief raised -bd 2] -fill x
    frame .qedit.box
    pack [text .qedit.text -width 80 -height 10 -wrap char \
        -yscroll ".qedit.s set"] \
        -side left -expand 1 -fill both -in .qedit.box
    .qedit.text insert 0.0 $text
    pack [scrollbar .qedit.s -command ".qedit.text yview"] \
        -side right -fill y -in .qedit.box
    pack .qedit.box -expand 1 -fill both
    set TKU(qedit) 0
    frame .qedit.frame
    pack [Tku_Button .qedit.ok Submit { File:SendSQLDisp [ .qedit.text get 1.0 end ]} {} 1] \
        -side left -padx 5m -pady 2m -in .qedit.frame
    pack [Tku_Button .qedit.frame.clear Clear {.qedit.text delete 1.0 end; return}] \
      -side left -padx 5m -pady 2m -in .qedit.frame
    pack [Tku_Button .qedit.cancel Close {destroy .qedit; return}] \
        -side left -padx 5m -pady 2m -in .qedit.frame
    pack .qedit.frame -fill x
    Tku_SetTabbing .qedit.text .qedit.ok.b .qedit.cancel.b
    Tku_Center .qedit
    focus .qedit.text
}

#
# Filedialog utilities
#

#
# Tku_FdGetHier - run a directory back to /
#
# input:
#    dir    top-level directory (e.g., /home/jimbo/pg)
#
# returns:
#    list of directories below dir (e.g., {/home/jimbo} {/home} {/})
#
proc Tku_FdGetHier {w dir} { 
    if { $dir == "/" } { return "" }
    set h ""
    set d [file dirname $dir]
    while {$d != "/"} {
        lappend h $d
        set d [file dirname $d]
    }
    lappend h /
    return $h
}
#
# Tku_FdUpdate -    updates the listbox, file and glob entries
#
# action:
#    globs the current directory, adds "/" to the end of directories,
#    fills in the listbox, file and glob entry boxes
#
proc Tku_FdUpdate {w} {
    global TKU
    set dir [pwd]
    # if there never was a glob, focus on the entry box
    # so the user can add one
    if {$TKU(fd,$w,glob) == "*" || $TKU(fd,$w,glob) == ""} {
        set TKU(fd,$w,all) 1
        $w.select.glob delete 0 end
    }
    # fill in the hierarchy menu button
#    if {[set l [string length $dir]] > $TKU(fd,$w,width)} {
#        set i [expr $l-$TKU(fd,$w,width)]
#        set d "...[string range $dir $i end]"
#    } else {
#        set d $dir
#    }
    set d $dir
    $w.box.menu_title configure -text $d
    Tku_Working Scanning...
    $w.box.menu.m delete 0 last
    foreach d [Tku_FdGetHier $w $dir] {
        $w.box.menu.m add command -label $d -command "Tku_FdCwd $w $d"
    }
    if [info exists TKU(fd,$w,lastdir)] {
        if {$TKU(fd,$w,lastdir) != "/" && $TKU(fd,$w,lastdir) != $dir} {
            $w.box.menu.m add separator
            $w.box.menu.m add command -label $TKU(fd,$w,lastdir) -command "Tku_FdCwd $w $TKU(fd,$w,lastdir)"
        }
    }
    # clear and fill the list box
    Tku_ListboxClear $w.box.listbox
    update
    if {$dir != "/"} {
        set files ../
    }
    if $TKU(fd,$w,all) {
        if $TKU(fd,$w,show) {set glob {.* *}} else {set glob *}
    } else {
        set glob $TKU(fd,$w,glob)
    }

    # get list of files
    # NB: check globbing results because [glob zz] returns "zz" whether or
    # not file zz exists
    set filelist [eval glob -nocomplain $glob]
    if {$filelist == $glob && ![file exists $glob]} {set filelist ""}
    foreach f [lsort $filelist] {
        if {$f == "." || $f == ".."} continue
        if [file isdirectory $f] {
            lappend files $f/
        } else {
            if !$TKU(fd,$w,SetDir) {lappend files $f}
        }
    }
    Tku_ListboxFill $w.box.listbox $files
    Tku_Waiting $w.title
}

#
# Tku_FdCwd    - change to a new directory
#
# input:
#    d    new directory
#
# action:
#    cd's to $d, sets TKU(fd,$w,lastdir), and calls Tku_FdUpdate
#
proc Tku_FdCwd {w d} {
    global TKU
    if ![file readable $d] {Tku_Notice "Can't open directory $d."; return}
    set TKU(fd,$w,lastdir) [pwd]
    cd $d
    set TKU(fd,$w,directory) [pwd]
    Tku_FdUpdate $w
}

#
# Tku_FdSelect -    narrows globbing to selected files
#
# action:
#    sets TKU(fd,$w,glob) from contents of glob entry box, sets TKU(fd,$w,all),
#    and calls Tku_FdUpdate.  If glob entry is empty, just focuses on it.
#
proc Tku_FdSelect {w} {
    global TKU
    set glob [$w.select.glob get]
    if {$glob == ""} {
        focus $w.select.glob
    } else {
        set TKU(fd,$w,all) 0
        set TKU(fd,$w,glob) $glob
        Tku_FdUpdate $w
    }
}

proc Tku_FdView {w} {
    global TKU
    set i [$w.box.listbox curselection]
    case [llength $i] in {
        0 return
        1 {set file [$w.box.listbox get $i]}
        default return
    }
    Tku_ViewFile $file
}

proc Tku_ViewFile {file} {
    if [catch {set fp [open $file]}] {
        Tku_Error "Can't open file: $file"
        return 0
    }
    set n 0
    while {[winfo exists .view$n]} {incr n}
    set win .view$n
    toplevel .view$n
    wm title .view$n "View file: $file"
    wm minsize .view$n 100 100
    Tku_MenuBar .view$n.menu
    Tku_MenuButton .view$n.menu "File -under 0" \
	"command -label Save -command {Tku_ViewFileSave $n} -accel C-S -under 0" \
	{Saves contents to a file} \
        "command -label Close -command {destroy .view$n} -accel C-W -under 0" \
	{Closes window} 
    pack .view$n.menu -fill x
    frame .view$n.frame
    pack [text .view$n.text -yscroll ".view$n.vs set" -relief sunken] \
        -side left -fill both -expand 1 -in .view$n.frame
    pack [scrollbar .view$n.vs -command ".view$n.text yview"] \
        -side right -fill y -in .view$n.frame
    pack .view$n.frame -fill both -expand 1
    set TKU(view$n,ro) disabled
    checkbutton .view$n.ro -text "Read Only" -var TKU(view$n,ro) \
	-on disabled -off normal -relief flat \
	-com ".view$n.text conf -state \$TKU(view$n,ro)"
    .view$n.ro select
    pack .view$n.ro -anchor e -padx 5m -pady 2m
    .view$n.text insert 0.0 [read $fp]
    close $fp
    .view$n.text conf -state disabled
    focus .view$n.text
}

proc Tku_ViewFileSave {n} {
    if ![winfo exists .view$n] return
    if {[set file [Tku_FdOpenW]] == ""} return
    puts $file [.view$n.text get 0.0 end]
    close $file
}

#
# Tku_FdOk    - ok button action
#
# action:
#    grabs entry from file entry box.  If it's a directory, calls Tku_FdCwd.
#    Otherwise, if TKU(fd,$w,access) = read, check for readability, if
#    = write, checks for writability and warns the user if the file
#    already exists.
#
proc Tku_FdOk {w} {
    global TKU
    set dir [pwd]
    set file [$w.entry get]
    if {$file == {}} {
        set i [$w.box.listbox curselection]
        case [llength $i] in {
            0    return
            1    {set file [$w.box.listbox get $i]}
            default    return
        }
    }
    if [file isdirectory $file] {
        $w.entry delete 0 end
        Tku_FdCwd $w $file
        return
    }

    if $TKU(fd,$w,SetDir) return

    if {[file dirname $file] != "."} {
        set dir [file dirname $file]
        cd $dir    
        set TKU(fd,$w,directory) $dir
        set file [file tail $file]
    }

    case $TKU(fd,$w,access) in {
    w* {
        if [file exists $file] {
            if ![file writable $file] {Tku_Error "Can't write $file."; return}
            if ![Tku_Confirm "$file exists, overwrite?"] return
        }
        if ![file writable $dir] {Tku_Error "Can't write in directory $dir"; return}
       }
    r* {
        if ![file exists $file] {Tku_Notice "$file does not exists."; return}
        if ![file readable $file] {Tku_Error "Can't read $file."; return}
        }
    }
    set TKU(fd,$w,filename) $file
    destroy $w
}

#
# Tku_FdCancel -    cancel action
#
# action:
#    cancel the file dialog and sets TKU(fd,$w,filename) to {}
#
proc Tku_FdCancel {w} {
    global TKU
    cd $TKU(fd,$w,savecwd)
    set TKU(fd,$w,filename) {}
    set TKU(fd,$w,directory) {}
    destroy $w
}

proc Tku_FdJump {w} {
    if {[set dir [Tku_Entry "Jump to directory" {} Jump]] == ""} return
    Tku_FdCwd $w $dir
}

proc Tku_FdBoxSelect {w} {
    $w.entry delete 0 end
    $w.entry insert 0 [$w.box.listbox get [$w.box.listbox cur]]
    Tku_FdOk $w
}

#
# Tku_Fd -        setup file dialog window
#
# inputs:
#    various args:
#        -file <filename>    default filename
#        -dir <directory>    default directory
#        -select    <pattern>    glob files matching pattern
#        -title <title>        wm title
#        -access <mode>        access mode, read or write
#        -show            show hidden files
#        -setdir            get a directory instead of a filename
#        -open            return an open file descriptor
#                    instead of just a valid filename
#
# action:
#    Constructs listbox, hierarchy menubutton, back-to button,
#    globbing radiobuttons and entry, filename entry, and
#    ok, cancel, home, refresh, and jump-to buttons.  Calls Tku_FdUpdate
#    and waits for $w window to exit before returning a valid
#    filename or open file descriptor.
#
#    Returns {} on cancel.
#
proc Tku_Fd {w args} {
    global TKU

    # set default arguements or restore arguments from last
    # file dialog
    set TKU(fd,$w,savecwd) [pwd]
    if ![info exists TKU(fd,$w,show)] {set TKU(fd,$w,show) 0}
    if ![info exists TKU(fd,$w,all)] {set TKU(fd,$w,all) 1}
    if ![info exists TKU(fd,$w,glob)] {set TKU(fd,$w,glob) {}}
    set file {}
    set select {}
    set access r
    set title "Select a File"
    set dir {}
    set msg {}
	set show 0
	set open 0
	set setdir 0

    # parse arguments
	for {set i 0} {$i < [llength $args]} {incr i} {
		switch -- [lindex $args $i] {
			-file	{set file [lindex $args [incr i]]}
			-dir	{set dir [lindex $args [incr i]]}
			-select	{set select [lindex $args [incr i]]}
			-access	{set access [lindex $args [incr i]]}
			-msg	{set msg [lindex $args [incr i]]}
			-title	{set title [lindex $args [incr i]]}
			-show	{set show 1}
			-open	{set open 1}
			-setdir	{set setdir 1}
			default	{error "strange arg to Tku_Fd: [lindex $args $i]"}
		}
	}

    # update TKU
    set TKU(fd,$w,filename) $file
    if {$dir != {} && [file isdirectory $dir] && [file readable $dir]} {
        set TKU(fd,$w,lastdir) $TKU(fd,$w,savecwd); cd $dir
    } else {
        set dir $TKU(fd,$w,savecwd)
    }
    if {$select != {}} {
        set TKU(fd,$w,glob) $select
        set TKU(fd,$w,all) 0
    }
    set TKU(fd,$w,access) $access
    if $show {set TKU(fd,$w,show) 1}
    set TKU(fd,$w,directory) $dir
    set TKU(fd,$w,SetDir) $setdir

    if [winfo exists $w] {destroy $w}
    toplevel $w
    wm title $w $title
    wm minsize $w 1 1

    # construct right-hand buttons
    frame $w.bottom
    frame $w.b
    if $TKU(fd,$w,SetDir) {
        Tku_Button $w.b.ok {Set Directory} "destroy $w" \
            "Returns the current directory" 1
        set label "Jump To..."
        set files directories
    } else {
        Tku_Button $w.b.ok OK "Tku_FdOk $w" "Selects the file" 1
        set label Filename
        set files files
    }
    Tku_Button $w.b.cancel Cancel "Tku_FdCancel $w"
    Tku_Button $w.b.refresh Update "Tku_FdUpdate $w" "Re-reads current directory"
    Tku_Button $w.b.home Home "Tku_FdCwd $w ~" "Jumps to your home directory"
    Tku_Button $w.b.view View "Tku_FdView $w" "Opens file for viewing"
    set bpack "-pady 1m -padx 5m -anchor w"
    eval pack $w.b.ok $w.b.cancel $w.b.refresh $w.b.home $w.b.view $bpack

    # construct left-hand labels, buttons, and listbox
    set width 40
    set TKU(fd,$w,width) $width

    Tku_Listbox $w.box "Tku_FdBoxSelect $w" -geom ${width}x10
    tk_listboxSingleSelect $w.box.listbox

    # replace the title with a groovy menubutton/directory heirarchy
    frame $w.box.menu_frame -relief flat -bd 0
    label $w.box.menu_title
#    menubutton $w.box.menu -menu $w.box.menu.m -font cursor -text \020 \
#	-relief raised -bd 2
	global MB
    menubutton $w.box.menu -menu $w.box.menu.m -bitmap [Tku_PullDown:Bitmap] \
	-relief raised -bd 2 -padx 2m
    menu $w.box.menu.m
    pack $w.box.menu -side left -in $w.box.menu_frame -fill y
    pack $w.box.menu_title -side left -in $w.box.menu_frame
    pack $w.box.menu_frame -before $w.box.title -anchor w -fill x
    destroy $w.box.title

    frame $w.show_file
    checkbutton $w.show -variable TKU(fd,$w,show) -relief flat \
        -variable TKU(fd,$w,show) -onvalue 1 -offvalue 0 \
        -text "Show hidden $files" -command "Tku_FdUpdate $w"
    label $w.filename -text $label -width $width -anchor w
    entry $w.entry -width $width
    $w.entry insert 0 $TKU(fd,$w,filename)
    pack $w.show $w.filename $w.entry -in $w.show_file -anchor w -padx 2m

    # construct file-pattern matching frame
    frame $w.select
    label $w.select.label -text "Select" -anchor w
    radiobutton $w.select.all -variable TKU(fd,$w,all) -value 1 \
        -text "all $files" -command "Tku_FdUpdate $w" -relief flat
    radiobutton $w.select.match -variable TKU(fd,$w,all) -value 0 \
        -text "$files matching" -command "Tku_FdSelect $w" -relief flat
    entry $w.select.glob -width 10
    $w.select.glob insert 0 $TKU(fd,$w,glob)
    pack $w.select.label -side left -padx 2m
    pack $w.select.all -anchor w
    pack $w.select.match $w.select.glob -side left
    bind $w.select.glob <Return> "Tku_FdSelect $w"

    set tablist [list $w.box.listbox $w.entry $w.select.glob $w.b.ok.b \
		$w.b.cancel.b $w.b.refresh.b $w.b.home.b $w.b.view.b]

    if !$TKU(fd,$w,SetDir) {
        Tku_Button $w.b.jump "Jump To..." "Tku_FdJump $w" \
	    "Jump to a specific directory"
        eval pack $w.b.jump $bpack
        bind $w.entry <Return> "$w.b.ok.b flash;$w.b.ok.b invoke"
        lappend tablist $w.b.jump.b
    } else {
        bind $w.entry <Return> "Tku_FdOk $w"
    }
    if {$msg == {}} {set msg $title}
    pack [Tku_StatusBar $w.title $msg] -fill x
    pack $w.b -side right -fill y
#    label $w.directory -anchor w -width $width -text {Current Directory} \
#	-relief raised -bd 2
#    pack $w.directory -anchor w -fill x
    pack $w.box -expand 1 -fill both
    pack $w.show_file $w.select -fill x -anchor w -ipady 1m
    eval Tku_SetTabbing $tablist

    # update listbox and wait for $w to be destroyed
    Tku_Center $w

    Tku_FdUpdate $w
    Tku_DefFocus $w.box.listbox
    tkwait window $w
    cd $TKU(fd,$w,savecwd)

    # cancel'ed early return
    if {$TKU(fd,$w,directory) == {}} {return {}}

    # directory early return
    if $TKU(fd,$w,SetDir) {
        return $TKU(fd,$w,directory)
    }

    # filename-only early return
    if !$open {return $TKU(fd,$w,directory)/$TKU(fd,$w,filename)}

    # attempt to open file and return file descriptor
    if [catch {set file [open $TKU(fd,$w,directory)/$TKU(fd,$w,filename) $TKU(fd,$w,access)]} err] {
        Tku_Error $err ; return {}
    } else {
        return $file
    }
}

#
# Tku_Fd macros -    simple macros for calling FD
#
# inputs:
#    title    wm title
#    args    additional Tku_Fd args (e.g., -file, -show, etc.)
#
#
# Return a writable filename
proc Tku_FdOpenWFilename {args}  {
    eval Tku_Fd .fdsave -access w -title {{Save File}} $args
}
#
# Return a readable filename
proc Tku_FdReadable {args} {
    eval Tku_Fd .fdread -access r -title {{Open File}} $args
}
#
# Return a writeable open file descriptor
proc Tku_FdOpenW {args}  {
    eval Tku_Fd .fdwrite -access w -open -title {{Save File}} $args
}
#
# Return a readable open file descriptor
proc Tku_FdOpenR {args} {
    eval Tku_Fd .fdopen -access r -open -title {{Open File}} $args
}
#
# Return a directory name
proc Tku_FdDir {args} {
    eval Tku_Fd .fddir -setdir -title {{Set Directory}} $args
}

# ---------------------------------------------------------------------------
# Tku_StatusBar - create a status bar which normally goes below a menu bar
#
# w:	status bar frame
# text:	left hand side title, e.g., "The Spiffy App"
# bitmap: bitmap to put just before text
# ---------------------------------------------------------------------------
proc Tku_StatusBar {w {text "Status"} {bitmap questhead}} {
    global errorInfo errorCode TKU

    frame $w
    pack [label $w.status -anchor e] -in $w -side right -padx 10 -pady 10

    if {$bitmap != ""} {
        pack [label $w.bm -bitmap $bitmap] -side left -padx 10 -pady 10 -in $w
    }
    if {$text != ""} {
        pack [label $w.title -text $text] -side left -padx 10 -pady 10 -in $w
    }
    lappend TKU(statusBars) $w
    bind $w.status <Destroy> "
	set TKU(statusBars) \[ldelete \$TKU(statusBars) $w\]
    "
    return $w
}

# ---------------------------------------------------------------------------
# Tku_Working/Waiting - set or clear a status bar messages - all
#	status bars will post the same message
#
# msg:	status bar frame
# delay: optional delay to pause so message can be seen
# ---------------------------------------------------------------------------
proc Tku_Working {{msg "Working..."} {delay ""}} {
    global TKU
    foreach w $TKU(statusBars) {$w.status configure -text $msg}
	update
    if {$delay != ""} {after $delay}
}
proc Tku_Waiting args {
    global TKU
    foreach w $TKU(statusBars) {$w.status configure -text ""}
	update
}

# ---------------------------------------------------------------------------
# Tku_Pid, Tku_TmpDir, Tku_Tmp - return a unique pid, temp directory,
#	or temp file
# ---------------------------------------------------------------------------
proc Tku_Pid {} {
    # make sure we get a unique number each time
    exec sh -c {echo $$}
}
proc Tku_TmpDir {} {
    global env
    if [info exists env(TMPDIR)] {return $env(TMPDIR)} else {return "/tmp"}
}
proc Tku_Tmp {} {return "[Tku_TmpDir]/tk_tmp.[Tku_Pid]"}

# ---------------------------------------------------------------------------
# Tku_Select - allow the user to select an item from a popup listbox
#
# msg:	select message, e.g., "Select a username"
# entries: select list, e.g., [list jimbo miley terry ...]
# ---------------------------------------------------------------------------
proc Tku_Select {msg entries} {
    global TKU

    if [info exists .select] {error "Tku_Select: already selecting something"}
    toplevel .select
    wm title .select "Select"

    Tku_Listbox .select.box {set TKU(Return) [selection get]}
    Tku_ListboxFill .select.box.listbox $entries
    .select.box.title conf -text $msg

	frame .select.butts
    Tku_Button .select.ok OK {set TKU(Return) [selection get]} \
	{Return the selected line} 1
    Tku_Button .select.cancel Cancel {set TKU(Return) ""}

    pack .select.ok .select.cancel -padx 5m -pady 2m -side left \
		-in .select.butts
	pack .select.butts -side bottom -fill x
    pack .select.box -expand 1 -fill both
    
    Tku_SetTabbing .select.box.listbox .select.ok.b .select.cancel.b
    Tku_Center .select
    focus .select.box.listbox

    tkwait variable TKU(Return)
    destroy .select
    return $TKU(Return)
}

# --------------------------------------------------------------------------
# Tku_DefFocus - change the default focus, setting a binding to restore
#	the previous default when the window is destroyed
# --------------------------------------------------------------------------
proc Tku_DefFocus {w} {
    set def [focus default]
    update
	if ![winfo exists $w] return
    focus default $w
    focus $w
    if {[set top [winfo toplevel $w]] != "." && $def != "none"} {
	bind $top <Destroy> "Tku_DefFocus $def"
    }
}

# --------------------------------------------------------------------------
# Tku_Flash - flash a widget
# --------------------------------------------------------------------------
proc Tku_Flash {w} {
	set bg [lindex [$w conf -bg] 4]
	set fg [lindex [$w conf -fg] 4]
	$w conf -bg $fg; $w conf -fg $bg; update
	after 50 "$w conf -bg $bg; $w conf -fg $fg"
}

# ---------------------------------------------------------------------------
# Tku_Sheet - create and manage spread-sheet like canvases
#
# cmd:	command, one of: create, putcol, insert, delete, etc.
# s:	toplevel sheet frame - passed to all the Sheet: routines
# args:	args for the various commands
# ---------------------------------------------------------------------------
proc Tku_Sheet {cmd s args} {
    global TKU

    set a1 [lindex $args 0]
    set a2 [lindex $args 1]
    case $cmd in {
	create	{Tku_Sheet:Create $s $a1 $a2 [lindex $args 2]}
	putcol {
	    Tku_ListboxFill $s.c.$a1 $a2
	}
	insert {
	    loop n 0 $TKU(sheetCols,$s) {
		$s.c.$n insert $a1 [lindex $a2 $n]
	    }
	    Tku_Sheet renum $s
	}
	clear {
		loop n 0 $TKU(sheetCols,$s) { Tku_ListboxClear $s.c.$n }
		Tku_ListboxClear $s.lines.box
        }
	delete {
	    loop n 0 $TKU(sheetCols,$s) {$s.c.$n delete $a1}
	    Tku_Sheet renum $s
	}
	renum {
	    set max 0
	    loop n 0 $TKU(sheetCols,$s) {
		if {[$s.c.$n size] > $max} {set max [$s.c.$n size]}
	    }
	    Tku_ListboxClear $s.lines.box
	    loop n 0 $max {$s.lines.box insert end $n}
	}
	get {$s.c.[lindex $args 0] get [lindex $args 1]}
	put {$s.c.$a1 insert $a2 [lindex $args 2]}
	replace {$s.c.$a1 delete $a2; $s.c.$a1 insert $a2 [lindex $args 2]}
	getcol {Tku_ListboxDump $s.c.$a1}
	getline {
	    loop n 0 $TKU(sheetCols,$s) {
		lappend res [Tku_Sheet get $s $n $a1]
	    }
	}
	name {return $TKU(sheetName,$s,$a1)}
	index {return $TKU(sheetIndex,$s,$a1)}
	cur* {$s.lines.box curselection}
	save {
    	    if {[set file [Tku_FdOpenWFilename]] == ""} return
    	    Tku_Sheet:Fmt $s $file
	}
	print {Tku_Sheet:Print $s}
	ignore {Tku_Sheet:Ignore $s $a1}
	selectclear {Tku_Sheet:SelClear $s $s.lines.box}
	selectall {
	    $s.lines.box select from 0
	    $s.lines.box select to end
	    Tku_Sheet:SelUpdate $s
	}
	default {error "unknown cmd: $cmd"}
    }
}

# ---------------------------------------------------------------------------
# Tku_Sheet:YSCmd - scroll sheet listboxes together
#
# i:	yview index passed by the scrollbar
# ---------------------------------------------------------------------------
proc Tku_Sheet:Create {s cols lines colinfo} {
    global TKU

    set TKU(sheetCols,$s) $cols
    set TKU(sheetLines,$s) $lines

    set vis 10
    set TKU(sheetVisLines,$s) $vis

    frame $s
    canvas $s.c -xscroll "$s.xs set" -bd 0 -close 5.0 \
	-cursor top_left_arrow
    scrollbar $s.xs -orient horiz -command "$s.c xview"
    scrollbar $s.ys -command "Tku_Sheet:YSCmd $s"

    catch {unset TKU(sheetNames,$s)}

    loop n 0 $cols {

	set info [lindex $colinfo $n]
	if {[set name [lindex $info 0]] == ""} {set name $n}
	if {[set len [lindex $info 1]] == ""} {set len 20}

	set TKU(sheetIgnore,$s,$n) 0
	set TKU(sheetName,$s,$n) $name
	lappend TKU(sheetNames,$s) $name
	set TKU(sheetIndex,$s,$name) $n
	set TKU(sheetSel,$s,$n) 0

	listbox $s.c.$n -geom ${len}x$vis -bd 0 -export 0 \
	    -yscroll "Tku_Sheet:YScroll $s"

	if {[set bind [lindex $info 2]] != ""} {
	    bind $s.c.$n <Double-Button-1> $bind
	}

        bind $s.c.$n <1> "Tku_Sheet:SelClear $s %W; Tku_Sheet:Sel $s %W %y from"
        bind $s.c.$n <B1-Motion> "Tku_Sheet:Sel $s %W %y to"
        bind $s.c.$n <Shift-1> "Tku_Sheet:Sel $s %W %y adjust"
        bind $s.c.$n <Shift-B1-Motion> "Tku_Sheet:Sel $s %W %y to"
    }

    if {[set linewidth [option get . linewidth Linewidth]] == ""} {
	set linewidth 2
    }
    label .xxx 
    set font [lindex [.xxx conf -font] 4]
    set titleheight [expr [winfo reqheight .xxx]+$linewidth]
    destroy .xxx
    set height [expr [winfo reqheight $s.c.0]]
#    set totalwidth [expr $width * $cols * 2]
    set totalheight [expr $height + $titleheight]
    set TKU(sheetHeight,$s) $totalheight

    set totalwidth 0
    loop n 0 $cols {
#	set x [expr $n * $width]
	set x $totalwidth 
    	incr totalwidth [expr [winfo reqwidth $s.c.$n]+$linewidth]
	set lx [expr $x-$linewidth/2]
	$s.c create text [expr $x+$linewidth] 0 -text $TKU(sheetName,$s,$n) \
	    -anchor nw -tags $s.$n -font $font
	$s.c create window $x $titleheight -window $s.c.$n \
	    -anchor nw -tags $s.$n
	if {$n > 0} {
	    $s.c create line $lx 0 $lx $totalheight -width $linewidth \
	    	-tags [list $s.$n vlines]
	}
    }

    set ly [expr $titleheight - $linewidth]
    set canvwid [expr $totalwidth * 1.25]
    set viswid [winfo reqwidth $s.c]
    if {$canvwid < $viswid} {set canvwid $viswid}
    $s.c create line 0 $ly $canvwid $ly -width $linewidth -tags $s.topline
    $s.c conf -scrollregion "0 0 $canvwid $totalheight" -height $totalheight


    canvas $s.lines -cursor sb_right_arrow -bd 0
    listbox $s.lines.box -geom 5x$vis -bd 0 -yscroll "Tku_Sheet:YScroll $s" \
	-export no
    set width [expr [winfo reqwidth $s.lines.box] + $linewidth]
    $s.lines create window 0 $titleheight -window $s.lines.box -anchor nw
    set lpos [expr $width - $linewidth / 2]
    $s.lines create line $lpos 0 $lpos $totalheight \
    	-width $linewidth -tags $s.sideline
    $s.lines create line 0 $ly $width $ly -width $linewidth

    $s.lines conf -width $width -height $totalheight
    loop n 0 $lines {$s.lines.box insert end $n}

    bind $s.lines.box <1> \
	"Tku_Sheet:SelClear $s %W; Tku_Sheet:Sel $s %W %y from"
    bind $s.lines.box <B1-Motion> "Tku_Sheet:Sel $s %W %y to"
    bind $s.lines.box <Shift-1> "Tku_Sheet:Sel $s %W %y adjust"
    bind $s.lines.box <Shift-B1-Motion> "Tku_Sheet:Sel $s %W %y to"

    $s.c bind vlines <1> "Tku_Sheet:StartDrag $s %x"
    $s.c bind vlines <B1-Motion> "Tku_Sheet:Drag $s %x"
    $s.c bind vlines <ButtonRelease-1> "Tku_Sheet:StopDrag $s"
    $s.c bind vlines <Any-Enter> "$s.c conf -cursor right_side"
    $s.c bind vlines <Any-Leave> "$s.c conf -cursor top_left_arrow"

    pack $s.xs -side bottom -fill x
    pack $s.ys -side right -fill y
    pack $s.lines -side left -anchor s -fill y
    pack $s.c -fill both -expand 1

    return $s
}

# ---------------------------------------------------------------------------
# Tku_Sheet:YSCmd - scroll sheet listboxes together
#
# i:	yview index passed by the scrollbar
# ---------------------------------------------------------------------------
proc Tku_Sheet:YSCmd {s i} {
    global TKU
    $s.lines.box yview $i
    loop n 0 $TKU(sheetCols,$s) {$s.c.$n yview $i}
}

# ---------------------------------------------------------------------------
# Tku_Sheet:YScroll - scroll listboxes together and handle vertical resizes
#
# Note: after a sheet canvas resizes, the y scrollbar will get resized and
# execute it's -command procedure (Tku_Sheet:YSCmd) to notify the listboxes
# which will in turn execute their -yview procedure (this procedure) at
# which point we can check for a size change of the canvas.  A bit
# weird but I couldn't find a binding to catch this directly.
#
# args:	args passed by the listboxes
# ---------------------------------------------------------------------------
proc Tku_Sheet:YScroll {s args} {
    global TKU
    if {[winfo height $s.c] != $TKU(sheetHeight,$s)} {Tku_Sheet:YResize $s}
    eval $s.ys set $args
}

# ---------------------------------------------------------------------------
# Tku_Sheet:YResize - resize listboxes to fit the canvas
# ---------------------------------------------------------------------------
proc Tku_Sheet:YResize {s} {
    global TKU

    set totalheight [winfo height $s.c]
    set titleheight [lindex [$s.c coords $s.topline] 0]
    set newheight [expr $totalheight - $titleheight]

    set height [winfo height $s.c.0]
    set g [lindex [$s.c.0 conf -geom] 4]
    set y [lindex [split $g x] 1]
    set sb [lindex [$s.c.0 conf -selectborder] 4]
    set bd [lindex [$s.c.0 conf -border] 4]
    set fluff [expr ($sb + $bd) * 2]
    set avpix [expr ($height - $fluff) / $y]
    
    set newy [expr $newheight / $avpix]
    set adjheight [expr $newy * $avpix + $titleheight]

    loop n 0 $TKU(sheetCols,$s) {
        set g [lindex [$s.c.$n conf -geom] 4]
        set x [lindex [split $g x] 0]
        $s.c.$n conf -geom ${x}x$newy
    }
    set g [lindex [$s.lines.box conf -geom] 4]
    set x [lindex [split $g x] 0]
    $s.lines.box conf -geom ${x}x$newy
    set TKU(sheetHeight,$s) $totalheight
    set sr [lindex [$s.c conf -scrollregion] 4]
    $s.c conf -scrollregion [lreplace $sr 3 3 $totalheight]
    foreach item [$s.c find withtag vlines] {
        eval $s.c coords $item [lreplace [$s.c coords $item] 3 3 $totalheight]
    }
    eval $s.lines coords $s.sideline \
        [lreplace [$s.lines coords $s.sideline] 3 3 $totalheight]
    $s.lines raise $s.sideline
}

# ---------------------------------------------------------------------------
# Tku_Sheet:Drag - start dragging a resize line
#
# x:	x position of mouse
# ---------------------------------------------------------------------------
proc Tku_Sheet:StartDrag {s x} {
    global TKU

    # record the line being dragged and it's start position
    set TKU(sheetBeginX) [set TKU(sheetLastX) [$s.c canvasx $x]]
    set TKU(sheetCur) [$s.c find withtag current]
    # double the line lenght so it's easy to see
    set wide [expr [lindex [$s.c itemconfig $TKU(sheetCur) -width] 4] * 2]
    $s.c itemconfig $TKU(sheetCur) -width $wide
}

# ---------------------------------------------------------------------------
# Tku_Sheet:Drag - continue dragging a resize line
#
# x:	x position of mouse
# ---------------------------------------------------------------------------
proc Tku_Sheet:Drag {s x} {
    global TKU

    # move line relative to start position
    set x [$s.c canvasx $x]
    $s.c move current [expr $x-$TKU(sheetLastX)] 0
    set TKU(sheetLastX) $x
}

# ---------------------------------------------------------------------------
# Tku_Sheet:StopDrag - stop dragging a resize line
# ---------------------------------------------------------------------------
proc Tku_Sheet:StopDrag {s} {
    global TKU

    # restore line width and calculate final movement
    set item $TKU(sheetCur)
    set tag [lfirst [lindex [$s.c itemconf $item -tags] 4]]
    set notwide [expr [lindex [$s.c itemconfig $item -width] 4] / 2]
    $s.c itemconfig $item -width $notwide
    set change [expr $TKU(sheetLastX) - $TKU(sheetBeginX)]
    if {$change == 0} return

    # get new width of the corresponding listbox
    set n [llast [split $tag .]]
    set lb $s.c.[expr $n - 1]
    set g [split [lindex [$lb conf -geom] 4] x]
    set gx [lindex $g 0]
    set gy [lindex $g 1]
    set sb [lindex [$lb conf -selectborder] 4]
    set bd [lindex [$lb conf -border] 4]
    set fluff [expr ($sb + $bd) * 2]
    set wid [winfo width $lb]
    set neww [expr $wid + $change - $fluff]

    # adjust listbox width to closest integral number of average
    # character widths, but don't allow the listbox to get smaller
    # then it's title
    set avpix [expr ($wid - $fluff) / $gx]
    set newgx [expr $neww / $avpix]
    set min [expr [string length $TKU(sheetName,$s,[expr $n-1])]+1]
    if {$newgx < $min} {set newgx $min}
    $lb conf -geom ${newgx}x$gy

    # adjust line position to actual listbox movement
    set actualw [winfo width $lb]
    set actualchange [expr $actualw - $wid]
    $s.c move $item [expr -1 * $change] 0

    # move all other listboxes and lines over
    loop n $n $TKU(sheetCols,$s) {$s.c move $s.$n $actualchange 0}
}

# ---------------------------------------------------------------------------
# Tku_Sheet:SelClear - clear the selection from the listboxes
#
# w:	listbox the mouse is selecting - all other will be cleared
# ---------------------------------------------------------------------------
proc Tku_Sheet:SelClear {s w} {
    global TKU

    if {$w == "$s.lines.box"} {set col -1} else {set col [winfo name $w]}
    $s.lines.box select clear
    loop n 0 $TKU(sheetCols,$s) {if {$n != $col} {set TKU(sheetSel,$s,$n) 0}}
    Tku_Sheet:SelUpdate $s
}

# ---------------------------------------------------------------------------
# Tku_Sheet:Sel - manage selection as the mouse moves around the listboxes
#
# w:	listbox the mouse is moving over
# y:	y position in listbox
# what: one of: "to" or "from" to start or extend selection
# ---------------------------------------------------------------------------
proc Tku_Sheet:Sel {s w y what} {
    global TKU

    # make sure the column the mouse is moving over is selected
    if {$w != "$s.lines.box"} {
	set col [winfo name $w]
    	if !$TKU(sheetSel,$s,$col) {
    	    set TKU(sheetSel,$s,$col) 1
	    Tku_Sheet:SelUpdate $s
	}
    }

    # now find the nearest element and update the selection in all boxes
    set near [$w nearest $y]
    $s.lines.box select $what $near
    loop n 0 $TKU(sheetCols,$s) {
	if $TKU(sheetSel,$s,$n) {$s.c.$n select $what $near}
    }
}

# ---------------------------------------------------------------------------
# Tku_Sheet:SelUpdate - update the selection if a column is selected
# ---------------------------------------------------------------------------
proc Tku_Sheet:SelUpdate {s} {
    global TKU
    if {[set selected [Tku_Sheet cur $s]] != ""} {
	set clear 0
    	set first [lfirst $selected]
    	set last [llast $selected]
    } else {
	set clear 1
    }
    loop n 0 $TKU(sheetCols,$s) {
	if {$TKU(sheetSel,$s,$n) && !$clear} {
	    $s.c.$n select from $first ; $s.c.$n select to $last
	} else {
	    $s.c.$n select clear
	}
    }
}


# ---------------------------------------------------------------------------
# Tku_Sheet:Print - print the selected columns
# ---------------------------------------------------------------------------
proc Tku_Sheet:Print {s} {
    global TKU

    set deflpr "enscript -rG"
    if ![info exists TKU(option,lpr)] {set TKU(option,lpr) $deflpr}
    if {$TKU(option,lpr) == ""} {set TKU(option,lpr) $deflpr}

    set TKU(option,lpr) [Tku_Entry "Print command" $TKU(option,lpr)]
    if {$TKU(option,lpr) == ""} return

    set file [Tku_Tmp]
    if ![Tku_Sheet:Fmt $s $file] return
    if [catch {eval exec $TKU(option,lpr) < $file} status] {
    	Tku_Notice "Print notice:\n$status"
    }
    exec rm -f $file
}

# ---------------------------------------------------------------------------
# Tku_Sheet:Fmt - dump the contents of the selected columns to a file
#
# file:	output file
# ---------------------------------------------------------------------------
proc Tku_Sheet:Fmt {s file} {
    global TKU

    # if no lines are selected, get all of them
    if {[set sel [Tku_Sheet cur $s]] == ""} {
		set sel [Tku_ListboxDump $s.lines.box]
    }

    # get selected columns - if no columns are selected, get all of them
	# except those marked to be ignored
	set all 1
    loop n 0 $TKU(sheetCols,$s) {
   		if $TKU(sheetSel,$s,$n) {set all 0; break}
	}
    loop n 0 $TKU(sheetCols,$s) {
   		if {$TKU(sheetSel,$s,$n)} {set col($n) 1} \
		elseif {$all &&  !$TKU(sheetIgnore,$s,$n)} {set col($n) 1} \
		else {set col($n) 0}
	}

    loop n 0 $TKU(sheetCols,$s) {
	if $col($n) {
	    set len($n) 0
	    foreach e $sel {
	        set length [string length [$s.c.$n get $e]]
	        if {$length > $len($n)} {set len($n) $length}
		}
	}
    }
    if ![info exists len] {
		Tku_Notice "Nothing to save or print."
		return 0
    }
    if [catch {set fp [open "$file" w]}] {
	Tku_Error "Can't open output file: $file."
	return 0
    }
    set line "  row"
    set sep   "     "
    foreach n [lsort [array names len]] {
	set namelen [string length $TKU(sheetName,$s,$n)]
	if {$namelen > $len($n)} {set len($n) $namelen}
	append line [format " | %-$len($n)s" $TKU(sheetName,$s,$n)]
	append sep  [format " + %-$len($n)s" " "]
    }
    regsub -all { } $sep {-} sep
    puts $fp "$line\n$sep"
    foreach e $sel {
	set line [format " %-4d" $e]
	foreach n [lsort [array names len]] {
	    append line [format " | %-$len($n)s" [$s.c.$n get $e]]
        }
	puts $fp $line
    }
    close $fp
    return 1
}

proc Tku_Sheet:Ignore {s name} {
	global TKU
	if {[set n [Tku_Sheet:ColNum $s $name 0 0]] == -1} {return 0}
	set TKU(sheetIgnore,$s,$n) 1
	return 1
}
# --------------------------------------------------------------------------
# Tku_Sheet:ColNum - get the query results column number of a desired field
#
# s	sheet
# name	name of desired field
# altok	if 1, prompt the user for an alternate field if
#	desired field doesn't exist in the query results
# --------------------------------------------------------------------------
proc Tku_Sheet:ColNum {s name {altok 1} {confirm 1}} {
    global TKU

    set none -1
    if ![winfo exists $s.c.0] {return $none}

    # search for the desired field name
    loop n 0 $TKU(sheetCols,$s) {
		if {$TKU(sheetName,$s,$n) == $name} {
			if !$confirm {return $n}
			if [Tku_Confirm "Use the $name column?"] {return $n}
		}
	}

    # if the desired field wasn't found and we can't use an alternate
    # or the user doesn't want to select one, return none
    if !$altok {return $none}
    if ![Tku_Confirm "This action requires a field called \"$name\" - would you like to select an alternate?"] {return $none}

    set name [Tku_Select "Alternate field:" $TKU(sheetNames,$s)]
    if {$name == ""} {
	return $none
    } else {
    	Tku_Sheet:SelUpdate $s
    	return $TKU(sheetIndex,$s,$name)
    }
}

# --------------------------------------------------------------------------
# Tku_Sheet:GetCol - get the contents of a query results window by number
#
# n	column number
# allok	if 1, it's o.k. to return all elements if none are selected
# --------------------------------------------------------------------------
proc Tku_Sheet:GetCol {s n {allok 1}} {
    global TKU

    if ![winfo exists $s.c.0] {return ""}

    # if nothings selected, ask the user if it's o.k. to return everything
	set TKU(sheetSel,$s,$n) 1
	Tku_Sheet:SelUpdate $s
    if {[$s.c.$n curselection] == "" && $allok} {
        set name $TKU(sheetName,$s,$n)
        if [Tku_Confirm "You haven't selected any lines from the $name column - do you want to select all of them?"] {
	    Tku_Sheet selectall $s
	 }
    }
    foreach i [$s.c.$n curselection] {lappend res [Tku_Sheet get $s $n $i]}
    return $res
}

# --------------------------------------------------------------------------
# Tku_Sheet:ColNames - return the column names of a sheet
# --------------------------------------------------------------------------
proc Tku_Sheet:ColNames {s} {global TKU ; return $TKU(sheetNames,$s)}


# --------------------------------------------------------------------------
# Set some one-time options and bindings (which means this file should
# probably be source'd before any windows are created just like tk.tcl)
# --------------------------------------------------------------------------
option add *Cursor top_left_arrow
option add *Button.borderWidth 2
option add *Button.padX 4
option add *Button.padY 2
option add *Frame.borderWidth 2
option add *Frame.relief raised
option add *Entry.relief sunken
option add *Scrollbar.relief sunken
option add *Listbox.relief sunken

bind Text <Up> {%W mark set insert "insert -1l"; %W yview -pickplace insert}
bind Text <Down> {%W mark set insert "insert +1l"; %W yview -pickplace insert}
bind Text <Control-p> [bind Text <Up>]
bind Text <Control-n> [bind Text <Down>]
bind Text <Left> {%W mark set insert "insert -1c"; %W yview -pickplace insert}
bind Text <Right> {%W mark set insert "insert +1c"; %W yview -pickplace insert}
bind Text <Control-b> [bind Text <Left>]
bind Text <Control-f> [bind Text <Right>]
bind Text <Control-a> {%W mark set insert "insert linestart"; %W yview -pickplace insert}
bind Text <Control-e> {%W mark set insert "insert lineend"; %W yview -pickplace insert}
bind Text <Control-u> {%W delete "insert linestart" insert; %W yview -pickplace insert}
bind Text <Escape> {Tku_Flash %W}

bind Entry <Control-a> {%W icursor 0}
bind Entry <Control-e> {%W icursor end}
bind Entry <Left> {%W icursor [expr [%W index insert]-1]}
bind Entry <Right> {%W icursor [expr [%W index insert]+1]}
bind Entry <Control-b> [bind Entry <Left>]
bind Entry <Control-f> [bind Entry <Right>]
bind Entry <Escape> {Tku_Flash %W}

bind Listbox <Any-Key> {Tku_Listbox:MoveTo %W %A}
bind Listbox <Up> {Tku_Listbox:MoveUp %W}
bind Listbox <Down> {Tku_Listbox:MoveDown %W}
bind Listbox <Control-p> [bind Listbox <Up>]
bind Listbox <Control-n> [bind Listbox <Down>]
bind Listbox <Shift-Up> {Tku_Listbox:MoveTop %W}
bind Listbox <Shift-Down> {Tku_Listbox:MoveBottom %W}
bind Listbox <Escape> {set TKU(inListbox) {}}

foreach TKU(class) {Listbox Entry Button Checkbutton Radiobutton Text} {
    tk_bindForTraversal $TKU(class)
    bind $TKU(class) <Any-Control-KeyPress> {Tku_MenuBar:Accel C %W %K}
    bind $TKU(class) <Any-Meta-KeyPress> {Tku_MenuBar:Accel M %W %K}
}; unset TKU(class)


bind Menu <Right> {
    set m %W    
    set i [$m index active]
    if ![catch {set c [lindex [$m entryconfig $i -menu] 4]}] {
    	if {$c != {}} {
	     $c activate 0
	     focus $c
	     return
	}
    }
    $m entryconfig $i -state normal
    tk_nextMenu 1
}

bind Menu <Left> {
    if {[winfo class [winfo parent %W]] == "Menu"} {
	%W entryconfig [%W index active] -state normal
	focus [winfo parent %W]
    } else {
	tk_nextMenu -1
    }
}

bind Menu <Up> {Tku_Menu:NextEntry %W -1}
bind Menu <Down> {Tku_Menu:NextEntry %W 1}

proc Tku_Menu:NextEntry {menu count} {
    if {[$menu index last] == "none"} {
	return
    }
    set length [expr [$menu index last]+1]
    set i [$menu index active]
    if {$i == "none"} {
	set i 0
    } else {
	incr i $count
    }
    while 1 {
	while {$i < 0} {
	    incr i $length
	}
	while {$i >= $length} {
	    incr i -$length
	}
	if {[catch {$menu entryconfigure $i -state} state] == 0} {
	    if {[lindex $state 4] != "disabled"} {
		break
	    }
	}
	incr i $count
    }
    $menu activate $i
}

set TKU(popup,enabled) 1
set TKU(popup,entry) ""
bind all <Destroy> +Tku_Popup:UnPost
bind all <Any-Leave> +Tku_Popup:UnPost
bind Menu <Any-Enter> {+Tku_Popup:Menu %W %y}
bind Menu <Any-Motion> {+Tku_Popup:Menu %W %y}
bind Menu <Unmap> [bind all <Any-Leave>]
bind Entry <Any-Enter> {Tku_Popup:Edge %W %X %Y}
bind Button <Any-Enter> {+Tku_Popup:Edge %W %X %Y}
bind Checkbutton <Any-Enter> {+Tku_Popup:Edge %W %X %Y}
bind Radiobutton <Any-Enter> {+Tku_Popup:Edge %W %X %Y}
bind MenuButton <Any-Enter> {+Tku_Popup:Edge %W %X %Y}
