#include "MiTcl.h"

#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>


static int
EnterConn(interp, conn)
	Tcl_Interp	*interp;
	MI_CONNECTION	*conn;
{
	int fd;
	FILE *fp;

	fd = mi_unix_connection(conn);
	if (fd == MI_ERROR) {
		interp->result = "could not get unix connection";
		return TCL_ERROR;
	}
	fp = fdopen(fd, "r+");
	if (fp == NULL) {
		interp->result = "could not fdopen unix connection";
		return TCL_ERROR;
	}
	setbuf(fp, NULL);
	Tcl_EnterFile(interp, fp, TCL_FILE_READABLE | TCL_FILE_WRITABLE);
	return TCL_OK;
}


int
Tmi_get_default_connection_info(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_CONNECTION_INFO cinfo;
    char            port[20];

    if (mi_get_default_connection_info(&cinfo) == MI_ERROR) {
	interp->result = "could not get default connection info";
	return TCL_ERROR;
    }
    sprintf(port, "%d", cinfo.server_port);
    Tcl_AppendElement(interp, cinfo.server_name);
    Tcl_AppendElement(interp, cinfo.server_host);
    Tcl_AppendElement(interp, port);
    return TCL_OK;
}


int
Tmi_set_default_connection_info(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_CONNECTION_INFO cinfo;

    cinfo.server_name = (char *) NULL;
    cinfo.server_host = (char *) NULL;
    cinfo.server_port = 0;

    switch (argc) {
    case 3:
	if (argv[1][0] != '\0') {
	    cinfo.server_host = argv[1];
	}
	if (Tcl_GetInt(interp, argv[2], &(cinfo.server_port)) != TCL_OK) {
	    return TCL_ERROR;
	}
	break;
    case 2:
	if (argv[1][0] != '\0') {
	    cinfo.server_name = argv[1];
	}
	break;
    default:
	Tcl_AppendResult(interp, "wrong # args:  should be \"",
			 argv[0], " {server | host port}\"", (char *) NULL);
	return TCL_ERROR;
    }

    if (mi_set_default_connection_info(&cinfo) == MI_ERROR) {
	interp->result = "could not set default connction info";
	return TCL_ERROR;
    }
    return TCL_OK;
}


int
Tmi_server_connect(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_CONNECTION  *conn;
    MI_CONNECTION_INFO cinfo;

    if (argc != 1) {
	if (Tmi_set_default_connection_info(dummy, interp, argc, argv)
	    != TCL_OK) {
	    return TCL_ERROR;
	}
    }
    if (mi_get_default_connection_info(&cinfo) == MI_ERROR) {
	interp->result = "could not get default connection info";
	return TCL_ERROR;
    }
    conn = mi_server_connect(&cinfo);
    if (conn == (MI_CONNECTION *) NULL) {
	interp->result = "could not connect to server";
	return TCL_ERROR;
    }
    MiSetId(interp->result, (void *) conn);
    return TCL_OK;
}


int
Tmi_get_connection_info(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_CONNECTION_INFO cinfo;
    MI_CONNECTION  *conn;
    char            port[20];

    if (MiArgv1Conn(interp, argc, argv, &conn) != TCL_OK) {
	return TCL_ERROR;
    }
    if (mi_get_connection_info(conn, &cinfo) == MI_ERROR) {
	interp->result = "couldn't get connection info";
	return TCL_ERROR;
    }
    sprintf(port, "%d", cinfo.server_port);
    Tcl_AppendElement(interp, cinfo.server_name);
    Tcl_AppendElement(interp, cinfo.server_host);
    Tcl_AppendElement(interp, port);
    return TCL_OK;
}


int
Tmi_get_default_database_info(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_DATABASE_INFO dbinfo;

    if (mi_get_default_database_info(&dbinfo) == MI_ERROR) {
	interp->result = "could not get default database info";
	return TCL_ERROR;
    }
    Tcl_AppendElement(interp, dbinfo.database_name);
    Tcl_AppendElement(interp, dbinfo.user_name);
    Tcl_AppendElement(interp, dbinfo.password);
    return TCL_OK;
}


int
Tmi_set_default_database_info(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_DATABASE_INFO dbinfo;

    if (mi_get_default_database_info(&dbinfo) == MI_ERROR) {
	interp->result = "could not get default database info!";
	return TCL_ERROR;
    }
    switch (argc) {
    case 4:
	if (argv[3][0] != '\0') {
	    dbinfo.password = argv[3];
	}
    case 3:
	if (argv[2][0] != '\0') {
	    dbinfo.user_name = argv[2];
	}
    case 2:
	if (argv[1][0] != '\0') {
	    dbinfo.database_name = argv[1];
	}
	break;
    default:
	Tcl_AppendResult(interp, "wrong # args:  should be \"",
			 argv[0], " db [user] [pass]\"", (char *) NULL);
	return TCL_ERROR;
    }

    if (mi_set_default_database_info(&dbinfo) == MI_ERROR) {
	interp->result = "could not set default database info";
	return TCL_ERROR;
    }
    return TCL_OK;
}


int
Tmi_login(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_CONNECTION  *conn;
    MI_DATABASE_INFO dbinfo;

    if (mi_get_default_database_info(&dbinfo) == MI_ERROR) {
	interp->result = "could not get default database info!";
	return TCL_ERROR;
    }
    switch (argc) {
    case 5:
	if (argv[4][0] != '\0') {
	    dbinfo.password = argv[4];
	}
    case 4:
	if (argv[3][0] != '\0') {
	    dbinfo.user_name = argv[3];
	}
    case 3:
	if (argv[2][0] != '\0') {
	    dbinfo.database_name = argv[2];
	}
    case 2:
	if (MiGetConn(interp, argv[1], &conn) != TCL_OK) {
	    return TCL_ERROR;
	}
	break;
    default:
	Tcl_AppendResult(interp, "wrong # args:  should be \"",
		argv[0], " connId [db] [user] [password]\"", (char *) NULL);
	return TCL_ERROR;
    }
    if (mi_login(conn, &dbinfo) == MI_ERROR) {
	interp->result = "could not login";
	return TCL_ERROR;
    }
    if (EnterConn(interp, conn) != TCL_OK) {
	return TCL_ERROR;
    }
    return TCL_OK;
}


int
Tmi_get_database_info(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_DATABASE_INFO dbinfo;
    MI_CONNECTION  *conn;

    if (MiArgv1Conn(interp, argc, argv, &conn) != TCL_OK) {
	return TCL_ERROR;
    }
    if (mi_get_database_info(conn, &dbinfo) == MI_ERROR) {
	interp->result = "couldn't get database info";
	return TCL_ERROR;
    }
    Tcl_AppendElement(interp, dbinfo.database_name);
    Tcl_AppendElement(interp, dbinfo.user_name);
    Tcl_AppendElement(interp, dbinfo.password);
    return TCL_OK;
}


int
Tmi_open(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    char           *db = (char *) NULL;
    char           *user = (char *) NULL;
    char           *pass = (char *) NULL;
    MI_CONNECTION  *conn;

    switch (argc) {
    case 4:
	if (argv[3][0]) {
	    pass = argv[3];
	}
    case 3:
	if (argv[2][0]) {
	    user = argv[2];
	}
    case 2:
	if (argv[1][0]) {
	    db = argv[1];
	}
    case 1:
	break;
    default:
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " [db] [user] [pass]\"", (char *) NULL);
	return TCL_ERROR;
    }

    if ((conn = mi_open(db, user, pass)) == (MI_CONNECTION *) NULL) {
	interp->result = "could not open database";
	return TCL_ERROR;
    }
    if (EnterConn(interp, conn) != TCL_OK) {
	return TCL_ERROR;
    }
    MiSetId(interp->result, (void *) conn);
    return TCL_OK;
}


int
Tmi_close(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_CONNECTION  *conn;
    char            closeCmd[0x100];

    if (MiArgv1Conn(interp, argc, argv, &conn) != TCL_OK) {
	return TCL_ERROR;
    }

    /* Create the Tcl close command. */
    sprintf(closeCmd, "close file%d", mi_unix_connection(conn));

    /* Close the connection. */
    if (mi_close(conn) == MI_ERROR) {
	interp->result = "couldn't close connection";
	return TCL_ERROR;
    }

    /* Evaluate the close command, ignoring the return value. */
    (void) Tcl_Eval(interp, closeCmd);
    Tcl_ResetResult(interp);
    return TCL_OK;
}
