#include "MiTcl.h"


static int
MiGetEventCode(interp, string, event)
    Tcl_Interp     *interp;
    char           *string;
    int            *event;
{
    if (STREQ(string, "All_Events")) {
	*event = MI_All_Events;
    }
    else if (STREQ(string, "Exception")) {
	*event = MI_Exception;
    }
    else if (STREQ(string, "Alerter_Fire_Msg")) {
	*event = MI_Alerter_Fire_Msg;
    }
    else if (STREQ(string, "Delivery_Status_Msg")) {
	*event = MI_Delivery_Status_Msg;
    }
    else if (STREQ(string, "Query_Interrupt_Ack")) {
	*event = MI_Query_Interrupt_Ack;
    }
    else if (STREQ(string, "Client_Library_Error")) {
	*event = MI_Client_Library_Error;
    }
    else if (STREQ(string, "Xact_State_Change")) {
	*event = MI_Xact_State_Change;
    }
    else if (STREQ(string, "Print")) {
	*event = MI_Print;
    }
    else if (STREQ(string, "Request")) {
	*event = MI_Request;
    }
    else if (STREQ(string, "Tape_Request")) {
	*event = MI_Tape_Request;
    }
    else {
	Tcl_AppendResult(interp, "unknown event \"",
			 string, "\":  should be ",
			 "All_Events, Exception, Alerter_Fire_Msg, ",
			 "Delivery_Status_Msg, Query_Interrupt_Ack, ",
			 "Client_Library_Error, Xact_State_Change, ",
			 "Print, Request, or Tape_Request", (char *) NULL);
	return TCL_ERROR;
    }
    return TCL_OK;
}


static void
MiMarkEvent(event, conn, eventData, userData)
    MI_EVENT_TYPE   event;
    MI_CONNECTION  *conn;
    void           *eventData;
    void           *userData;
{
    MiEvent        *d = (MiEvent *) userData;
    char            nalerts[50];
    char            buf[512];

    /* Set the connection id */
    MiSetId(d->connId, (void *) conn);

    /*
     * set the eventname and build up the event data keyed list based on the
     * type of event
     */
    Tcl_DStringInit(d->eventData);
    switch (event) {

    case MI_Alerter_Fire_Msg:
	strcpy(d->eventName, "Alerter_Fire_Msg");

	/* set the event name */
	Tcl_DStringStartSublist(d->eventData);
	Tcl_DStringAppendElement(d->eventData, "name");
	Tcl_DStringAppendElement(d->eventData, mi_alert_name(eventData));
	Tcl_DStringEndSublist(d->eventData);

	/* set the number of alerts */
	Tcl_DStringStartSublist(d->eventData);
	sprintf(nalerts, "%d", mi_alert_nalerts(eventData));
	Tcl_DStringAppendElement(d->eventData, "nalerts");
	Tcl_DStringAppendElement(d->eventData, nalerts);
	Tcl_DStringEndSublist(d->eventData);

	/* set the status */
	Tcl_DStringStartSublist(d->eventData);
	Tcl_DStringAppendElement(d->eventData, "status");
	switch (mi_alert_status(eventData)) {
	case MI_ALERTER_FIRED:
	    Tcl_DStringAppendElement(d->eventData, "FIRED");
	    break;
	case MI_ALERTER_DROPPED:
	    Tcl_DStringAppendElement(d->eventData, "DROPPED");
	    break;
	default:
	    Tcl_DStringAppendElement(d->eventData, "unknown?");
	    break;
	}
	Tcl_DStringEndSublist(d->eventData);

	break;

    case MI_Delivery_Status_Msg:
	strcpy(d->eventName, "Delivery_Status_Msg");
	break;

    case MI_Exception:
    case MI_Client_Library_Error:
	if (event == MI_Exception) {
	    strcpy(d->eventName, "Exception");
	}
	else {
	    strcpy(d->eventName, "Client_Library_Error");
	}

	/* record the error level */
	Tcl_DStringStartSublist(d->eventData);
	Tcl_DStringAppendElement(d->eventData, "error_level");
	switch (mi_error_level(eventData)) {
	case MI_MESSAGE:
	    Tcl_DStringAppendElement(d->eventData, "MESSAGE");
	    break;
	case MI_EXCEPTION:
	    Tcl_DStringAppendElement(d->eventData, "EXCEPTION");
	    break;
	case MI_FATAL:
	    Tcl_DStringAppendElement(d->eventData, "FATAL");
	    break;
	case MI_LIB_BADARG:
	    Tcl_DStringAppendElement(d->eventData, "LIB_BADARG");
	    break;
	case MI_LIB_USAGE:
	    Tcl_DStringAppendElement(d->eventData, "LIB_USAGE");
	    break;
	case MI_LIB_INTERR:
	    Tcl_DStringAppendElement(d->eventData, "LIB_INTERR");
	    break;
	case MI_LIB_NOIMP:
	    Tcl_DStringAppendElement(d->eventData, "LIB_NOIMP");
	    break;
	case MI_LIB_DROPCONN:
	    Tcl_DStringAppendElement(d->eventData, "LIB_DROPCONN");
	    break;
	default:
	    Tcl_DStringAppendElement(d->eventData, "UNKNOWN");
	    break;
	}
	Tcl_DStringEndSublist(d->eventData);

	/* record the SQL error code */
	Tcl_DStringStartSublist(d->eventData);
	Tcl_DStringAppendElement(d->eventData, "error_sql_code");
	mi_errmsg(eventData, buf, TCL_DSTRING_STATIC_SIZE);
	Tcl_DStringAppendElement(d->eventData, buf);
	Tcl_DStringEndSublist(d->eventData);

	/* record the error message */
	Tcl_DStringStartSublist(d->eventData);
	Tcl_DStringAppendElement(d->eventData, "errmsg");
	mi_errmsg(eventData, buf, TCL_DSTRING_STATIC_SIZE -
		  Tcl_DStringLength(d->eventData));
	Tcl_DStringAppendElement(d->eventData, buf);
	Tcl_DStringEndSublist(d->eventData);

	break;

    case MI_Print:
	strcpy(d->eventName, "Print");
	break;

    case MI_Query_Interrupt_Ack:
	strcpy(d->eventName, "Query_Interrupt_Ack");
	break;

    case MI_Request:
	strcpy(d->eventName, "Request");
	break;

    case MI_Tape_Request:
	strcpy(d->eventName, "Tape_Request");
	break;

    case MI_Xact_State_Change:
	strcpy(d->eventName, "Xact_State_Change");

	/* setup the Xact_State_Change event data keyed list */
	Tcl_DStringStartSublist(d->eventData);
	Tcl_DStringAppendElement(d->eventData, "state");
	switch (mi_xact_state(eventData)) {
	case MI_XACT_BEGIN:
	    Tcl_DStringAppendElement(d->eventData, "BEGIN");
	    break;
	case MI_XACT_END:
	    Tcl_DStringAppendElement(d->eventData, "END");
	    break;
	case MI_XACT_ABORT:
	    Tcl_DStringAppendElement(d->eventData, "ABORT");
	    break;
	}
	Tcl_DStringEndSublist(d->eventData);

	break;
    case MI_All_Events:
    default:
	fprintf(stderr,
	    "Impossible event:  %d (%s line %d)", event, __FILE__, __LINE__);
	exit(1);
	/* NOTREACHED */
	break;
    }


    /* mark the procedure to be run */
    Tcl_AsyncMark(d->asyncHandler);
}


static int
MiDoEvent(clientData, interp, code)
    ClientData      clientData;
    Tcl_Interp     *interp;
    int             code;
{
    Tcl_DString     proc;
    Tcl_DString     result;
    Tcl_DString     eInfo;
    Tcl_DString     eCode;
    char           *p;
    MiEvent        *d = (MiEvent *) clientData;

    /* save the state of the interpreter */
    Tcl_DStringInit(&result);
    Tcl_DStringAppend(&result, interp->result, -1);

    Tcl_DStringInit(&eCode);
    p = Tcl_GetVar(interp, "errorCode", TCL_GLOBAL_ONLY);
    if (p != (char *) NULL) {
	Tcl_DStringAppend(&eCode, p, -1);
    }
    Tcl_DStringInit(&eInfo);
    p = Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
    if (p != (char *) NULL) {
	Tcl_DStringAppend(&eInfo, p, -1);
    }
    /*
     * create the string to evaluate, appending the connId, eventName, and
     * eventData as the last argument
     */
    Tcl_DStringInit(&proc);
    Tcl_DStringAppend(&proc, d->proc, -1);
    Tcl_DStringAppendElement(&proc, d->connId);
    Tcl_DStringAppendElement(&proc, d->eventName);
    Tcl_DStringAppendElement(&proc, d->eventData->string);

    /* evalute the procedure */
    if (Tcl_GlobalEval(interp, proc.string) != TCL_OK) {
	fprintf(stderr, "Error handling callback:\n%s\n",
		Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY));
    }
    Tcl_DStringFree(&proc);

    /* restore the interpreter */
    if (Tcl_DStringLength(&eCode) > 0) {
	(void) Tcl_SetVar(interp, "errorCode", eCode.string, TCL_GLOBAL_ONLY);
    }
    else {
	(void) Tcl_UnsetVar(interp, "errorCode", TCL_GLOBAL_ONLY);
    }
    Tcl_DStringFree(&eCode);

    if (Tcl_DStringLength(&eInfo) > 0) {
	(void) Tcl_SetVar(interp, "errorInfo", eInfo.string, TCL_GLOBAL_ONLY);
    }
    else {
	(void) Tcl_UnsetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
    }
    Tcl_DStringFree(&eInfo);

    Tcl_DStringResult(interp, &result);

    return code;
}


int
Tmi_add_callback(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_EVENT_TYPE   event;
    MiEvent        *d;

    if (argc != 3) {
	Tcl_AppendResult(interp, "wrong # args:  should be \"",
			 argv[0], " event proc\"", (char *) NULL);
	return TCL_ERROR;
    }
    if (MiGetEventCode(interp, argv[1], &event) != TCL_OK) {
	return TCL_ERROR;
    }
    d = (MiEvent *) mi_alloc(sizeof(MiEvent));
    d->eventData = (Tcl_DString *) mi_alloc(sizeof(Tcl_DString));
    d->asyncHandler = Tcl_AsyncCreate(MiDoEvent, (ClientData) d);
    d->proc = (char *) mi_alloc(strlen(argv[2]) + 1);
    strcpy(d->proc, argv[2]);
    if (mi_add_callback(event, (MI_VOID) MiMarkEvent, (void *) d) ==
	MI_ERROR) {
	interp->result = "could not add callback";
	return TCL_ERROR;
    }
    return TCL_OK;
}


int
Tmi_default_callback(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_EVENT_TYPE   event;
    int             connect = 0;

    if ((argc != 2) && (argc != 3)) {
	Tcl_AppendResult(interp, "wrong # args:  should be \"",
			 argv[0], " event [connect]\"", (char *) NULL);
	return TCL_ERROR;
    }
    if (argc == 3) {
	if (Tcl_GetInt(interp, argv[2], &connect) != TCL_OK) {
	    return TCL_ERROR;
	}
    }
    if (MiGetEventCode(interp, argv[1], &event) != TCL_OK) {
	return TCL_ERROR;
    }
    if (mi_add_callback(event, mi_default_callback, (void *) &connect) ==
	MI_ERROR) {
	interp->result = "could not set default handler";
	return TCL_ERROR;
    }
    return TCL_OK;
}


int
Tmi_disable_callbacks(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    if (mi_disable_callbacks() == MI_ERROR) {
	return TCL_ERROR;
    }
    return TCL_OK;
}


int
Tmi_enable_callbacks(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    if (mi_enable_callbacks() == MI_ERROR) {
	return TCL_ERROR;
    }
    return TCL_OK;
}


int
Tmi_delete_callback(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_EVENT_TYPE   event;

    if (argc != 2) {
	Tcl_AppendResult(interp, "wrong # args:  should be \"",
			 argv[0], " event\"", (char *) NULL);
	return TCL_ERROR;
    }
    if (MiGetEventCode(interp, argv[1], &event) != TCL_OK) {
	return TCL_ERROR;
    }
    if (mi_delete_callback(event) == MI_ERROR) {
	interp->result = "could not delete callback";
	return TCL_ERROR;
    }
    return TCL_OK;
}


int
Tmi_processing_needed(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_CONNECTION  *conn;

    if (MiArgv1Conn(interp, argc, argv, &conn) != TCL_OK) {
	return TCL_ERROR;
    }
    switch (mi_processing_needed(conn, 0)) {
    case MI_AS_NONE:
	interp->result = "NONE";
	break;
    case MI_AS_RESULT:
	interp->result = "RESULT";
	break;
    case MI_AS_ROW:
	interp->result = "ROW";
	break;
    case MI_AS_ERROR:
	interp->result = "ERROR";
	break;
    case MI_AS_DEAD:
	interp->result = "DEAD";
	break;
    case MI_AS_UNKNOWN:
    default:
	interp->result = "unknown return from mi_processing_needed";
	return TCL_ERROR;
    }
    return TCL_OK;
}
