#include "MiTcl.h"

static int MiGetRowValues _ANSI_ARGS_ ((Tcl_Interp *interp,
    MI_ROW         *row, Tcl_DString    *dsp, int             type));


int
Tmi_exec(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    int             control = 0;
    MI_CONNECTION  *conn;

    if (argc != 3) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " connId query\"", (char *) NULL);
	return TCL_ERROR;
    }
    if (MiGetConn(interp, argv[1], &conn) != TCL_OK) {
	return TCL_ERROR;
    }
    if (mi_exec(conn, argv[2], control) == MI_ERROR) {
	Tcl_AppendResult(interp, "couldn't send query \"",
			 argv[2], "\"", (char *) NULL);
	return TCL_ERROR;
    }
    switch (mi_get_result(conn)) {
    case MI_ERROR:
	interp->result = "MI_ERROR";
	break;
    case MI_NO_MORE_RESULTS:
	interp->result = "MI_NO_MORE_RESULTS";
	break;
    case MI_ROWS:
	interp->result = "MI_ROWS";
	break;
    case MI_DML:
	interp->result = "MI_DML";
	break;
    case MI_DDL:
	interp->result = "MI_DDL";
	break;
    case MI_FUNCTION_RESULTS:
	interp->result = "MI_FUNCTION_RESULTS";
	break;
    default:
	interp->result = "unknown return from mi_get_result";
	return TCL_ERROR;
    }
    return TCL_OK;
}


int
Tmi_query_finish(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_CONNECTION  *conn;

    if (MiArgv1Conn(interp, argc, argv, &conn) != TCL_OK) {
	return TCL_ERROR;
    }
    if (mi_query_finish(conn) == MI_ERROR) {
	interp->result = "couldn't finish query";
	return TCL_ERROR;
    }
    return TCL_OK;
}


int
Tmi_query_interrupt(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_CONNECTION  *conn;
    int             block = 0;

    if ((argc != 2) && (argc != 3)) {
	Tcl_AppendResult(interp, "wrong # args:  should be \"",
			 argv[0], " connId ?block?\"", (char *) NULL);
	return TCL_ERROR;
    }
    if (MiGetConn(interp, argv[1], &conn) != TCL_OK) {
	return TCL_ERROR;
    }
    if (argc == 3) {
	if (Tcl_GetInt(interp, argv[2], &block) != TCL_OK) {
	    return TCL_ERROR;
	}
    }
    if (mi_query_interrupt(conn, block) == MI_ERROR) {
	interp->result = "couldn't interrupt query";
	return TCL_ERROR;
    }
    return TCL_OK;
}


int
Tmi_command_is_finished(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_CONNECTION  *conn;

    if (MiArgv1Conn(interp, argc, argv, &conn) != TCL_OK) {
	return TCL_ERROR;
    }
    switch (mi_command_is_finished(conn)) {
    case 0:
	interp->result = "0";
	break;
    case 1:
	interp->result = "1";
	break;
    case MI_ERROR:
	interp->result = "could not tell if command is finished";
	return TCL_ERROR;
    default:
	interp->result = "unknown return from mi_command_is_finished";
	return TCL_ERROR;
    }
    return TCL_OK;
}


int
Tmi_result_command_name(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_CONNECTION  *conn;

    if (MiArgv1Conn(interp, argc, argv, &conn) != TCL_OK) {
	return TCL_ERROR;
    }
    interp->result = mi_result_command_name(conn);
    if (interp->result == (char *) NULL) {
	interp->result = "could not get last command";
	return TCL_ERROR;
    }
    return TCL_OK;
}


int
Tmi_result_object_id(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_CONNECTION  *conn;

    if (MiArgv1Conn(interp, argc, argv, &conn) != TCL_OK) {
	return TCL_ERROR;
    }
    if (mi_result_object_id(conn, interp->result, TCL_RESULT_SIZE)
	== MI_ERROR) {
	interp->result = "could not get result object id!";
	return TCL_ERROR;
    }
    return TCL_OK;
}


int
Tmi_result_row_count(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_CONNECTION  *conn;
    int             n;

    if (MiArgv1Conn(interp, argc, argv, &conn) != TCL_OK) {
	return TCL_ERROR;
    }
    n = mi_result_row_count(conn);
    if (n == MI_ERROR) {
	interp->result = "could not get number of affected rows";
	return TCL_ERROR;
    }
    sprintf(interp->result, "%d", n);
    return TCL_OK;
}


int
Tmi_column_count(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_ROW_DESC    *rdesc;
    int             n;

    if (argc != 2) {
	Tcl_AppendResult(interp, "wrong # args:  should be \"",
			 argv[0], " rdescId\"", (char *) NULL);
	return TCL_ERROR;
    }
    if (MiGetRdesc(interp, argv[1], &rdesc) != TCL_OK) {
	return TCL_ERROR;
    }
    n = mi_column_count(rdesc);
    if (n == MI_ERROR) {
	interp->result = "could not get column count";
	return TCL_ERROR;
    }
    sprintf(interp->result, "%d", n);
    return TCL_OK;
}


int
Tmi_column_id(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_ROW_DESC    *rdesc;
    int             n;

    if (argc != 3) {
	Tcl_AppendResult(interp, "wrong # args:  should be \"",
			 argv[0], " rdescId colname \"", (char *) NULL);
	return TCL_ERROR;
    }
    if (MiGetRdesc(interp, argv[1], &rdesc) != TCL_OK) {
	return TCL_ERROR;
    }
    n = mi_column_id(rdesc, argv[2]);
    if (n == MI_ERROR) {
	Tcl_AppendResult(interp, "could not get column id for \"",
			 argv[2], "\"", (char *) NULL);
	return TCL_ERROR;
    }
    sprintf(interp->result, "%d", n);
    return TCL_OK;
}


int
Tmi_column_is_composite(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_ROW_DESC    *rdesc;
    int             n;

    if (MiArgv2ColId(interp, argc, argv, &rdesc, &n) != TCL_OK) {
	return TCL_ERROR;
    }
    switch (mi_column_is_composite(rdesc, n)) {
    case 0:
	interp->result = "0";
	break;
    case 1:
	interp->result = "1";
	break;
    case MI_ERROR:
	interp->result = "could not determine if column is composite";
	return TCL_ERROR;
	break;
    default:
	interp->result = "unknown return from mi_column_is_composite";
	return TCL_ERROR;
	break;
    }
    return TCL_OK;
}


int
Tmi_column_name(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_ROW_DESC    *rdesc;
    int             n;

    if (MiArgv2ColId(interp, argc, argv, &rdesc, &n) != TCL_OK) {
	return TCL_ERROR;
    }
    interp->result = mi_column_name(rdesc, n);
    if (interp->result == (char *) NULL) {
	interp->result = "could not get column name";
	return TCL_ERROR;
    }
    return TCL_OK;
}


int
Tmi_column_type_id(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_ROW_DESC    *rdesc;
    int             n;

    if (MiArgv2ColId(interp, argc, argv, &rdesc, &n) != TCL_OK) {
	return TCL_ERROR;
    }
    interp->result = mi_column_type_id(rdesc, n);
    if (interp->result == (char *) NULL) {
	interp->result = "could not get column type id";
	return TCL_ERROR;
    }
    return TCL_OK;
}


int
Tmi_column_type_length(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_ROW_DESC    *rdesc;
    int             n;
    int             len;

    if (MiArgv2ColId(interp, argc, argv, &rdesc, &n) != TCL_OK) {
	return TCL_ERROR;
    }
    len = mi_column_type_length(rdesc, n);
    if (len == MI_ERROR) {
	interp->result = "could not get column length";
	return TCL_ERROR;
    }
    sprintf(interp->result, "%d", len);
    return TCL_OK;
}


int
Tmi_column_type_name(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_ROW_DESC    *rdesc;
    int             n;

    if (MiArgv2ColId(interp, argc, argv, &rdesc, &n) != TCL_OK) {
	return TCL_ERROR;
    }
    interp->result = mi_column_type_name(rdesc, n);
    if (interp->result == (char *) NULL) {
	interp->result = "could not get column type name";
	return TCL_ERROR;
    }
    return TCL_OK;
}


int
Tmi_get_row_desc(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_ROW         *row;
    MI_ROW_DESC    *rdesc;

    if (argc != 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " rowId\"", (char *) NULL);
	return TCL_ERROR;
    }
    if (MiGetRow(interp, argv[1], &row) != TCL_OK) {
	return TCL_ERROR;
    }
    rdesc = mi_get_row_desc(row);
    if (rdesc == (MI_ROW_DESC *) NULL) {
	interp->result = "couldn't get row descriptor";
	return TCL_ERROR;
    }
    MiSetId(interp->result, (void *) rdesc);
    return TCL_OK;
}


int
Tmi_get_row_desc_without_row(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_CONNECTION  *conn;
    MI_ROW_DESC    *rdesc;

    if (MiArgv1Conn(interp, argc, argv, &conn) != TCL_OK) {
	return TCL_ERROR;
    }
    rdesc = mi_get_row_desc_without_row(conn);
    if (rdesc == (MI_ROW_DESC *) NULL) {
	interp->result = "couldn't get row descriptor";
	return TCL_ERROR;
    }
    MiSetId(interp->result, (void *) rdesc);
    return TCL_OK;
}


int
Tmi_composite_column_desc(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_ROW_DESC    *rdesc;
    MI_ROW_DESC    *crdesc;
    int             n;

    if (MiArgv2ColId(interp, argc, argv, &rdesc, &n) != TCL_OK) {
	return TCL_ERROR;
    }
    crdesc = mi_composite_column_desc(rdesc, n);
    if (crdesc == (MI_ROW_DESC *) NULL) {
	interp->result = "could not get composite column row descriptor";
	return TCL_ERROR;
    }
    MiSetId(interp->result, (void *) crdesc);
    return TCL_OK;
}


int
Tmi_next_row(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_CONNECTION  *conn;
    MI_ROW         *row;
    int             error;

    if (MiArgv1Conn(interp, argc, argv, &conn) != TCL_OK) {
	return TCL_ERROR;
    }
    row = mi_next_row(conn, &error);
    if (row == (MI_ROW *) NULL) {
	switch (error) {
	case 0:
	    return TCL_OK;
	case MI_ERROR:
	    interp->result = "couldn't get row descriptor";
	    return TCL_ERROR;
	default:
	    interp->result = "unknown error code from mi_next_row";
	    return TCL_ERROR;
	}
    }
    MiSetId(interp->result, (void *) row);
    return TCL_OK;
}


int
Tmi_value(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_ROW         *row;
    int             n;
    int             len;
    char           *val;

    if (argc != 4) {
	Tcl_AppendResult(interp, "wrong # args:  should be \"",
			 argv[0], " rowId colId var\"", (char *) NULL);
	return TCL_ERROR;
    }
    if (MiGetRow(interp, argv[1], &row) != TCL_OK) {
	return TCL_ERROR;
    }
    if (Tcl_GetInt(interp, argv[2], &n) != TCL_OK)
	return TCL_ERROR;

    switch (mi_value(row, n, &val, &len)) {
    case MI_NULL_VALUE:
	interp->result = "NULL";
	break;
    case MI_NORMAL_VALUE:
	interp->result = "NORMAL";
	break;
    case MI_ROW_VALUE:
	MiSetId(interp->result, (void *) val);
	val = interp->result;
	interp->result = "ROW";
	break;
    case MI_ERROR:
	interp->result = "could not get value";
	return TCL_ERROR;
    default:
	interp->result = "unknown return from mi_value";
	return TCL_ERROR;
    }
    Tcl_SetVar(interp, argv[3], val, 0);
    return TCL_OK;
}


int
Tmi_get_row_values(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    Tcl_DString     dsp;
    MI_ROW         *row;
    int             type;

    if (argc != 3) {
	Tcl_AppendResult(interp, "wrong # args:  should be \"",
			 argv[0], " rowId flat\"", (char *) NULL);
	return TCL_ERROR;
    }
    if (MiGetRow(interp, argv[1], &row) != TCL_OK) {
	return TCL_ERROR;
    }
    if (Tcl_GetInt(interp, argv[2], &type) != TCL_OK)
	return TCL_ERROR;
    switch (type) {
    case 0:
	type = GROUPED_ROW;
	break;
    case 1:
	type = FLAT_ROW;
	break;
    default:
	sprintf(interp->result, "unknown type:  %d", type);
	return TCL_ERROR;
    }

    Tcl_DStringInit(&dsp);
    if (MiGetRowValues(interp, row, &dsp, type) != TCL_OK) {
	Tcl_DStringFree(&dsp);
	return TCL_ERROR;
    }
    Tcl_DStringResult(interp, &dsp);
    return TCL_OK;
}


static int
MiGetRowValues(interp, row, dsp, type)
    Tcl_Interp     *interp;
    MI_ROW         *row;
    Tcl_DString    *dsp;
    int             type;
{
    int             ncols;
    int             i;
    char           *val;
    int             len;

    ncols = mi_column_count(mi_get_row_desc(row));
    for (i = 0; i < ncols; ++i) {
	switch (mi_value(row, i, &val, &len)) {
	case MI_NULL_VALUE:
	    Tcl_DStringAppendElement(dsp, "(null)");
	    break;
	case MI_NORMAL_VALUE:
	    Tcl_DStringAppendElement(dsp, val);
	    break;
	case MI_ROW_VALUE:
	    switch (type) {
	    case GROUPED_ROW:
		{
		    Tcl_DString     ndsp;
		    Tcl_DStringInit(&ndsp);
		    if (MiGetRowValues(interp, (MI_ROW *) val, &ndsp, type) !=
			TCL_OK) {
			Tcl_DStringFree(&ndsp);
			return TCL_ERROR;
		    }
		    Tcl_DStringResult(interp, &ndsp);
		    Tcl_DStringAppendElement(dsp, interp->result);
		}
		break;
	    case FLAT_ROW:
		if (MiGetRowValues(interp, (MI_ROW *) val, dsp, type) !=
		    TCL_OK)
		    return TCL_ERROR;
		break;
	    default:
		interp->result = "weird type of row";
		return TCL_ERROR;
	    }
	    break;
	case MI_ERROR:
	    interp->result = "can't get value";
	    return TCL_ERROR;
	default:
	    interp->result = "unknow return from mi_value";
	    return TCL_ERROR;
	}
    }
    return TCL_OK;
}
