#include "MiTcl.h"

static int MiSSGet _ANSI_ARGS_((Tcl_Interp *, int, char **, MI_ROW *(*)()));


int
Tmi_save_set_count(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_SAVE_SET    *ss;
    int		    num;

    if (argc != 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " ssId\"", (char *) NULL);
	return TCL_ERROR;
    }
    if (MiGetSS(interp, argv[1], &ss) != TCL_OK) {
	return TCL_ERROR;
    }
    num = mi_save_set_count(ss);
    if (num == MI_ERROR) {
	interp->result = "couldn't count save set";
	return TCL_ERROR;
    }
    sprintf(interp->result, "%d", num);
    return TCL_OK;
}


int
Tmi_save_set_create(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_CONNECTION  *conn;
    MI_SAVE_SET    *ss;

    if (argc != 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " connId\"", (char *) NULL);
	return TCL_ERROR;
    }
    if (MiGetConn(interp, argv[1], &conn) != TCL_OK) {
	return TCL_ERROR;
    }
    ss = mi_save_set_create(conn);
    if (ss == NULL) {
	interp->result = "couldn't create save set";
	return TCL_ERROR;
    }
    MiSetId(interp->result, (void *) ss);
    return TCL_OK;
}


int
Tmi_save_set_delete(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_ROW    *row;

    if (argc != 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " rowId\"", (char *) NULL);
	return TCL_ERROR;
    }
    if (MiGetRow(interp, argv[1], &row) != TCL_OK) {
	return TCL_ERROR;
    }
    if (mi_save_set_delete(row) == MI_ERROR) {
	interp->result = "couldn't delete row from save set";
	return TCL_ERROR;
    }
    return TCL_OK;
}


int
Tmi_save_set_destroy(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_SAVE_SET    *ss;

    if (argc != 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " ssId\"", (char *) NULL);
	return TCL_ERROR;
    }
    if (MiGetSS(interp, argv[1], &ss) != TCL_OK) {
	return TCL_ERROR;
    }
    if (mi_save_set_destroy(ss) == MI_ERROR) {
	interp->result = "couldn't destroy save set";
	return TCL_ERROR;
    }
    return TCL_OK;
}


int
Tmi_save_set_get_first(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    return MiSSGet(interp, argc, argv, mi_save_set_get_first);
}


int
Tmi_save_set_get_next(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    return MiSSGet(interp, argc, argv, mi_save_set_get_next);
}


int
Tmi_save_set_get_previous(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    return MiSSGet(interp, argc, argv, mi_save_set_get_previous);
}


static int
MiSSGet(interp, argc, argv, func)
    Tcl_Interp     *interp;
    int		    argc;
    char	  **argv;
    MI_ROW	  *(*func)();
{
    MI_SAVE_SET    *ss;
    MI_ROW	   *row;
    int		    error;

    if (argc != 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " ssId\"", (char *) NULL);
	return TCL_ERROR;
    }
    if (MiGetSS(interp, argv[1], &ss) != TCL_OK) {
	return TCL_ERROR;
    }
    row = (*func)(ss, &error);
    if (row == NULL) {
	if (error == MI_ERROR) {
	    interp->result = "couldn't get save set row";
	    return TCL_ERROR;
	}
    }
    else {
	MiSetId(interp->result, (void *) row);
    }
    return TCL_OK;
}


int
Tmi_save_set_insert(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_SAVE_SET    *ss;
    MI_ROW         *row;

    if (argc != 3) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " ssId rowId\"", (char *) NULL);
	return TCL_ERROR;
    }
    if (MiGetSS(interp, argv[1], &ss) != TCL_OK) {
	return TCL_ERROR;
    }
    if (MiGetRow(interp, argv[2], &row) != TCL_OK) {
	return TCL_ERROR;
    }
    row = mi_save_set_insert(ss, row);
    if (row == NULL) {
	interp->result = "couldn't insert row into save set";
	return TCL_ERROR;
    }
    MiSetId(interp->result, (void *) row);
    return TCL_OK;
}


int
Tmi_save_set_member(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_ROW         *row;

    if (argc != 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
			 argv[0], " rowId\"", (char *) NULL);
	return TCL_ERROR;
    }
    if (MiGetRow(interp, argv[1], &row) != TCL_OK) {
	return TCL_ERROR;
    }
    switch (mi_save_set_member(row)) {
    case 0:
	interp->result = "0";
	break;
    case 1:
	interp->result = "1";
	break;
    case MI_ERROR:
	interp->result = "couldn't test if row is a member of a save set";
	return TCL_ERROR;
	break;
    default:
	interp->result = "strange result from mi_save_set_member";
	return TCL_ERROR;
	break;
    }
    return TCL_OK;
}
