#include "MiTcl.h"


/* return the MITCL_LIBRARY enviornment variable or compiled in default */
int
Tmi_library(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    char           *getenv();
    char           *lib;

    lib = getenv("MITCL_LIBRARY");
    if (lib == NULL) {
	lib = MITCL_LIBRARY;
    }
    Tcl_SetResult(interp, lib, TCL_VOLATILE);
    return TCL_OK;
}


/* return the Illustra libmi library version */
int
Tmi_library_version(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    if (mi_library_version(interp->result, TCL_RESULT_SIZE) == MI_ERROR) {
	interp->result = "could not get libmi version";
	return TCL_ERROR;
    }
    return TCL_OK;
}


int
Tmi_exc_handled(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_CONNECTION  *conn;

    if (MiArgv1Conn(interp, argc, argv, &conn) != TCL_OK) {
	return TCL_ERROR;
    }
    mi_exc_handled(conn);
    return TCL_OK;
}


int
Tmi_unix_connection(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_CONNECTION  *conn;
    int             fd;

    if (MiArgv1Conn(interp, argc, argv, &conn) != TCL_OK) {
	return TCL_ERROR;
    }
    fd = mi_unix_connection(conn);
    if (fd == MI_ERROR) {
	interp->result = "couldn't get Unix file descriptor for connection";
	return TCL_ERROR;
    }
    sprintf(interp->result, "file%d", fd);
    return TCL_OK;
}


int
Tmi_event_pending(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    MI_CONNECTION  *conn;
    int             events;

    if (MiArgv1Conn(interp, argc, argv, &conn) != TCL_OK) {
	return TCL_ERROR;
    }
    events = mi_event_pending(conn);
    if (events == MI_ERROR) {
	interp->result = "could not check for events!";
	return TCL_ERROR;
    }
    sprintf(interp->result, "%d", events);
    return TCL_OK;
}


int
Tmi_db_error_raise(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    int             msg_type;
    MI_CONNECTION  *conn;
    mi_text        *p1 = (mi_text *) NULL;
    mi_text        *p2 = (mi_text *) NULL;
    mi_text        *p3 = (mi_text *) NULL;

    /* NB:  we only allow two optional parameters for now */
    if ((argc != 4) && (argc != 6)) {
	Tcl_AppendResult(interp, "wrong # args:  should be \"",
			 argv[0], " connId msgType msg ?p1 p2?\"",
			 (char *) NULL);
	return TCL_ERROR;
    }
    if (MiGetConn(interp, argv[1], &conn) != TCL_OK) {
	return TCL_ERROR;
    }
    if (STREQ(argv[2], "MESSAGE")) {
	msg_type = MI_MESSAGE;
    }
    else if (STREQ(argv[2], "EXCEPTION")) {
	msg_type = MI_EXCEPTION;
    }
    else if (STREQ(argv[2], "FATAL")) {
	msg_type = MI_FATAL;
    }
    else if (STREQ(argv[2], "SQL")) {
	msg_type = MI_SQL;
    }
    else {
	Tcl_AppendResult(interp, "bad msgType \"", argv[2],
			 "\":  should be MESSAGE, EXCEPTION, FATAL, or SQL",
			 (char *) NULL);
	return TCL_ERROR;
    }

    if (argc == 6) {
	p1 = mi_string_to_text(argv[4]);
	p2 = mi_string_to_text(argv[5]);
    }
    mi_db_error_raise(conn, msg_type, argv[3], p1, p2, p3);
    if (argc == 6) {
	mi_free((char *) p1);
	mi_free((char *) p2);
    }
    return TCL_OK;
}


int
Tmi_getsysparam(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    char           *server;

    if ((argc != 2) && (argc != 3)) {
	Tcl_AppendResult(interp, "wrong # args:  should be \"",
			 argv[0], " name ?server?\"", (char *) NULL);
	return TCL_ERROR;
    }
    if (argc == 3) {
	server = argv[2];
    }
    else {
	server = mi_sysname((char *) NULL);
	if (server == NULL) {
	    interp->result = "can't determine current system";
	    return TCL_ERROR;
	}
    }
    interp->result = mi_getsysparam(server, argv[1]);
    if (interp->result == NULL) {
	interp->result = "can't determine parameter value";
	return TCL_ERROR;
    }
    return TCL_OK;
}


int
Tmi_client(dummy, interp, argc, argv)
    ClientData      dummy;
    Tcl_Interp     *interp;
    int             argc;
    char          **argv;
{
    switch (mi_client()) {
    case 1:
	interp->result = "1";
	break;
    case 0:
	interp->result = "0";
	break;
    default:
	interp->result = "string result from mi_client";
	return TCL_ERROR;
	break;
    }
    return TCL_OK;
}


int
Mi_Init(interp)
    Tcl_Interp     *interp;
{
    static char     initCmd[] =
    "if [file exists [mi_library]/init.tcl] {\n\
            source [mi_library]/init.tcl\n\
        } else {\n\
            error \"can't find LIBMI init file:  [mi_library]/init.tcl\"\n\
        }";

    /* create the commands */
#include "MiInits.cpp"

    /* initialize the library */
    return Tcl_Eval(interp, initCmd);
}
